package cpmodel;

import choco.integer.IntDomainVar;
import choco.integer.search.AssignVar;
import choco.integer.search.DomOverDeg;
import choco.integer.search.IncreasingDomain;
import choco.integer.search.MinDomain;
import choco.Solver;
import choco.Constraint;
import choco.Propagator;
import choco.global.BoundAllDiff;
import choco.global.matching.GlobalCardinalityVar;
import cpmodel.constraints.SimpleInverse;
import cpmodel.constraints.ModelChannel;
import cpmodel.constraints.Colouring;
import tools.InstanceReader;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.BitSet;
import java.util.ArrayList;
import java.util.Iterator;

import instance.TimetablingInstance;
import instance.clique.MaxCliqueFinder;
import instance.clique.CompoundClique;
import magicsearch.impact.ImpactStrategie;
import magicsearch.impact.Iref;
import magicsearch.impact.impactMAC.IBSProblem;
import magicsearch.impact.impactMAC.ImpactBasedVarSearch;
import magicsearch.impact.impactMAC.ImpactBasedValSearch;
import magicsearch.impact.impactMAC.IBSAssignVar;
import magicsearch.domwdeg.domwdegMAC.DWDProblem;
import magicsearch.domwdeg.domwdegMAC.DWDAssignVar;
import magicsearch.domwdeg.DomOverWdeg;
import magicsearch.domwdeg.DWDPlugin;

/**
 * The model that maintains the matching as a gcc
 * and only need to branch on the et variables
 */
public class HeavyModel extends PrimalModel {

	/**
	 * the number of events assigned in each timeslot
	 */
	protected IntDomainVar[] cardTime;

	/**
	 * The number of empty rooms in each timeslot
	 */
	protected IntDomainVar[] availableSpace;

	/**
	 * boolean variable indicating if event i has been placed in timeslot ech
	 */
	protected IntDomainVar[][] bik;

	/**
	 * roomEvent[i][j] is a variable to give the event assigned in room j
	 * of timeslot j.
	 */
	protected IntDomainVar[][] roomEvent;


	public HeavyModel(String s) {
		super(s);
	}

	public void init(String s) {
		name = s;
		InstanceReader reader = null;
		try {
			reader = new InstanceReader(new FileInputStream(s));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		ins = reader.getInstance();
		ins.setName(s);

		timetable = new int[nHours * nDays][ins.getNRooms()];
		solution = new int[ins.getNEvents()];
	}

	public void postConflicts(int coltype) {
		if(heuristictype == DomWDeg &&
				(coltype == CONFLICTS || coltype == R_ALLDIFF) ) {
			// 1. Events that share a student cannot coincide in time
			for (int i = 0; i < ins.getNEvents() - 1; i++)
				for (int j = i + 1; j < ins.getNEvents(); j++) {
					if (ins.eventsShareStudents(i, j)) {
						pb.post(pb.neq(eventTime[i], eventTime[j]));
					}
				}
		} else super.postConflicts(coltype);
	}


	public void buildMacthing2() {
		// Takes care of the rooms allocation through gcc constraints
		// MatchingInterface [rooms] x [events] for each time slot
		IntDomainVar[][] roomEvent = new IntDomainVar[nDays * nHours][ins.getNRooms()];
		for (int i = 0; i < nDays * nHours; i++) {
			for (int j = 0; j < ins.getNRooms(); j++) {
				roomEvent[i][j] = pb.makeEnumIntVar("Re" + i + " " + j, 0, ins.getNEvents());
				for (int k = 0; k < ins.getNEvents(); k++) {
					pb.post(new SimpleInverse(
							roomEvent[i][j],
							k + 1,
							event[k],
							(j + 1) + i * ins.getNRooms()));
				}
			}
		}
		cardTime = new IntDomainVar[nDays * nHours];
		availableSpace = new IntDomainVar[nDays * nHours];

		for (int i = 0; i < nDays * nHours; i++) {
			cardTime[i] = pb.makeBoundIntVar("Ct" + i, 0, ins.getNRooms());
			availableSpace[i] = pb.makeBoundIntVar("Avs" + i, 0, ins.getNRooms());
			pb.post(pb.occurrence(eventTime, (i + 1), cardTime[i]));
			pb.post(pb.eq(availableSpace[i], pb.minus(ins.getNRooms(), cardTime[i])));
		}

		//cardinality variable and the corresponding channeling
		IntDomainVar[][] bik = new IntDomainVar[ins.getNEvents()][nDays * nHours];
		for (int i = 0; i < nDays * nHours; i++) {
			for (int j = 0; j < ins.getNEvents(); j++) {
				
				//System.out.println( eventTime[j].getDomain() );
				
				if( eventTime[j].canBeInstantiatedTo(i+1) )
				{
					bik[j][i] = pb.makeEnumIntVar("b_" + i + "," + j, 0, 1);
					pb.post(pb.boolChanneling(bik[j][i],eventTime[j],i+1));
				}
				else
				{
					bik[j][i] = pb.makeEnumIntVar("b_" + i + "," + j, 0, 0);
				}
//				bik[j][i] = pb.makeEnumIntVar("b_" + i + "," + j, 0, 1);
//				pb.post(pb.boolChanneling(bik[j][i],eventTime[j],i+1));
			}
		}
	

		for (int i = 0; i < nDays *nHours; i++) {
			IntDomainVar[] occ = new IntDomainVar[ins.getNEvents() + 1];
			occ[0] = availableSpace[i];
			for (int j = 0; j < ins.getNEvents(); j++) {
				occ[j+1] = bik[j][i];
			}
			pb.post(new GlobalCardinalityVar(roomEvent[i],0,ins.getNEvents(),occ));
		}
	}
	
	
	public void buildMacthing() {
		// Takes care of the rooms allocation through gcc constraints
		// MatchingInterface [rooms] x [events] for each time slot
		IntDomainVar[][] roomEvent = new IntDomainVar[nDays * nHours][ins.getNRooms()];
		for (int i = 0; i < nDays * nHours; i++) {
			for (int j = 0; j < ins.getNRooms(); j++) {
				roomEvent[i][j] = pb.makeEnumIntVar("Re" + i + " " + j, 0, ins.getNEvents());
				for (int k = 0; k < ins.getNEvents(); k++) {
					pb.post(new SimpleInverse(
							roomEvent[i][j],
							k + 1,
							event[k],
							(j + 1) + i * ins.getNRooms()));
				}
			}
		}

		//cardinality variable and the corresponding channeling
		IntDomainVar[][] bik = new IntDomainVar[ins.getNEvents()][nDays * nHours];
		for (int i = 0; i < nDays * nHours; i++) {
			for (int j = 0; j < ins.getNEvents(); j++) {
				if( eventTime[j].canBeInstantiatedTo(i+1) )
				{
					bik[j][i] = pb.makeEnumIntVar("b_" + i + "," + j, 0, 1);
					pb.post(pb.boolChanneling(bik[j][i],eventTime[j],i+1));
				}
				else
				{
					bik[j][i] = pb.makeEnumIntVar("b_" + i + "," + j, 0, 0);
				}
			}
		}
	

		for (int i = 0; i < nDays *nHours; i++) {
			IntDomainVar[] occ = new IntDomainVar[ins.getNEvents()+1];
			occ[0] = pb.makeBoundIntVar("nom", 0, ins.getNRooms());
			for (int j = 0; j < ins.getNEvents(); j++) {
				occ[j+1] = bik[j][i];
			}
			pb.post(new GlobalCardinalityVar(roomEvent[i],0,ins.getNEvents(),occ));
		}
	}


	public void setHeuristic() {
		Solver s = pb.getSolver();
		s.generateSearchSolver(pb);
		if (heuristictype == Impact) {
			restart = false;
			initialization = false;
			ImpactStrategie ibs = new Iref(pb,eventTime);
			((IBSProblem) pb).setIbs(ibs);
			if (initialization) {
				long tps = System.currentTimeMillis();
				((IBSProblem) pb).initImpact(timelimit);
				inittime = (int) (System.currentTimeMillis() - tps);
				System.out.println("initialization time " + inittime);
			}

			s.attachGoal(new IBSAssignVar(
								new ImpactBasedVarSearch(eventTime, ibs),
						        new ImpactBasedValSearch(ibs)));
			s.addGoal(new AssignVar(new MinDomain(pb,eventRoom),
									new IncreasingDomain()));
		} else if (heuristictype == DomWDeg) {
			s.attachGoal(new DWDAssignVar(new DomOverWdeg(pb,eventTime),new IncreasingDomain()));
			s.addGoal(new DWDAssignVar(new DomOverWdeg(pb, eventRoom),
									   new IncreasingDomain()));
		} else {
			s.attachGoal(new AssignVar(new DomOverDeg(pb, eventTime), new IncreasingDomain()));
			s.addGoal(new AssignVar(new MinDomain(pb, eventRoom), new IncreasingDomain()));						
		}
	}

	public void solve() {
		super.solve();
		//Solver.setVerbosity(Solver.SEARCH);
		Solver s = pb.getSolver();
		//s.setLoggingMaxDepth(1000);
		setHeuristic();
		s.launch();
		
//		if (pb.isFeasible()) {
//			checkSolution();
//		}
//		//System.out.println("NB solution found " + pb.getSolver().getNbSolutions() + " " + this.getTime());
	}

	public void seeWeights() {
		if (heuristictype == DomWDeg) {
			Iterator it = pb.getIntConstraintIterator();
			while(it.hasNext()) {
				Propagator c = (Propagator) it.next();
				DWDPlugin plug = (DWDPlugin) c.getPlugIn();
				System.out.println(c + " : " + plug.getWdeg());
			}
		}
	}

	/**
	 * build only the coloring, room allocation and matching
	 */
	public void buildModel() {
		super.buildModel();
		this.buildColoring(PrimalModel.CONFLICTS);
		this.buildAllocation();

		// The Links between coloring and allocation
		// 3. Two events cannot take place in the same room at the same time
		//pb.post(new BoundAllDiff(event, true)); 
		pb.post( pb.allDifferent(event) );
		// 6. Channeling between event, eventRoom and eventTime
		for (int i = 0; i < ins.getNEvents(); i++) {
			pb.post(new ModelChannel(event[i], eventRoom[i], eventTime[i], ins.getNRooms(), TimetablingInstance.N_TIMESLOTS));
		}

		//this.buildMacthing();
	}

	
}
