package cpmodel;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Comparator;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.complete.HardSolver;
import localsearch.solvers.complete.SearchLoop;
import localsearch.graphalgos.matchings.Matching;
import cpmodel.constraints.NotAllEqual;
import cpmodel.heuristics.MyValueSelector;
import cpmodel.heuristics.WeightedDomOverDeg;
import choco.Solver;
import choco.integer.IntDomainVar;
import magicsearch.impact.ImpactStrategie;
import magicsearch.impact.Iref;
import magicsearch.impact.impactMAC.IBSProblem;
import magicsearch.impact.impactMAC.ImpactBasedVarSearch;
import magicsearch.impact.impactMAC.ImpactBasedValSearch;
import magicsearch.impact.impactMAC.IBSSolver;
import instance.TimetablingInstance;


/**
 * The model that forgets the room constraints and will only
 * check the matching in the end
 */
public class LightModel extends PrimalModel {

	public boolean isFeasible = false;
	protected int softCost = Integer.MAX_VALUE;

	protected SortableVar[] SortedEvents;
	protected int[] roomSizes;
	protected int[] usedRooms;
	protected AbstractSolver local;


	public class SortableVar {
		public int value;
		public IntDomainVar X;

		public SortableVar(IntDomainVar x_, int v_) {
			X = x_;
			value = v_;
		}
	}

	public class EventComparator implements Comparator {
		public int compare(Object o1, Object o2) {
			int rvalue = 0;
			SortableVar s1 = (SortableVar) o1;
			SortableVar s2 = (SortableVar) o2;
			if (s1.value < s2.value) rvalue = -1;
			else if (s1.value > s2.value) rvalue = 1;
			return rvalue;
		}
	}

	public LightModel(String s, int m) {
		super(s);
		local = new HardSolver(s, 0);
		heuristictype = m;
		fastinit();
	}

	public LightModel(TimetablingInstance s, AbstractSolver solv, int m) {
		super(s);
		local = solv;
		heuristictype = m;
		fastinit();
	}

	public void fastinit() {
		usedRooms = new int[nHours * nDays];
		nIterations = 0;
		nNogoods = 0;
		malgo = new Matching[nHours * nDays];
	}

	public String toString() {
		switch (heuristictype) {
			case DODLEX:
				return "dom/deg + lex          ";
			case DODCON:
				return "dom/deg + contention   ";
			case CONLEX:
				return "contention + lex       ";
			case CONCON:
				return "contention + contention";
			case CODLEX:
				return "contention/deg + lex       ";
			case CODCON:
				return "contention/deg + contention";
			case DomWDeg:
				return "DomWdeg + lex              ";
			case Impact:
				return "Impact                     ";
			default:
				return "model type unknown";
		}
	}

	//eventSizeComparator()

	/**
	 * build only the coloring
	 */
	@SuppressWarnings("unchecked")
	public void buildModel(int coltype) {
		super.buildModel();
		this.buildColoring(coltype);

		// Maximum occurrence of event per time slot (bound gcc)
//		int[] lobound = new int[(nHours * nDays)];
//		int[] upbound = new int[(nHours * nDays)];
//		for (int i = 0; i < (nHours * nDays); i++) {
//			lobound[i] = 0;
//			upbound[i] = ins.getNRooms();
//		}
//		pb.post(pb.boundGcc(eventTime, 1, (nHours * nDays), lobound, upbound));

		// Maximum occurrence of event per time slot
		IntDomainVar maxRooms = pb.makeBoundIntVar("MR", ins.getNRooms(), ins.getNRooms());
		for (int i = 0; i < nHours * nDays; i++) {
			pb.post(pb.occurenceMax(eventTime, (i + 1), maxRooms));
		}

	}

	public void setParameters(int discrepency) {
		super.solve();
		Solver s = pb.getSolver();
		inittime = 0;
		//for logs
		//Solver.setVerbosity(Solver.SEARCH);
		//s.setLoggingMaxDepth(10000);

		if (heuristictype != Impact && heuristictype != DomWDeg) {
			restart = false;
			//s.generateSearchSolver(pb);

			WeightedDomOverDeg varo = new WeightedDomOverDeg(eventTime,
					ins.getNEvents(), nDays * nHours, ins.getNRooms());
			MyValueSelector valo = new MyValueSelector(varo.timeContention);

			if (colorct != null) { //todo : update this code once the graph structure is ready
				varo.initvgraph();
				for (int i = 0; i < eventTime.length; i++) {
					varo.setNbNeigbhours(i, colorct.getNbN(i));
				}
				for (int i = 0; i < ins.getNEvents() - 1; i++)
					for (int j = i + 1; j < ins.getNEvents(); j++)
						if (ins.eventsShareStudents(i, j)) {
							varo.addNeigbhours(i, j);
						}
			}

			switch (heuristictype) {
				case DODLEX:
					varo.setDod();
					valo.setLex();
					break;
				case DODCON:
					varo.setDod();
					valo.setCon();
					break;
				case CONLEX:
					varo.setCon();
					valo.setLex();
					break;
				case CONCON:
					varo.setCon();
					valo.setCon();
					break;
				case CODLEX:
					varo.setCod();
					valo.setLex();
					break;
				case CODCON:
					varo.setCod();
					valo.setCon();
					break;
				default:
					throw new Error("heuristic " + heuristictype + " unknown");
			}

			//s.attachGoal(new AssignVar(varo, valo));
			//s.launch();
			s.setVarSelector(varo);
			s.setValSelector(valo);
			//pb.solve();

		} else if (heuristictype == Impact) {
			initialization = true;
			restart = true;
			setAndInitializeImpact(discrepency);
		} else if (heuristictype == DomWDeg) {
			throw new Error("domWdeg is still to do");
		}
	}


	public void setParameters() {
		setParameters(45);
	}

	public void setAndInitializeImpact() {
		setAndInitializeImpact(45);
	}

	public void setAndInitializeImpact(int discrepency) {
		Solver s = pb.getSolver();
		if (((IBSProblem) pb).getIbs() == null) {
			ImpactStrategie ibs = new Iref(pb, eventTime);
			((IBSProblem) pb).setIbs(ibs);
			if (discrepency < 45)
				((IBSSolver) s).setLDS(discrepency);
			s.setVarIntSelector(new ImpactBasedVarSearch(eventTime, ibs));
			s.setValIntSelector(new ImpactBasedValSearch(ibs));
			if (initialization) {
				long tps = System.currentTimeMillis();
				((IBSProblem) pb).initImpact(timelimit);
				inittime = (int) (System.currentTimeMillis() - tps);
				//System.out.println("initialization time " + inittime);
			}
		} else {
			if (discrepency < 45)
				((IBSSolver) pb.getSolver()).setLDS(discrepency);			
		}
	}

	//avoid rebuilding the matching !
	protected Matching explainMatching;
	protected Matching[] malgo;


	public int explain(int[] conflict) {
		if (explainMatching == null)
			explainMatching = new Matching(local.space);
		else explainMatching.reinit();
		Matching m = explainMatching;


		int _size = 0;
		int oldsize;
		int aux;

		boolean not_in_conflict = true;

		while (_size == 0 || (not_in_conflict && m.solve())) {
			oldsize = _size;
			do {
				m.addEvent(conflict[_size++]);
			} while (m.solve());

			if (_size - 1 > oldsize) {
				aux = conflict[_size - 1];
				conflict[_size - 1] = conflict[oldsize];
				conflict[oldsize] = aux;

				m.removeInterval(oldsize, _size - 2);
			} else not_in_conflict = false;

			_size = oldsize + 1;
		}

		return _size;
	}

	public boolean allocateRooms() {

		//System.out.println( "Found a coloring, try to allocate rooms" );

		for (int i = 0; i < (nHours * nDays); ++i) {
			usedRooms[i] = 0;
			for (int j = 0; j < ins.getNRooms(); ++j) {
				timetable[i][j] = -1;
			}
		}

		for (int i = 0; i < ins.getNEvents(); ++i) {
			solution[i] = -1;
			int j = (eventTime[i].getVal() - 1);
			//System.out.println("evt " + i + " in " + j);			
			timetable[j][usedRooms[j]++] = i;
		}

		boolean canAllocate = true;

		for (int i = 0; i < nHours * nDays; ++i) {
			if (malgo[i] == null)
				malgo[i] = new Matching(local.space);
			else malgo[i].reinit();

			for (int j = 0; j < usedRooms[i]; ++j) {
				malgo[i].addEvent(timetable[i][j]);
			}
			if (malgo[i].solve()) {

				for (int j = 0; j < malgo[i].getLeftSize(); j++) {
					int e = malgo[i].getLeft().get(j).getRef();
					int r = malgo[i].getLeft().get(j).getMatchNode();
					timetable[i][r] = e;
					solution[e] = r;
				}

			} else {
				canAllocate = false;

				usedRooms[i] = explain(timetable[i]);
				IntDomainVar[] conflictingEvents = new IntDomainVar[usedRooms[i]];
				for (int j = 0; j < usedRooms[i]; ++j) {
					conflictingEvents[j] = eventTime[timetable[i][j]];
				}
				++nNogoods;
				pb.postCut(new NotAllEqual(conflictingEvents));
				//System.out.println("post not all equal " + Arrays.toString(conflictingEvents));
			}
		}

//		if( !canAllocate )
//			System.out.println( "Cannot Allocate Rooms" );

		return canAllocate;
	}


	public void solve() {

		setParameters();
		boolean timeout = false;

		if (restart) {
			restartpol = new RestartPolicy(pb, timelimit - inittime);
			timeout = (restartpol.restartSolve() == null);
		} else {
			pb.solve();
			timeout = pb.getSolver().isEncounteredLimit();
		}

		while (!timeout && !allocateRooms()) {
			++nIterations;
			timeout = (pb.nextSolution() == null);
		}


		isFeasible = !timeout;
	}


	public void checkSolution() {
		SearchLoop.searchLoopTrace = true;

		try {
			String noutput = name.substring(0, name.length() - 4);
			BufferedWriter bw = new BufferedWriter(new FileWriter(new File(noutput + ".sln")));

			for (int i = 0; i < ins.getNEvents(); i++) {
				int t = (eventTime[i].getVal() - 1);
				bw.write(t + " " + solution[i]);
				bw.newLine();
			}

			bw.flush();
			bw.close();
			if (softCost == Integer.MAX_VALUE)
				AbstractSolver.runCheckerFromJava(noutput, -1);
			else AbstractSolver.runCheckerFromJava(noutput, softCost);

		} catch (IOException e) {
			e.printStackTrace();
		}

	}

// CODE for redundant GCC 
//		int nE;
//		int nR;
//		int nFeatures = ins.getNFeatures();
//		for (int i = 0; i < nFeatures; ++i) {
//			nE = ins.eventsRequiringFeature(i).cardinality();
//			nR = ins.roomsSatisfyingFeature(i).cardinality();
//			if (nR < nE && nR < ins.getNRooms()) {
//				//System.out.println( "Gcc(" + nE + ", " + nR + ")" );
//				IntDomainVar[] eventFeature = new IntDomainVar[nE];
//				int j = 0;
//				int e = ins.eventsRequiringFeature(i).nextSetBit(0);
//				while (e >= 0) {
//					eventFeature[j++] = eventTime[e];
//					e = ins.eventsRequiringFeature(i).nextSetBit(e + 1);
//				}
//				for (int k = 0; k < (nHours * nDays); k++) {
//					upbound[k] = nR;
//				}
//				//pb.post(pb.boundGcc(eventFeature, 1, (nHours * nDays), lobound, upbound));
//			}
//		}
//
//		SortedEvents = new SortableVar[ins.getNEvents()];
//		roomSizes = new int[ins.getNRooms()];
//		for (int i = 0; i < ins.getNEvents(); ++i) {
//			SortedEvents[i] = new SortableVar(eventTime[i], ins.getNbStudentInEvent(i));
//		}
//		for (int i = 0; i < ins.getNRooms(); ++i) {
//			roomSizes[i] = ins.getRoomSize(i);
//		}
//
//		Arrays.sort(roomSizes);
//		Arrays.sort(SortedEvents, new EventComparator());
////		for(int i = 0; i < ins.getNEvents(); ++i) {
////			System.out.println( i + " " + SortedEvents[i].X + " (" + SortedEvents[i].value + ")");
////		}
////		for(int i = 0; i < ins.getNRooms(); ++i) {
////			System.out.println( i + " " + roomSizes[i] );
////		}
//
//		// smaller index of the larger event
//		int maxESI = ins.getNEvents() - 1;
//
//		// the index of the largest room that cannot accommodate
//		// the event at index maxESI.
//		int maxRSI = ins.getNRooms() - 1;
//
//		while (maxESI >= 0 && maxRSI >= 0) {
////
////			System.out.println( "=======> " + maxESI + " " + maxRSI + "   "
////					+ SortedEvents[maxESI].value + " " + roomSizes[maxRSI] );
////
//			// find out the largest room that cannot fit maxESI
//			while (maxESI >= 0 && maxRSI >= 0 &&
//					SortedEvents[maxESI].value <= roomSizes[maxRSI]) {
//				--maxRSI;
//			}
//			// find out the smallest event that cannot fit into the room at maxRSI;
//			while (maxESI >= 0 && maxRSI >= 0 &&
//					SortedEvents[maxESI].value > roomSizes[maxRSI]) {
//				--maxESI;
//			}
//
//			if ((ins.getNEvents() - maxESI) > (ins.getNRooms() - maxRSI)
//					&& maxRSI >= 0) {
////				System.out.println( "=======> " + maxESI + " " + maxRSI + "   "
////						+ SortedEvents[maxESI+1].value + " " + roomSizes[maxRSI+1] );
////
//				IntDomainVar[] scope = new IntDomainVar[ins.getNEvents() - maxESI - 1];
//				for (int i = maxESI + 1; i < ins.getNEvents(); ++i) {
//					scope[i - maxESI - 1] = SortedEvents[i].X;
//				}
//				//System.out.println( Arrays.toString(scope) + " <= " + (ins.getNRooms()-maxRSI-1) );
////				System.out.println( "[" + scope[0] + ".." + scope[ins.getNEvents()-maxESI-2]
////				                        + "] <= " + (ins.getNRooms()-maxRSI-1) );
////
//				nR = (ins.getNRooms() - maxRSI - 1);
//				for (int k = 0; k < (nHours * nDays); k++) {
//					upbound[k] = nR;
//				}
//
//				//pb.post(pb.boundGcc(scope, 1, (nDays * nHours), lobound, upbound));
//			}
//		}
//
////		IntDomainVar[][] eventFeature = new IntDomainVar[nFeatures][];
////		for(int i = 0; i < nFeatures; ++i)
//
////		ins.eventRequiresFeature(event, feature)
////		ins.getNbStudentInEvent(event)
////		ins.featuresSatisfiedByRoom(room)
////		ins.featuresRequiredByEvent(event)
////		ins.getRoomSize(room)
////		ins.roomsCompatibleWithEvent(event)

}
