package cpmodel;


import instance.Precedence;
import instance.TimetablingInstance;
import instance.clique.CompoundClique;
import instance.clique.MaxCliqueFinder;


import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.BitSet;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.complete.SearchLoop;
import tools.InstanceReader;
import util.IntSet;
import choco.ContradictionException;
import choco.Problem;
import choco.Solver;
import choco.integer.IntDomainVar;
import choco.integer.search.AssignVar;
import choco.integer.search.DomOverDeg;
import choco.integer.search.IncreasingDomain;
import choco.integer.search.MinDomain;
import cpmodel.chocoext.TimeTablingProblem;
import cpmodel.constraints.Colouring;
import magicsearch.domwdeg.domwdegMAC.DWDProblem;
import magicsearch.impact.impactMAC.IBSProblem;


/**
 * An abstract class to propose some common services needed
 * by all models
 */
public abstract class PrimalModel {

	// stats
	protected int nIterations;
	protected int nNogoods;

	//heuristic type
	public final static int DODLEX = 0;
	public final static int DODCON = 1;
	public final static int CONLEX = 2;
	public final static int CONCON = 3;
	public final static int CODLEX = 4;
	public final static int CODCON = 5;
	public final static int DomWDeg = 6;
	public final static int Impact = 7;


	public final static int CONFLICTS = 0;
	public final static int R_CONFLCT = 1;
	public final static int R_ALLDIFF = 2;
	public final static int A_ALLDIFF = 3;

	public int heuristictype = 0;
	protected boolean initialization = false;
	protected int inittime;
	protected int startTime;
	protected boolean restart = false;
	protected RestartPolicy restartpol;
	protected int timelimit = 300000;

	//instance
	protected TimetablingInstance ins;
	protected int nDays;
	protected int nHours;
	protected String name;

	//choco data
	protected Problem pb;
	protected IntDomainVar[] event;
	protected IntDomainVar[] eventRoom;
	protected IntDomainVar[] eventTime;

	//solution
	protected int timetable[][];
	protected int solution[];

	public PrimalModel() {
		this("noname");
	}

	public PrimalModel(String s) {
		startTime = (int) System.currentTimeMillis();
		InstanceReader reader = null;
		try {
			reader = new InstanceReader(new FileInputStream(s));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		ins = reader.getInstance();
		ins.setName(s);
		init(ins);
	}

	public PrimalModel(TimetablingInstance ins) {
		startTime = (int) System.currentTimeMillis();
		this.ins = ins;
		init(ins);
	}

	public void init(TimetablingInstance ins) {
		nDays = 5;
		nHours = 9;
		name = ins.getName();
		timetable = new int[nHours * nDays][ins.getNRooms()];
		solution = new int[ins.getNEvents()];
	}


    public Problem getChocoProblem() {
        return pb;
    }

    public IntDomainVar[] getEventTime() {
        return eventTime;
    }

    /**
	 * Build the corresponding model (medium, light, heavy)
	 */
	public void buildModel() {
		if (heuristictype == DomWDeg) {
			pb = new DWDProblem();
		} else if (heuristictype == Impact) {
			pb = new IBSProblem();
		} else {
			pb = new TimeTablingProblem();
		}
	}

	/**
	 * Different models use different heuristics
	 */
	public void solve() {
		System.out.println("start solve nbc: " + pb.getNbIntConstraints() + " nbv: " + pb.getNbIntVars() +" Tused:" + ((int) System.currentTimeMillis() - startTime) + " ms " + " Tleft: " + (timelimit - ((int) System.currentTimeMillis() - startTime)));
		pb.getSolver().setTimeLimit(timelimit - ((int) System.currentTimeMillis() - startTime));
	}


	public void setTimeLimit(int t) {
		timelimit = t;
	}


	//store the coloring constraint to remember the graph
	protected Colouring colorct;


	public void postConflicts(int coltype) {

		if (coltype != A_ALLDIFF) {

			// 1. Events that share a student cannot coincide in time
			colorct = new Colouring(eventTime);
			for (int i = 0; i < ins.getNEvents() - 1; i++)
				for (int j = i + 1; j < ins.getNEvents(); j++) {
					if (ins.eventsShareStudents(i, j)) {
						colorct.setIncomp(i, j);
						//pb.post(pb.neq(eventTime[i], eventTime[j]));
					}
				}
			pb.post(colorct);
		}

		if (coltype == R_CONFLCT) {
			// 1. Events that share a student cannot coincide in time
			colorct = new Colouring(eventTime);
			for (int i = 0; i < ins.getNEvents() - 1; i++)
				for (int j = i + 1; j < ins.getNEvents(); j++) {
					if (ins.eventsShareStudents(i, j)
							|| ins.isRoomConflictBetweenEvents(i, j)) {
						colorct.setIncomp(i, j);
						//pb.post(pb.neq(eventTime[i], eventTime[j]));
					}
				}
			pb.post(colorct);
		} else if (coltype == A_ALLDIFF) {
			int i;
			MaxCliqueFinder mc = new MaxCliqueFinder(ins);
			mc.findCliques(true);
			CompoundClique[] cliques = mc.getCliques();
			for (CompoundClique clique : cliques) {

				BitSet nodes = clique.retrieveNodes();

				IntDomainVar[] scope = new IntDomainVar[clique.getSize()];

				i = 0;
				for (int n = nodes.nextSetBit(0); n > -1; n = nodes.nextSetBit(n + 1)) {
					scope[i++] = eventTime[n];

					//System.out.print( n + " " );
				}
				pb.post(pb.boundAllDifferent(scope, true));

				//System.out.println();
			}
			//mc.printCliques();
			//System.out.println();
		} else if (coltype == R_ALLDIFF) {
			int i, j, k;
			MaxCliqueFinder mc = new MaxCliqueFinder(ins);
			IntSet criticalRooms = mc.getCriticalRooms();

			for (i = criticalRooms.first(); i > -1; i = criticalRooms.next(i)) {
				IntSet criticalEvents = mc.getEvents(i);
				if (criticalEvents.size() > 1) {
					IntDomainVar[] scope = new IntDomainVar[criticalEvents.size()];

					k = 0;
					for (j = criticalEvents.first(); j > -1; j = criticalEvents.next(j)) {
						scope[k++] = eventTime[j];
						//System.out.print( j + " " );
					}
					pb.post(pb.boundAllDifferent(scope, true));
					//System.out.println();
				}
				//mc.printCliques();
				//System.out.println();
			}
		}
	}


	/**
	 * build the coloring problematic common to all models involving
	 * the et variables
	 */
	public void buildColoring(int coltype) {
		// Takes care of the graph coloring + precedence aspect
		// The graph coloring
		eventTime = new IntDomainVar[ins.getNEvents()];
		for (int i = 0; i < ins.getNEvents(); i++) {
			eventTime[i] = pb.makeEnumIntVar("Et" + (i + 1), 1, nDays * nHours);
		}


		postConflicts(coltype);

		// 4. Lecturer's availability
		for (int i = 0; i < ins.getNEvents(); i++) {
			for (int j = 0; j < nDays * nHours; j++) {
				if (!ins.eventIsCompatibleWithTimeslot(i, j)) {
					try {
						eventTime[i].remVal(j + 1);
					} catch (ContradictionException e) {
						throw new Error("buildColoring: inconsistent initial model ??");
					}
				}
			}
		}
//        // 5. Precedences
		for (Precedence p : ins.eventPrecedences()) {
			pb.post(pb.lt(eventTime[p.firstEvent], eventTime[p.secondEvent]));
			//System.out.println(eventTime[p.firstEvent] + " < " + eventTime[p.secondEvent]);
		}
	}


	/**
	 * build the room allocation problematic needed by the Heavy and Medium model
	 */
	public void buildAllocation() {
		// Takes care of the rooms allocation through auxiliary vars and channeling
		// The room allocation
		eventRoom = new IntDomainVar[ins.getNEvents()];
		event = new IntDomainVar[ins.getNEvents()];
		for (int i = 0; i < ins.getNEvents(); i++) {
			event[i] = pb.makeEnumIntVar("E" + (i + 1), 1, ins.getNRooms() * nDays * nHours);
			eventRoom[i] = pb.makeEnumIntVar("Er" + (i + 1), 1, ins.getNRooms());
		}
		// 2. Room's features
		for (int i = 0; i < ins.getNEvents(); i++) {
			for (int j = 0; j < ins.getNRooms(); j++) {
				if (!ins.isRoomAvailableForEvent(i, j)) {
					//pb.post(pb.neq(eventRoom[i], j + 1));
					try {
						//for(int k = 0; k<45; ++k)
						//	event[i].remVal(k*ins.getNRooms() + j + 1);
						eventRoom[i].remVal(j + 1);
					} catch (ContradictionException e) {
						throw new Error("buildallocation: inconsistent initial model ??");
					}
				}
			}
		}

	}

	public void printColoring() {
		for (int i = 0; i < nDays; i++) {
			for (int j = 0; j < nHours; j++) {
				for (int k = 0; k < ins.getNEvents(); k++) {
					if ((eventTime[k].getVal()) == (i * nHours + j + 1))
						System.out.print(k + " ");
				}
				System.out.println();
			}
			System.out.println();
		}
		System.out.println();

		System.out.println("isFeas: " + pb.isFeasible());
		System.out.println("nbNode: " + pb.getSolver().getSearchSolver().getNodeCount());
		System.out.println("time: " + pb.getSolver().getSearchSolver().getTimeCount());
	}

	public void printFullTimetable() {
		System.out.println("feasible: " + pb.isFeasible());
		//System.out.println(pb.pretty());
		if (pb.isFeasible()) {
			int timetable[] = new int[nHours * nDays * ins.getNRooms()];
			for (int i = 0; i < (nHours * nDays * ins.getNRooms()); i++)
				timetable[i] = -1;

			for (int i = 0; i < ins.getNEvents(); i++) {
				timetable[((eventTime[i].getVal() - 1) % nHours) * (nDays * ins.getNRooms()) +
						((eventTime[i].getVal() - 1) / nHours) * ins.getNRooms() +
						eventRoom[i].getVal() - 1] = i;
			}

			int k;
			for (int i = 0; i < nDays; i++) {
				for (int j = 0; j < nHours; j++) {
					for (int r = 0; r < ins.getNRooms(); r++) {
						k = timetable[j * (nDays * ins.getNRooms()) + i * ins.getNRooms() + r];
						if (k < 0)
							System.out.print("[ ]");
						else
							System.out.print("[" + k + "]");
					}
					System.out.println();
				}
				System.out.println();
			}
		}
		System.out.println("nbNode: " + pb.getSolver().getSearchSolver().getNodeCount());
		System.out.println("time: " + pb.getSolver().getSearchSolver().getTimeCount());
	}

	/**
	 * call the c++ checker and output the report in the console
	 */
	public void checkSolution() {

		SearchLoop.searchLoopTrace = true;

		try {
			String noutput = name.substring(0, name.length() - 4);
			BufferedWriter bw = new BufferedWriter(new FileWriter(new File(noutput + ".sln")));
			for (int i = 0; i < ins.getNEvents(); i++) {

				//bw.write((eventTime[i].getVal() - 1) + " " + (eventRoom[i].getVal() - 1));

				//assert( (eventTime[i].getVal() - 1) == ((event[i].getVal()-1)/ins.getNRooms()) );
				//assert( (eventRoom[i].getVal() - 1) == ((event[i].getVal()-1)%ins.getNRooms()) );

				//System.out.println( (event[i].getVal() - 1) + "  " + (eventTime[i].getVal() - 1) + "  " + (eventRoom[i].getVal() - 1) );
				bw.write( (eventTime[i].getVal() - 1) + " " + ((event[i].getVal()-1)%ins.getNRooms()));

				bw.newLine();
			}
			bw.flush();
			bw.close();
			AbstractSolver.runCheckerFromJava(noutput, -1);
		} catch (IOException e) {
			e.printStackTrace();
		}

	}


	public int getNDays() {
		return nDays;
	}

	public void setNDays(int days) {
		nDays = days;
	}

	public int getNHours() {
		return nHours;
	}

	public void setNHours(int hours) {
		nHours = hours;
	}

	public int getTime() {
		return (int) System.currentTimeMillis() - startTime;
	}

	public int getNode() {
		//System.out.println("" + pb.getSolver().getSearchSolver().getNodeCount());
		if (pb.getSolver().getSearchSolver() != null) {
			int n = pb.getSolver().getSearchSolver().getNodeCount();
			if (!restart) return n;
			else return restartpol.getNbNodeTot() + n;
		} else return 0;
	}

//******************************************************** //
//******************* Main  ******************************* //
//******************************************************** //


	public static void main(String[] args) {
		//MediumModel pm = new MediumModel("data/test/comp2-1-5-1.txt");
		//HeavyModel pm = new HeavyModel("data/Track2/comp-2007-2-3.tim");
		//LightModel pm = new LightModel("data/test/comp2-1-5-1.txt");

		//MediumModel pm = new MediumModel("data/test/comp2-1-5-1.txt");
		HeavyModel pm = new HeavyModel("data/small/small1.tim");//,LightModel.Impact);
		//LightModel pm = new LightModel("data/Track2/comp-2007-2-4.tim", 2);

		//LightModel pm = new LightModel("data/test/comp-2007-2-2-5-1.tim", 2);

		pm.nDays = 5;
		pm.nHours = 9;
		pm.heuristictype = PrimalModel.Impact;

		pm.buildModel(); //PrimalModel.CONFLICTS);
		pm.solve();
		System.out.println(""+ pm.getNode());
		//pm.printColoring();
		//pm.printFullTimetable();
		//pm.checkSolution();
	}

//******************************************************** //
//******************* Debug tools ************************ //
//******************************************************** //

	public void solveDebug() {
		// Solving
		System.out.print("start solve");
		Solver s = pb.getSolver();
		s.setVerbosity(Solver.SEARCH);
		s.setLoggingMaxDepth(100);
		s.generateSearchSolver(pb);
		//s.attachGoal(new AssignVar(new StaticVarOrder(event), new IncreasingDomain()));
		//s.attachGoal(new AssignVar(new MinDomain(pb,event), new IncreasingDomain()));
		s.attachGoal(new AssignVar(new DomOverDeg(pb, eventTime), new IncreasingDomain()));
		s.addGoal(new AssignVar(new MinDomain(pb, eventRoom), new IncreasingDomain()));
//        s.attachGoal(new AssignVar(new MinDomPrint(pb,eventTime,
//        		nHours,nDays,
//        		ins.getNEvents(),
//        		ins.getNRooms(), 1), new IncreasingDomain()));
//        s.addGoal(new AssignVar(new MinDomPrint(pb,eventRoom,
//        		nHours,nDays,
//        		ins.getNEvents(),
//        		ins.getNRooms(), 2), new IncreasingDomain()));
		s.launch();
		//s.flushLogs();
		// Display
	}


}
