package cpmodel;

import choco.AbstractProblem;
import choco.Solver;
import choco.search.NodeLimit;
import choco.search.TimeLimit;

import java.util.logging.Logger;
import java.util.logging.Level;

// **************************************************
// *                   J-CHOCO                      *
// *   Copyright (C) F. Laburthe, 1999-2003         *
// **************************************************
// *  an open-source Constraint Programming Kernel  *
// *     for Research and Education                 *
// **************************************************

public class RestartPolicy {

    protected AbstractProblem pb;
    protected int currentlimit;
    protected double grow = 1.4;
    protected long maxTime;

    /**
     * Given a problem pb and a maximum time as a default limit
     * The restartPolicy consist in executing the solve() method as long
     * as an other limit in nodes is reached (this limit increase from one execution to
     * another and the time of ALL executions combined is limited by maxTime)
     */
    public RestartPolicy(AbstractProblem pb, long maxTime) {
        this.pb = pb;
        this.maxTime = maxTime;
        currentlimit = pb.getNbIntVars();
    }

    public RestartPolicy(AbstractProblem pb, long maxTime, int climit) {
        this.pb = pb;
        this.maxTime = maxTime;
        currentlimit = climit;
    }

    public RestartPolicy(AbstractProblem pb, long maxTime, int climit, double grow) {
        this.pb = pb;
        this.maxTime = maxTime;
        this.currentlimit = climit;
        this.grow = grow;
    }

    protected int nbNodeTot;

    protected int nbNodeLastIter;

    protected int tottime;

    protected int nbRestart;

    public int getNbNodeTot() {
        return nbNodeTot;
    }

    public int getNbNodeLastIter() {
        return nbNodeLastIter;
    }

    public int getTottime() {
        return tottime;
    }

    public int getNbRestart() {
        return nbRestart;
    }

    public void reinitSolver() {
        pb.getSolver().getSearchSolver().nextMove = -1;
    }

    public void setNodeLimitOnTheFly(int clim) {
        NodeLimit nodelimit = (NodeLimit) pb.getSolver().getSearchSolver().limits.get(1);
        nodelimit.reset(true);
        nodelimit.setNbMax(clim);
    }

    public void setTimeLimitOnTheFly(long clim) {
        TimeLimit timelimit = (TimeLimit) pb.getSolver().getSearchSolver().limits.get(0);
        timelimit.reset(true);
        timelimit.setNbMax((int) (clim));
    }

    public void setTime(long clim) {
        TimeLimit timelimit = (TimeLimit) pb.getSolver().getSearchSolver().limits.get(0);
        timelimit.reset(true);
        timelimit.setNbMax((int) (clim));
    }

    public Boolean restartSolve() {
        Logger.getLogger("choco").setLevel(Level.OFF);
        Logger.getLogger("choco.search").setLevel(Level.OFF);
        nbNodeTot = 0;
        nbNodeLastIter = 0;
        tottime = 0;
        nbRestart = 0;
        long startTime = System.currentTimeMillis();
        long currentTime = 0;
        boolean limitReached = true;
        boolean first = true;

        while (limitReached && currentTime < maxTime) {
            //System.out.println("start a resolution with limit " + currentlimit + "[" + currentTime + "]");
            if (first) {
                pb.getSolver().setNodeLimit(currentlimit);
                pb.getSolver().setTimeLimit((int) (maxTime - currentTime));
                Solver s = pb.getSolver();
                s.generateSearchSolver(pb);
                s.launch();
                first = false;
            } else {
                nbRestart++;
                setNodeLimitOnTheFly(currentlimit);
                setTimeLimitOnTheFly(maxTime - currentTime);
                reinitSolver();
                pb.nextSolution();
            }

            nbNodeLastIter = pb.getSolver().getSearchSolver().getNodeCount();
            nbNodeTot += nbNodeLastIter;
            limitReached = pb.getSolver().getSearchSolver().isEncounteredLimit();
            if (limitReached)
                pb.worldPopUntil(pb.getSolver().getSearchSolver().baseWorld);
            currentTime = System.currentTimeMillis() - startTime;
            currentlimit = (int) (grow * currentlimit);
        }

        tottime = (int) (System.currentTimeMillis() - startTime);
        if (limitReached) return null;
        else {
	        setNodeLimitOnTheFly(Integer.MAX_VALUE);	        
            return (pb.getSolver().getNbSolutions() > 0);
        }
    }



}
