package cpmodel;

import cpmodel.constraints.ConsecutiveConstraint;
import cpmodel.constraints.Cost1Constraint;
import cpmodel.constraints.Cost3Constraint;
import cpmodel.heuristics.OptimizationValueSelector;
import choco.ContradictionException;
import choco.Solver;
import choco.Propagator;
import choco.integer.IntDomainVar;
import instance.TimetablingInstance;
import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.complete.SoftSolver;
import magicsearch.impact.ImpactStrategie;
import magicsearch.impact.Iref;
import magicsearch.impact.impactMAC.IBSProblem;
import magicsearch.impact.impactMAC.ImpactBasedVarSearch;
import magicsearch.impact.impactMAC.ImpactBasedValSearch;
import magicsearch.impact.impactMAC.IBSSolver;

public class SoftLightModel extends LightModel {

	protected IntDomainVar[] cost;
	protected IntDomainVar objective;

	protected int[] storedTime;
	protected int[] storedRoom;

	protected int storedCost1;
	protected int storedCost2;
	protected int storedCost3;

	protected boolean unsat = false;
	protected boolean timeout = false;
	protected boolean stopAtFirstSol = false;

	protected ConsecutiveConstraint conscons;

	public SoftLightModel(String s, int m) {
		super(s, m);
		initBuild();
	}


	public SoftLightModel(TimetablingInstance ins, AbstractSolver s, int m) {
		super(ins, s, m);
		initBuild();
	}

	public void initBuild() {
		storedTime = new int[ins.getNEvents()];
		storedRoom = new int[ins.getNEvents()];
	}


	public void setAndInitializeImpact() {
		setAndInitializeImpact(45);
	}

	public void setAndInitializeImpact(int discrepency) {
		if (((IBSProblem) pb).getIbs() == null) {
			Solver s = pb.getSolver();
			IntDomainVar[] ivars = new IntDomainVar[eventTime.length + 1];
			System.arraycopy(eventTime, 0, ivars, 0, eventTime.length);
			ivars[eventTime.length] = objective;

			ImpactStrategie ibs = new Iref(pb, ivars);
			((IBSProblem) pb).setIbs(ibs);
			if (discrepency < 45)
				((IBSSolver) s).setLDS(discrepency);
			s.setVarIntSelector(new ImpactBasedVarSearch(eventTime, ibs));
			if (SoftSolver.optHeuristic)
				s.setValIntSelector(new OptimizationValueSelector(ins, eventTime, conscons, ibs));
			else s.setValIntSelector(new ImpactBasedValSearch(ibs));

			if (initialization) {
				int level = pb.getWorldIndex();
				long tps = System.currentTimeMillis();
				((IBSProblem) pb).initImpact(timelimit);
				inittime = (int) (System.currentTimeMillis() - tps);
				pb.worldPopUntil(level);
			}
		} else {
			if (discrepency < 45)
				((IBSSolver) pb.getSolver()).setLDS(discrepency);
		}
	}

	public void buildModel(int coltype) {
		super.buildModel(coltype);


		cost = new IntDomainVar[3];
		cost[0] = pb.makeBoundIntVar("Cost_0", 0, 5 * ins.getNStudents());
		cost[1] = pb.makeBoundIntVar("Cost_1", 0, (35 * ins.getNStudents()));
		cost[2] = pb.makeBoundIntVar("Cost_2", 0, (5 * ins.getNStudents()));

		// objective function
		objective = pb.makeBoundIntVar("Obj", 0,  (45 * ins.getNStudents()));
		pb.post(pb.eq(objective, pb.sum(cost)));

		// cost 1 (last timeslots)
		pb.post(new Cost1Constraint(eventTime, cost[0], ins));

		// cost 2 (consecutive courses)
		conscons = new ConsecutiveConstraint(eventTime, cost[1], ins);
		pb.post(conscons);

		// cost 3 (almost empty days)
		pb.post(new Cost3Constraint(eventTime, cost[2], ins));
	}


	protected void saveImprovingSolution() {
		for (int i = 0; i < ins.getNEvents(); ++i) {
			// store eventTime
			storedTime[i] = eventTime[i].getVal() - 1;

			// store rooms
			storedRoom[i] = solution[i];

			// store cost 1;
			storedCost1 = cost[0].getInf();

			// store cost 2;
			storedCost2 = cost[1].getInf();

			// store cost 3;
			storedCost3 = cost[2].getInf();
		}
		isFeasible = true;
	}

	private void restoreSavedSolution() throws ContradictionException {
		pb.worldPopUntil(0);
		for (int i = 0; i < ins.getNEvents(); ++i) {
			eventTime[i].setVal(storedTime[i] + 1);
			solution[i] = storedRoom[i];
		}
	}

	public static boolean printCPResolution = true;

	public void solve() {
		softCost = Integer.MAX_VALUE;		
		solve(45);
		restoreSol();
	}

	/**
	 * solve the problem by increasing discrepencies
	 */
	public void solveLDS() {
		int d = 0;
		softCost = Integer.MAX_VALUE;		
		while (((int) System.currentTimeMillis() - startTime) < timelimit) {
			System.out.println("DEPENDENCY: " + d);
			pb.feasible = null;
			solve(d);
			d++;
		}
		restoreSol();
	}

	public void solveLDS(int uniqued) {
		softCost = Integer.MAX_VALUE;
		solve(uniqued);
		restoreSol();
	}


	public void solve(int discrepency) {
		//Solver.setVerbosity(Solver.SEARCH);
		//pb.getSolver().setLoggingMaxDepth(100);
		setParameters(discrepency);
		mainSoftLoop();
	}

	public void restoreSol() {
		if (printCPResolution && unsat)
			System.out.println("Optimal solution!");

		if (isFeasible) {
			try {
				restoreSavedSolution();
			} catch (ContradictionException e) {
				throw new Error("solution stored not AC ?");
			}
		}
		if (printCPResolution)
			System.out.println("Final costs: " + storedCost2 + ", " + storedCost3 + ", " + storedCost1);
	}


	public void mainSoftLoop() {
		timeout = false;
		unsat = false;
		if (restart) {
			restartpol = new RestartPolicy(pb, timelimit - ((int) System.currentTimeMillis() - startTime));
			Boolean b = restartpol.restartSolve();
			timeout = (b == null);
			unsat = (b == Boolean.FALSE);
		} else {
			pb.solve();
			timeout = pb.getSolver().isEncounteredLimit();
			unsat = (pb.isFeasible() == Boolean.FALSE);
		}
		Boolean returnValue;
		Propagator cut = null;
	
		while (!unsat && !timeout && softCost > 0) {

			while (!unsat && !timeout && !allocateRooms()) {
				++nIterations;

				returnValue = pb.nextSolution();
				timeout = (returnValue == null);
				unsat = (returnValue == Boolean.FALSE);

				//if (printCPResolution && !timeout)
				//	System.out.println("\tfound a coloring " + nNogoods);
			}

			if (!unsat && !timeout) {

				saveImprovingSolution();
				if (printCPResolution)
					System.out.print("found an improving solution: "
							+ softCost + " -> ");

				softCost = computeCurrentCost();

				if (printCPResolution)
					System.out.println(objective.getInf() + " - " + this.getTime() + "ms / " + this.getNode() + " nodes " + nNogoods + " nogoods");

				if (cut != null)
					cut.delete();

				cut = (Propagator) pb.lt(objective, softCost);
				pb.postCut(cut);
				
				if (stopAtFirstSol) break;
				
				//if (!restart)
				returnValue = pb.nextSolution();
				//else returnValue = restartpol.restartSolve();
				timeout = (returnValue == null);
				unsat = (returnValue == Boolean.FALSE);
			}
		}
		if (cut != null) cut.delete();
	}

	protected boolean isFinal(int timeslot) {
		return (timeslot % 9) == 0;
	}

	protected int computeCurrentCost() {
		return (storedCost1 + storedCost2 + storedCost3);
	}

	public static void main(String[] args) {
		String benchi = "data/small2008/small1.tim";
		SoftLightModel model = new SoftLightModel(benchi, Impact);

		System.out.println("solve " + benchi + " with " + model);
		SoftSolver.optHeuristic = true;
		printCPResolution = true;
		model.buildModel(PrimalModel.R_ALLDIFF);
		//model.buildModel(PrimalModel.A_ALLDIFF);
		model.setTimeLimit(4200000);
		model.solve();

		int time = model.getTime();
		int node = model.getNode();

		System.out.println(node + " NDS, " + time + " ms, "
				+ (model.nIterations + 1) + " iterations, "
				+ model.nNogoods + " nogoods learned\n");
		if (model.isFeasible) {
			//model.printColoring();
			model.checkSolution();
			//Benchmark.testChocoSol(true,bench[i],model.eventTime);
			//apply matching

			System.out.print("SAT     ");
		} else System.out.print("UNKnown ");
	}

}
