package cpmodel.constraints;

import java.util.BitSet;
import java.util.Arrays;
import java.util.Random;
import java.util.Iterator;

import instance.TimetablingInstance;
import choco.ContradictionException;
import choco.Var;
import choco.Problem;
import choco.Constraint;
import choco.util.IntIterator;
import choco.bool.AbstractCompositeConstraint;
import choco.integer.IntDomainVar;
import choco.integer.search.RandomIntVarSelector;
import choco.integer.search.RandomIntValSelector;
import choco.integer.constraints.AbstractIntConstraint;
import choco.integer.constraints.AbstractLargeIntConstraint;
import choco.mem.IStateInt;
import choco.mem.trailing.EnvironmentTrailing;
import choco.mem.trailing.StoredInt;
import util.IntSet;


/**
 * The constraint used to maintain
 * the second cost
 * (number of triplets of consecutive events).
 * <p/>
 * Semantic: X0,X1,X3 take consecutive values <-> X4 = 1
 */
public class ConsecutiveConstraint extends AbstractLargeIntConstraint {


	/**
	 * The variables corresponding to the event timeslots,
	 */
	protected IntDomainVar[] events;

	/**
	 * The variable corresponding to the cost.
	 */
	protected IntDomainVar cost;

	/**
	 * Number of events
	 */
	protected static int N;

	/**
	 * List of pairs < events, cost > for each ordered pair of events;
	 */
	public static int[][] costList;
	public static int[] tripletCost;


	protected TimetablingInstance instance;

	/**
	 * Structs used for computing the lower bound
	 */

	protected int[][] reversible_groundEvents;
	protected StoredInt[] reversible_card;
	protected StoredInt reversible_lower_bound;

	protected int common_lower_bound;
	protected int[] common_card;

	
///////////////////////////////////
	/**
	 * Structs used for computing better bound, and pruning.
	 */
	protected int[][] penalties;
	protected StoredInt[] minPenalty;
	protected StoredInt[] maxPenalty;
///////////////////////////////////
	
	

	public ConsecutiveConstraint(IntDomainVar[] events, IntDomainVar cost, TimetablingInstance instance) {

		super(toSingleArray(cost, events));
		this.events = events;
		this.cost = cost;
		this.instance = instance;
		this.N = events.length;
		reversible_groundEvents = new int[45][instance.getNRooms()];

		reversible_lower_bound = new StoredInt((EnvironmentTrailing) this.getProblem().getEnvironment(), 0);
		reversible_card = new StoredInt[45];

		for (int i = 0; i < 45; ++i)
			reversible_card[i] = new StoredInt((EnvironmentTrailing) this.getProblem().getEnvironment(), 0);

		common_card = new int[45];
		
///////////////////////////////////
		penalties = new int[N][45];
		minPenalty = new StoredInt[N];
		maxPenalty = new StoredInt[N];
		for (int i = 0; i < N; ++i) {
			minPenalty[i] = new StoredInt((EnvironmentTrailing) this.getProblem().getEnvironment(), 0);
			maxPenalty[i] = new StoredInt((EnvironmentTrailing) this.getProblem().getEnvironment(), 0);
		}
///////////////////////////////////

		initCostList();
	}

	private static IntDomainVar[] toSingleArray(IntDomainVar var, IntDomainVar[] vars) {
		IntDomainVar[] array = new IntDomainVar[vars.length + 1];
		System.arraycopy(vars, 0, array, 0, vars.length);
		array[vars.length] = var;
		return array;
	}

	private void initCostList() {

		BitSet student_i = new BitSet(1000);
		BitSet student_ij = new BitSet(1000);
		BitSet student_ijk = new BitSet(1000);

		tripletCost = new int[N * (N - 1) * (N - 2) / 6];
		costList = new int[N * (N - 1) / 2][];
		int[] tmp_costList = new int[2 * N];
		int l = 0, ntc;

		for (int i = 0; i < N - 1; ++i) {
			student_i.clear();
			student_i.or(instance.studentsAttendingEvent(i));

			for (int j = i + 1; j < N; ++j) {
				student_ij.clear();
				student_ij.or(instance.studentsAttendingEvent(j));
				student_ij.and(student_i);

				if (!student_ij.isEmpty()) {

					ntc = 0;
					for (int k = 0; k < N; ++k)
						if (k != i && k != j) {
							student_ijk.clear();
							student_ijk.or(instance.studentsAttendingEvent(k));
							student_ijk.and(student_ij);

							if (!student_ijk.isEmpty()) {
								tmp_costList[ntc++] = student_ijk.cardinality();
								tmp_costList[ntc++] = k;
								if (k > j)
									tripletCost[orderedIndex(i, j, k)] = student_ijk.cardinality();//tmp_costList[ntc-1];
							}
						}
					if (ntc > 0) {
						costList[l] = new int[ntc];
						for (int k = 0; k < ntc; ++k) {
							costList[l][k] = tmp_costList[k];
						}
						++l;
					} else {
						costList[l++] = null;
					}
				} else {
					costList[l++] = null;
				}
			}
		}
	}

	public static int get2Index(final int x, final int y) {
		int res;
		if (x < y)
			res = ((x * (N - 1)) - (x * (x + 1) / 2) + y - 1);
		else
			res = ((y * (N - 1)) - (y * (y + 1) / 2) + x - 1);
		return res;
	}


	public static int get3Index(final int x, final int y, final int z) {
		int a, b, c;
		if (x < y) {
			if (z < x) {
				a = z;
				b = x;
				c = y;
			} else {
				a = x;
				if (y < z) {
					b = y;
					c = z;
				} else {
					b = z;
					c = y;
				}
			}
		} else {
			if (z < y) {
				a = z;
				b = y;
				c = x;
			} else {
				a = y;
				if (x < z) {
					b = x;
					c = z;
				} else {
					b = z;
					c = x;
				}
			}
		}


		return orderedIndex(a, b, c);
	}

	public static int orderedIndex(int x, int y, int z) {
		y -= (x + 1);
		z -= (x + 1);
		return ((x * N * N) - ((x * (x + 2)) * N) + (x * (x + 1) * (x + 2) / 3)) / 2
				+ ((y * (N - 2 - x)) - (y * (y + 1) / 2) + z - 1);
	}
	
	public int updateEvent( int evt )
	{
		//assert( !events[evt].isInstantiated() );
		
		int oldmincost = minPenalty[evt].get();
		int newmincost = getMinPenalty(evt);
	
		if( newmincost > oldmincost ) 
		{
			reversible_lower_bound.add( newmincost-oldmincost );
			minPenalty[evt].set( newmincost );
		}
		return (newmincost - oldmincost);
	}

	private void updateInstantiatedEvents(final int evt, final int tst) {
		reversible_groundEvents[tst][reversible_card[tst].get()] = evt;
		reversible_card[tst].add(1);
	}

	public int getPenalty(final int evt, final int tst) {
		
		int c=0, n = Math.min(tst + 2, 44);
		for (int i = Math.max(tst - 2, 0); i <= n; ++i) {
			common_card[i] = reversible_card[i].get();
			++c;
		}

		int extracost = 0;
		if( c > 1 ) {
			if (tst > 1 && sameDay(tst - 2, tst - 1, tst))
				extracost += getPenalty(evt, tst - 2, tst - 1);
			if (tst > 0 && tst < 44 && sameDay(tst - 1, tst, tst + 1))
				extracost += getPenalty(evt, tst - 1, tst + 1);
			if (tst < 43 && sameDay(tst, tst + 1, tst + 2))
				extracost += getPenalty(evt, tst + 1, tst + 2);
		}
		return extracost;
	}

	private int getPenalty(final int evt, final int t1, final int t2) {

		int x, y, extracost = 0;
		if (common_card[t1] > 0 && common_card[t2] > 0)
			for (x = 0; x < common_card[t1]; ++x) {
				if (costList[get2Index(evt, reversible_groundEvents[t1][x])] != null)
					for (y = 0; y < reversible_card[t2].get(); ++y)
						extracost += tripletCost[get3Index(reversible_groundEvents[t1][x],
								reversible_groundEvents[t2][y], evt)];
			}
		return extracost;
	}

//	public void pruneEvents(final int new_evt, final int ground_t,
//	                        final int prune_t, final int lb) throws ContradictionException {
//
//		int i, j, k, ub = cost.getSup();
//		for (i = 0; i < common_card[ground_t]; ++i) {
//			k = get2Index(new_evt, reversible_groundEvents[ground_t][i]);
//			if (costList[k] != null)
//				for (j = 0; j < costList[k].length; j += 2)
//					if (costList[k][j] + lb > ub && !events[costList[k][j + 1]].isInstantiated())
//						events[costList[k][j + 1]].remVal(prune_t);
//		}
//	}
	
	public void filterFromLowerBound() throws ContradictionException
	{
		int evt, vali, pen;
		for(evt=0; evt<N; ++evt)
		{		
			if( !events[evt].isInstantiated() 
						&& (common_lower_bound - minPenalty[evt].get() + maxPenalty[evt].get() > cost.getSup() ) )
			{				
				common_lower_bound -= minPenalty[evt].get();
				
				vali = events[evt].getInf();
				do {
					pen = getPenalty(evt, vali-1);
					
					//assert( pen <= maxPenalty[evt].get() );
					
					if( common_lower_bound + pen  > cost.getSup() ) {
						//System.out.println( "prune " + vali + " from " + events[evt].pretty() );
						events[evt].remVal( vali );
					}
					vali = events[evt].getNextDomainValue(vali);
				} while( vali <= events[evt].getSup() );
				
				common_lower_bound += minPenalty[evt].get();
			}
		}
	}
	
	
	public void checkPenalties()
	{
		int evt, vali, pen, realmincost=Integer.MAX_VALUE, realmaxcost=0;
		//int tmincost = realmincost;
		//int tmaxcost = realmaxcost;
		for(evt=0; evt<N; ++evt)
		{		
			if( !events[evt].isInstantiated() )
			{		
				realmincost=Integer.MAX_VALUE;
				realmaxcost=0;
				vali = events[evt].getInf();
				do {
					pen = getPenalty(evt, vali-1);
	
					if( pen < realmincost ) 
						{
							realmincost = pen;
							//if( realmincost < tmincost ) tmincost = realmincost;
						}
					if( pen > realmaxcost ) 
						{
							realmaxcost = pen;
							//if( realmaxcost > tmaxcost ) tmaxcost = realmaxcost;
						}
	
					vali = events[evt].getNextDomainValue(vali);
				} while( vali <= events[evt].getSup() );

				if( realmincost != minPenalty[evt].get() || realmaxcost != maxPenalty[evt].get() )
				{
					System.out.println( "evt" + evt + " should be [" + realmincost + ".." + realmaxcost + "], is [" 
							+ minPenalty[evt].get() + ".." + maxPenalty[evt].get() + "]" );
				}
				//assert( realmincost == minPenalty[evt].get() && realmaxcost == maxPenalty[evt].get() );
				
			}
		}
		
		
//		if( tmaxcost != maxPotPenalty.get() )
//		{
//			System.out.println( "total should be [" + tmaxcost + "], is [" 
//					 + maxPotPenalty.get() + "]" );
//		}
		
	}
	
	
	public boolean updateLowerBound( int tst ) throws ContradictionException
	{
		int evt, increase=0;
		for(evt=0; evt<N; ++evt)
			if( !events[evt].isInstantiated() && relatedTo(evt, tst) )
				increase+=updateEvent(evt);
		if( increase > 0 )
			cost.updateInf(cost.getInf()+increase, cIndices[N]);
		return increase > 0;
	}
	
	public int getMinPenalty(int evt)
	{
		int newmaxcost=0, vali=events[evt].getInf(), pen, newmincost=Integer.MAX_VALUE;
		do {
			pen = getPenalty(evt, vali-1);
			if( pen < newmincost ) newmincost = pen;
			if( pen > newmaxcost ) newmaxcost = pen;
			vali = events[evt].getNextDomainValue(vali);
		} while( vali <= events[evt].getSup() );
		
		maxPenalty[evt].set( newmaxcost );
		
		return newmincost;
	}

	public boolean relatedTo(int evt, int tst)
	{
		boolean related = false;
		int minv = Math.max(tst-2, (tst/9)*9);
		int maxv = Math.min(tst+2, ((tst/9)+1)*9);
		//System.out.println( minv + ".." + tst + ".." + maxv );
		while( !related && minv <= maxv ) {
			related = events[evt].canBeInstantiatedTo(minv);
			++minv;	
		}
		return related;
	}
	
	@Override
	public void awakeOnInf(int evt) throws ContradictionException {
		if( evt < N ) {
			int increase = updateEvent( evt );
			if( increase > 0 )
				cost.updateInf(cost.getInf()+increase, cIndices[N]);
		}
	}
	
	@Override
	public void awakeOnRemovals(int evt, IntIterator values) throws ContradictionException {
		
		//System.out.println( "update evt" + evt );
		
		int increase = updateEvent( evt );
		
		//checkPenalties();
		
		if( increase > 0 )
			cost.updateInf(cost.getInf()+increase, cIndices[N]);
	}

	@Override
	public void awakeOnRem(int evt, int arg1) throws ContradictionException {
	}

	@Override
	public void awakeOnSup(int evt) throws ContradictionException {
		if( evt < N ) {
			int increase = updateEvent( evt );
			if( increase > 0 )
				cost.updateInf(cost.getInf()+increase, cIndices[N]);
		}
	}

	public boolean isSatisfied() {
		throw new Error("isSatified should not be called");
	}
		
	public void awakeOnInst(int idx) throws ContradictionException {
		if (idx < N) {

			int timeslot = events[idx].getVal() - 1;

			updateInstantiatedEvents(idx, timeslot);

			int penalty = getPenalty(idx, timeslot);
			int increase = ( penalty - minPenalty[idx].get());
			
			//System.out.println( idx + ": " + minPenalty[idx].get() + " -> " + penalty );
			
			if( increase > 0 ) {
				reversible_lower_bound.add( increase );
				minPenalty[idx].set( penalty );				
				common_lower_bound = reversible_lower_bound.get();// + addFromScratch();
				cost.updateInf(common_lower_bound, cIndices[N]);
			}
			
			if( updateLowerBound( timeslot ) || increase > 0 )
				filterFromLowerBound();

		}

	}

	public void propagate() throws ContradictionException {
	}

	private boolean sameDay(final int t1, final int t2, final int t3) {
		return (t1 % 9 < t2 % 9) && (t2 % 9 < t3 % 9);
	}

	public int addFromScratch() {
		int totadd = 0;
		for (int i = 0; i < events.length; i++) {
			IntDomainVar event = events[i];
			if (!event.isInstantiated()) {
				IntIterator it = event.getDomain().getIterator();
				int mini = Integer.MAX_VALUE;
				while (mini != 0 && it.hasNext()) {
					int val = it.next();
					int penalty = getPenalty(i, val - 1);
					if (mini > penalty) {
						mini = penalty;
					}
				}
				totadd += mini;
			}
		}
		//if (totadd != 0)
		//	System.out.println("" + totadd);
		return totadd;
	}


}
