/**
 *
 */
package cpmodel.constraints;

import instance.TimetablingInstance;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.*;

import tools.InstanceReader;
import choco.ContradictionException;
import choco.Problem;
import choco.util.IntIterator;
import choco.integer.IntDomainVar;
import choco.integer.constraints.AbstractLargeIntConstraint;
import choco.integer.search.RandomIntValSelector;
import choco.integer.search.RandomIntVarSelector;
import choco.mem.IStateBitSet;
import choco.mem.IStateInt;
import localsearch.graphalgos.Hongrois;


/**
 * The constraint that maintains the first cost
 * (number of students in a final timeslot).
 *
 * @author alexp
 */
public class Cost1Constraint extends AbstractLargeIntConstraint {

	/**
	 * The variables corresponding to the event timeslots,
	 */
	protected IntDomainVar[] events;

	/**
	 * The size (the number of students in it) of each event.
	 */
	protected int[] sizes;

	/**
	 * The variable corresponding to the cost.
	 */
	protected IntDomainVar cost;

	protected TimetablingInstance instance;

	protected IStateInt lb/*, ub*/;

	/**
	 * nFinalTimeslots[i] is the number of final timeslots
	 * in the domain of events[i].
	 */
	protected IStateInt[] nFinalTimeslots;

	/**
	 * nNonFinalTimeslots[i] is the number of non-final
	 * timeslots in the domain of events[i].
	 */
	protected IStateInt[] nNonFinalTimeslots;

	/**
	 * domainCopy(45*i+j) indicates if the value (j+1) is in
	 * the domain of the event i.
	 */
	protected IStateBitSet domainCopy;

	/**
	 * Creates a new constraint that maintains the cost 1.
	 *
	 * @param events   the variables corresponding to the event timeslots,
	 *                 in the same order as in the instance.
	 * @param cost     the variable corresponding to the cost.
	 * @param instance the instance
	 */
	public Cost1Constraint(IntDomainVar[] events, IntDomainVar cost, TimetablingInstance instance) {
		super(toSingleArray(cost, events));
		this.events = events;
		this.cost = cost;
		this.instance = instance;
		this.problem = vars[0].getProblem();

		sizes = new int[events.length];
		for (int i = 0; i < events.length; i++)
			sizes[i] = instance.studentsAttendingEvent(i).cardinality();
		initTotalStudent();

		lb = problem.getEnvironment().makeInt();
//		ub = problem.getEnvironment().makeInt();

		nFinalTimeslots = new IStateInt[events.length];
		for (int i = 0; i < events.length; i++)
			nFinalTimeslots[i] = problem.getEnvironment().makeInt(0);

		nNonFinalTimeslots = new IStateInt[events.length];
		for (int i = 0; i < events.length; i++)
			nNonFinalTimeslots[i] = problem.getEnvironment().makeInt(0);

		domainCopy = problem.getEnvironment().makeBitSet(events.length * 45);

	}

	private static IntDomainVar[] toSingleArray(IntDomainVar var, IntDomainVar[] vars) {
		IntDomainVar[] array = new IntDomainVar[vars.length + 1];
		System.arraycopy(vars, 0, array, 1, vars.length);
		array[0] = var;
		return array;
	}

	/**
	 * Indicates if the given timeslot is in
	 * the domain copy of the event.
	 *
	 * @param event    an event
	 * @param timeslot a timeslot
	 * @return whether the timeslot is in the domain copy.
	 */
	protected boolean isInDomainCopy(int event, int timeslot) {
		return domainCopy.get(45 * event + timeslot - 1);
	}

	/**
	 * Adds the given timeslot in
	 * the domain copy of the event.
	 *
	 * @param event    an event
	 * @param timeslot a timeslot
	 */
	protected void addInDomainCopy(int event, int timeslot) {
		if (!domainCopy.get(45 * event + timeslot - 1)) {
			domainCopy.set(45 * event + timeslot - 1);
			if (isFinal(timeslot))
				nFinalTimeslots[event].add(1);
			else
				nNonFinalTimeslots[event].add(1);
		}
	}

	/**
	 * Removes the given timeslot from
	 * the domain copy of the event.
	 *
	 * @param event    an event
	 * @param timeslot a timeslot
	 */
	protected void removeFromDomainCopy(int event, int timeslot) {
		if (domainCopy.get(45 * event + timeslot - 1)) {
			domainCopy.clear(45 * event + timeslot - 1);
			if (isFinal(timeslot))
				nFinalTimeslots[event].add(-1);
			else
				nNonFinalTimeslots[event].add(-1);
		}
	}

//	/**
//	 * Returns the index of the day of the
//	 * given timeslot.
//	 */
//	protected int timeslotDayIndex(int timeslot){
//		return (timeslot - 1) / 9;
//	}

	/**
	 * Returns the index of the final timeslot
	 * of the specified day.
	 *
	 * @param day a day index (between 0 and 4);
	 * @return the last timeslot of that day
	 */
	protected int finalTimeslotOfDay(int day) {
		assert (day >= 0 && day < 5);
		return (day + 1) * 9;
	}

	/**
	 * Hard coded function (doesn't depend on TimetablingInstance.N_TIMESLOTS).
	 */
	protected boolean isFinal(int timeslot) {
		return timeslot % 9 == 0;
	}

	protected void initStructures() {
		for (int e = 0; e < events.length; e++) {
			IntDomainVar var = events[e];
			for (int val = var.getInf(); val <= var.getSup(); val = var.getNextDomainValue(val))
				addInDomainCopy(e, val);
		}
	}

	protected void initLb() {
		int lbs = 0;
		for (int e = 0; e < events.length; e++)
			if (mustGoToFinal(e)) lbs += sizes[e];
		lb.set(lbs);
	}

	protected boolean checkLbConsistency() {
		int lbs = 0;
		for (int e = 0; e < events.length; e++)
			if (mustGoInFinalFromScratch(e)) lbs += sizes[e];
		if (lb.get() != lbs) throw new Error(lb.get() + " " + lbs);
		return lb.get() == lbs;
	}

	public boolean mustGoInFinalFromScratch(int e) {
		IntIterator it = vars[e + 1].getDomain().getIterator();
		while (it.hasNext()) {
			int val = it.next();
			if (!isFinal(val)) return false;
		}
		return true;
	}

//	protected boolean checkUbConsistency(){
//		int ubs = 0;
//		for(int e=0;e<events.length;e++){
//			if(canGoToFinal(e))
//				ubs += sizes[e];
//		}
//		return ub.get() == ubs;
//	}

	protected boolean checkStructureConsistency() {
		for (int e = 0; e < events.length; e++) {
			IntDomainVar var = events[e];
			int nNFTInDomain = 0,
					nNFTInDomainCopy = 0,
					nFTInDomain = 0,
					nFTInDomainCopy = 0;
			for (int val = var.getInf(); val <= var.getSup(); val = var.getNextDomainValue(val)) {
				if (!isInDomainCopy(e, val))
					return false;//A value in the domain is not in the copy
				if (isFinal(val))
					++nFTInDomain;
				else
					++nNFTInDomain;
			}
			for (int t = 1; t <= 45; t++)
				if (isInDomainCopy(e, t))
					if (isFinal(t)) ++nFTInDomainCopy;
					else ++nNFTInDomainCopy;

			if (nFTInDomain != nFTInDomainCopy)
				return false;//A value in the domain copy is not in the domain
			if (nNFTInDomain != nNFTInDomainCopy)
				return false;//A value in the domain copy is not in the domain
			if (nFTInDomain != nFinalTimeslots[e].get())
				return false;//The count of final timeslots does not correspond to the domains (actual and copy).
			if (nNFTInDomain != nNonFinalTimeslots[e].get())
				return false;//The count of non-final timeslots does not correspond to the domains (actual and copy).
		}
		return true;
	}


	/**
	 * The domain of the event has one value
	 * corresponding to the final timeslots.
	 */
	protected boolean canGoToFinal(int event) {
		return nFinalTimeslots[event].get() > 0;
	}

	/**
	 * The domain of the event has one value
	 * corresponding to non-final timeslots.
	 */
	protected boolean canGoToNonFinal(int event) {
		return nNonFinalTimeslots[event].get() > 0;
	}

	/**
	 * The domain of the event has only values
	 * corresponding to the final timeslots.
	 */
	protected boolean mustGoToFinal(int event) {
		return canGoToFinal(event) && !canGoToNonFinal(event);
//		return !canGoToNonFinal(event);
	}

	/**
	 * The domain of the event has only values
	 * corresponding to non-final timeslots.
	 */
	protected boolean mustGoToNonFinal(int event) {
		return canGoToNonFinal(event) && !canGoToFinal(event);
	}

	/**
	 * Updates the internal structures according to
	 * the removal of the value from the domain of the event.
	 */
	protected void updateValueRemoval(int event, int value) {
		removeFromDomainCopy(event, value);
	}

	/**
	 * Updates the internal structures according to
	 * the instantiation of the event to the value.
	 */
	protected void updateEventInstanciation(int event, int value) {
		for (int t = 1; t <= 45; t++)
			if (t != value) removeFromDomainCopy(event, t);
	}

	/**
	 * Updates the internal structures according to
	 * the current state of the event.
	 */
	protected void updateForEvent(int event) {
		for (int t = 1; t <= 45; t++)
			if (isInDomainCopy(event, t)
					&& !events[event].canBeInstantiatedTo(t))
				removeFromDomainCopy(event, t);
	}

	/**
	 * Removes the values corresponding to the final
	 * timeslots in the domain of the specified event.
	 *
	 * @throws ContradictionException
	 */
	protected boolean removeFinal(int event) throws ContradictionException {
		if (canGoToFinal(event)) {
//			System.out.println("remove final "+(event+1));
			for (int i = 0; i < 5; i++) {
				int t = finalTimeslotOfDay(i);//Final timeslots
				if (events[event].canBeInstantiatedTo(t)) {
					events[event].removeVal(t, cIndices[event + 1]);
					updateValueRemoval(event, t);
				}
			}
			return true;
		}
		return false;
	}


	public boolean isSatisfied() {
		int s = 0;
		for (int i = 0; i < events.length; i++)
			if (isFinal(events[i].getVal()))
				s += sizes[i];
		return s == cost.getVal();
	}

	@Override
	public void awakeOnVar(int idx) throws ContradictionException {
//		System.out.println(problem.getWorldIndex()+" Cost1Constraint.awakeOnVar() "+vars[idx].pretty());
		if (idx > 0) {
			int event = idx - 1;
			boolean couldGoToNonFinal = canGoToNonFinal(event);
			updateForEvent(event);
			if (mustGoToFinal(event) && couldGoToNonFinal) {
				lb.add(sizes[event]);
				cost.updateInf(lb.get(), cIndices[0]);
				constAwake(false);
			}
		} else constAwake(false);
	}


	@Override
	public void awakeOnRem(int idx, int x) throws ContradictionException {
//		System.out.println(problem.getWorldIndex()+" Cost1Constraint.awakeOnRem() "+idx+" "+x+" "+vars[idx].pretty());
		if (idx != 0) {
			int event = idx - 1;
			boolean couldGoToNonFinal = canGoToNonFinal(event);
			updateValueRemoval(event, x);
			if (mustGoToFinal(event) && couldGoToNonFinal) {//x was the only value of a non-final timeslot
				lb.add(sizes[event]);
				cost.updateInf(lb.get(), cIndices[0]);
				constAwake(false);
			}
		} else {/*Nothing to do, only bound events on the cost*/}
	}

	@Override
	public void awakeOnInst(int idx) throws ContradictionException {
//		System.out.println(problem.getWorldIndex()+" Cost1Constraint.awakeOnInst() "+idx+" "+vars[idx].getVal());
		if (idx != 0) {
			int event = idx - 1;
			boolean couldGoToNonFinal = canGoToNonFinal(event);//Before instantiation
			updateEventInstanciation(event, events[event].getVal());
			if (mustGoToFinal(event) && couldGoToNonFinal) {
				lb.add(sizes[event]);
				cost.updateInf(lb.get(), cIndices[0]);
				constAwake(false);
			}
		} else constAwake(false);
//			propagate();
	}

	@Override
	public void awakeOnInf(int idx) throws ContradictionException {
//		System.out.println(problem.getWorldIndex()+" Cost1Constraint.awakeOnInf() "+vars[idx].pretty());
		awakeOnVar(idx);
	}

	@Override
	public void awakeOnSup(int idx) throws ContradictionException {
//		System.out.println(problem.getWorldIndex()+" Cost1Constraint.awakeOnSup() "+vars[idx].pretty());
		awakeOnVar(idx);
	}

	@Override
	public void propagate() throws ContradictionException {
		//System.out.println("PROPAGATE");
		int nlb = lb.get();// + computeDeltaLB();
		cost.updateInf(nlb, cIndices[0]);
		for (int i = 0; i < events.length; i++) {
			if (canGoToNonFinal(i) && lb.get() + sizes[i] > cost.getSup()) {
				removeFinal(i);
			}
		}
	}

	@Override
	public void awake() throws ContradictionException {
		initStructures();
		initLb();
		propagate();
	}

//********************************************************************//
//******************** Improved LB ***********************************//
//********************************************************************//

	public int totalNBstudent = 0;

	public void initTotalStudent() {
		int sum = 0;
		for (int i = 0; i < events.length; i++) {
			sum += sizes[i];
		}
		totalNBstudent = sum;
	}

	public int computeDeltaLB() {
		if (events[0].getProblem().getWorldIndex() > 50) {
			//System.out.println("start compute deltaLB");
			int size = instance.getNRooms() * 45;
			int[][] hcost = new int[size][size];
			for (int i = 0; i < size; i++) {
				for (int j = 0; j < size; j++) {
					if (i < instance.getNEvents()) {
						hcost[i][j] = 100000;
					} else hcost[i][j] = 0;
				}
			}

			//for all non final timeslots, compute the set of events that can still go in this timeslot
			for (int i = 1; i <= 45; i++) {
				LinkedList<Integer> levts = new LinkedList<Integer>();
				for (int j = 0; j < events.length; j++) {
					if (events[j].canBeInstantiatedTo(i) || events[j].isInstantiatedTo(i)) {
						levts.add(j);
					}
				}
				for (Iterator<Integer> it = levts.iterator(); it.hasNext();) {
					int evt = it.next();
					for (int j = 0; j < instance.getNRooms(); j++) {
						if (instance.eventFitsInRoom(evt, j)) {
							if (isFinal(i))
								hcost[evt][j + (i-1) * instance.getNRooms()] = 0;
							else hcost[evt][j + (i-1) * instance.getNRooms()] = -instance.getNbStudentInEvent(evt);
						}
					}
				}
			}
			Hongrois hung = new Hongrois(hcost);
			hung.solve();

			//System.out.println(" " + hung.getCost());
			int maxFitStudent = -hung.getCost();
			//if (maxFitStudent > remainings) break;


			int qdelta = totalNBstudent - maxFitStudent - lb.get();
			//int qdelta = Math.max(lbf, 0);
			//System.out.println(problem.getWorldIndex() + " " + lbf + " vs " + lb.get() + " with " + nbevents + " implying " + qtoRemove);
			//if (qdelta > 0) {
			//	System.out.println(problem.getWorldIndex() + " add delta " + qdelta);
			//}
			if(qdelta < 0 ){
				throw new Error((totalNBstudent - maxFitStudent) + " >=? " + lb.get());				
			}
			assert (qdelta >= 0);
			//System.out.println("end compute deltaLB " + qdelta);
			return qdelta;
		} else return 0;
	}

	/**
	 * @param args
	 * @throws ContradictionException
	 */
	public static void main(String[] args) throws ContradictionException {

		for (int seed = 0; seed < 1; seed++) {
			TimetablingInstance t = anInstance();

			Problem pb = new Problem();
			IntDomainVar[] events = pb.makeEnumIntVarArray("e", t.getNEvents(), 1, 45);

			/*events[0].setVal(9);
			for(int i=1;i<events.length;i++) events[i].setVal(1);*/

			IntDomainVar cost = pb.makeBoundIntVar("cost", 0, 5 * t.getNStudents());
			Cost1Constraint c = new Cost1Constraint(events, cost, t);
			pb.post(c);

			pb.getSolver().setVarSelector(new RandomIntVarSelector(pb, seed));
			pb.getSolver().setValSelector(new RandomIntValSelector(seed));
			/*c.awake();
			events[1].remVal(44);
			c.awakeOnRem(2, 44);*/

			pb.solve();
			if (pb.isFeasible())
				do {
					System.out.println(c.isSatisfied() + " " + cost + " " + Arrays.toString(events));
				} while (pb.nextSolution() == Boolean.TRUE);
			System.out.println("nbSol = " + pb.getSolver().getNbSolutions());
		}
	}

	private static TimetablingInstance anInstance() {

		TimetablingInstance t;
		try {
			t = new InstanceReader(new FileInputStream("data/Track2/comp-2007-2-1.tim")).getInstance();
		} catch (FileNotFoundException e) {
			return null;
		}
		return t;

//		int nEvent = 10, nStudent = 100;
//		Random r = new Random(600);
//		TimetablingInstance t = new TimetablingInstance(nEvent, 0, 0, nStudent);		
//		for(int s=0;s<nStudent;s++){
//			//take a random subset of events and put it to the student
//			IntSet events = new IntSet(10);for(int e=0;e<nEvent;e++)events.add(e);
//			events.setRandom(r);
//			int nEventS = r.nextInt(nEvent)+1;
//			for(int e=0; e<nEventS;e++){
//				int re = events.randomValue();
//				events.remove(re);
//				t.setStudentForEvent(s, re);
//			}
//			
//		}
//		
//		return t;
	}
}

/*
	public void checkMissingDeductions() {
		int lbf = 0;
		BitSet evtInLb = new BitSet();
		for (int e = 0; e < events.length; e++) {
			if (mustGoInFinalFromScratch(e)) {
				lbf += sizes[e];
				evtInLb.set(e);
			}
		}
		//System.out.println(lb.get() + " " + lbf);
		if (lbf > cost.getSup()) throw new Error("contradiction missed");
		for (int i = 0; i < events.length; i++) {
			if (canGoToNonFinalFromScratch(i) &&
				!evtInLb.get(i) &&
				lbf + sizes[i] > cost.getSup()) {
				if (canGoToFinal(i))
					throw new Error("missing deduction on event " + i + " " + canGoToNonFinal(i));
			}
		}
	}


	public boolean canGoToNonFinalFromScratch(int e) {
		IntIterator it = vars[e + 1].getDomain().getIterator();
		while (it.hasNext()) {
			int val = it.next();
			if (!isFinal(val)) return true;
		}
		return false;
	}

	public boolean canGoToFinalFromScratch(int e) {
		IntIterator it = vars[e + 1].getDomain().getIterator();
		while (it.hasNext()) {
			int val = it.next();
			if (isFinal(val)) return true;
		}
		return false;
	}
*/
