/**
 * 
 */
package instance;

import instance.colouring.DegeneracyOrder;
import instance.colouring.MinDomMaxDegSelector;
import instance.colouring.MinDomOverDegSelector;
import instance.colouring.MinNodeSelector;

import java.io.FileInputStream;
import java.io.PrintStream;
import java.util.Iterator;

import tools.InstanceReader;
import util.IntSet;

/**
 * Representation of graph whose vertices has
 * a list of possible colours. Additionally,
 * each colour has a maximal number of vertices
 * to which it can be assigned.
 * 
 * @author alexp
 *
 */
public class ColourableConflictGraph extends ConflictGraph {

	
	/**
	 * colours[i] contains the list
	 * of currently possible colours for the node i.
	 */
	protected IntSet[] colours;
	
	/**
	 * maxOccurrences[c] indicates the maximal number of
	 * nodes the colour c must be assigned to. 
	 */
	protected int[] maxOccurrences;
	
	protected IntSet unassignedNodes;
	
	/**
	 * Creates a new graph with n nodes.
	 * @param n the number of nodes
	 * @param k the maximum number of possible colours
	 * for each node.
	 * @param o the maximal occurrence of the colours
	 */
	public ColourableConflictGraph(int n, int k, int o) {
		super(n);
		colours = new IntSet[n];
		for(int i=0;i<n;i++){
			colours[i] = new IntSet(k);
			colours[i].setRandom(random);
		}
		
		maxOccurrences = new int[k];
		for(int c=0;c<k;c++)
			maxOccurrences[c] = o;
		
		unassignedNodes = new IntSet(n);
		unassignedNodes.setRandom(random);
		for(int i=0;i<n;i++) unassignedNodes.add(i);
	}
	
	/**
	 * Fill with all possible colours the
	 * specified node.
	 * @param n
	 */
	public void fillColours(int n){
		assert(checkNodeIndex(n));
		for(int i=0;i<colours[n].capacity();i++)colours[n].add(i);
	}
	
	/**
	 * Fill with all possible colours the
	 * nodes.
	 */
	public void fillAllNodesColours(){
		for(int i=0;i<getNNodes();i++)
			fillColours(i);
	}
	
	/**
	 * Returns the list of currently available
	 * colours for the specified node.
	 * @param n
	 * @return
	 */
	public IntSet colourList(int n){
		assert(checkNodeIndex(n));
		return colours[n];
	}
	
	/**
	 * Sets the specified node to the specified colour
	 * (ie removes all other values from its domain),
	 * if it belongs to its current domain.
	 * @param node a valid node index
	 * @param colour a valid colour index
	 * @return <code>false</code> if the colour
	 * did not belong to the set of feasible colours
	 */
	public boolean setToColour(int node, int colour){
		assert(checkNodeIndex(node));
		
		IntSet colours = colourList(node);
		if(!colours.contains(colour))
			return false;
		for(Iterator<Integer> it=colours.iterator();it.hasNext();)
			if(it.next() != colour)
				it.remove();
		return true;
	}
	
	private boolean checkColourIndex(int c){
		return c >= 0 && c < maxOccurrences.length;
	}
	
	public int getNColours(){
		return maxOccurrences.length;
	}
	
	public void setOccurrences(int c, int o){
		assert(checkColourIndex(c));
		maxOccurrences[c] = o;
	}
	
	public int getOccurrences(int c){
		assert(checkColourIndex(c));
		return maxOccurrences[c];
	}
	
	/**
	 * Returns the set of the nodes that
	 * has not been assigned a colour.
	 * @return
	 */
	public IntSet getUnassignedNodes(){
		return unassignedNodes;
	}
	
	public int getNumberOfAssignedNodes(){
		return getNNodes()-unassignedNodes.size();
	}
	
	public int getNumberOfUnassignedNodes(){
		return unassignedNodes.size();
	}
	
	public void writeRaw(PrintStream out){
		//Number of nodes
		out.println(getNNodes());
		//Number of colours
		out.println(getNColours());
		//Space separated list of values for each node index
		for(int i=0;i<getNNodes();i++){
			IntSet colours = colourList(i);
			String sep = "";
			for(int c=colours.first();c>-1;c=colours.next(c)){
				out.print(sep+c);
				sep = " ";
			}
			out.println();
		}
		//Max occurrences of each colour index
		{
			String sep = "";
			for (int c = 0; c < getNColours(); c++) {
				out.print(sep + getOccurrences(c));
				sep = " ";
			}
			out.println();
		}
		//List of neighbours for each node index
		for(int i=0;i<getNNodes();i++){
			String sep = "";
			for(int j=firstNeighbour(i);j>-1;j=nextNeighbour(i, j)){
				out.print(sep+j);
				sep = " ";
			}
			out.println();
		}
				
	}
	
	
	public void writeAsDotty(PrintStream out){
		
		out.printf(
			"graph{\n"
		);
		//Write nodes (in case they're unconnected).
		out.printf("\t");
		for(int i=0;i<getNNodes();i++)
			out.printf("n%d [label=\"%s\"];", i, colourList(i));
		out.println();
		//Write edges
		for(int i=0;i<getNNodes();i++)
			for(int j=firstNeighbour(i);j>-1;j=nextNeighbour(i, j))
				if(j>i) out.printf("\tn%d -- n%d;%n",i,j);		
		out.println();
				
		out.printf("}\n");
		
	}
	
	/**
	 * @param args
	 * @throws Exception 
	 */
	public static void main(String[] args) throws Exception {
		testHeuristics();

	}
	
	public static void testHeuristics() throws Exception {
		for(int i=1;i<=8;i++){			
			for (int j = 0; j < 10; j++) {
				String instance = "comp-2007-2-"+i;
				TimetablingInstance ti = new InstanceReader(new FileInputStream("./data/Track2/"+instance+".tim")).getInstance();
				System.out.println("Graph " + instance
						+ " is being processed");
				{
					ColourableConflictGraph graph = ti.getFullConflictGraph();
					
					System.out.print("min dom "+new MinNodeSelector(graph).greedyList());
					double ratio = (double)graph.getNumberOfAssignedNodes()/graph.getNNodes();
					System.out.println(" "+ratio);
				}
				{
					ColourableConflictGraph graph = ti.resetFullConflictGraph();
					
					System.out.print("min dom/deg "+new MinDomOverDegSelector(graph).greedyList());
					double ratio = (double)graph.getNumberOfAssignedNodes()/graph.getNNodes();
					System.out.println(" "+ratio);
				}
				{
					ColourableConflictGraph graph = ti.resetFullConflictGraph();
					
					System.out.print("min dom+max deg "+new MinDomMaxDegSelector(graph).greedyList());
					double ratio = (double)graph.getNumberOfAssignedNodes()/graph.getNNodes();
					System.out.println(" "+ratio);
				}
				{
					ColourableConflictGraph graph = ti.resetFullConflictGraph();
					
					System.out.print("degeneracy "+new DegeneracyOrder(graph).greedyList());
					double ratio = (double)graph.getNumberOfAssignedNodes()/graph.getNNodes();
					System.out.println(" "+ratio);
				}
			}
		}

	}
	

}
