/**
 * 
 */
package instance;

import java.io.PrintStream;
import java.util.Random;

import util.IntSet;

/**
 * Represents the graph of the conflicts between events.
 * Two events are connected iff they must be assigned
 * to different timeslots.
 * 
 * @author alexp
 *
 */
public class ConflictGraph {
	
	/**
	 * For the node i, neighbours[i] contain
	 * the set of neighbours.
	 */
	protected IntSet[] neighbours;
	
	/**
	 * The maintained number of edges.
	 */
	protected int nEdges;
	
	/**
	 * The maximal number of edges.
	 */
	protected int maxNEdges;
	
	/**
	 * The random generator used by this class
	 * (mainly by IntSet instances).
	 */
	protected Random random = new Random();

	/**
	 * Creates a new graph with n nodes.
	 * @param n
	 */
	public ConflictGraph(int n){
		neighbours = new IntSet[n];
		for (int i = 0; i < neighbours.length; i++){
			neighbours[i] = new IntSet(n);
			neighbours[i].setRandom(random);
		}
		maxNEdges = n*(n-1)/2;
	}
	
	/**
	 * Specify the random 
	 * @param r
	 */
	public void setRandom(Random r){
		random = r;
	}
	
	public int getNNodes(){
		return neighbours.length;
	}
	
	public int getNEdges(){
		return nEdges;
	}
	
	/**
	 * Returns the density of the graph, that is
	 * the ratio between the actual and maximal
	 * number of edges.
	 * @return the density of the graph.
	 */
	public double getDensity(){
		return (double)nEdges/maxNEdges;
	}
	
	protected boolean checkNodeIndex(int n){
		return n >=0 && n < getNNodes();
	}
	
	protected boolean checkEdgeIntegrity(int n1, int n2){
		return /*n1 != n2 &&*///We don't need n1 and n2 to be different
			( (neighbours[n1].contains(n2) && neighbours[n2].contains(n1))
			|| (!neighbours[n1].contains(n2) && !neighbours[n2].contains(n1))
			);
	}
	
	public void addEdge(int n1, int n2){
		assert(checkNodeIndex(n1));
		assert(checkNodeIndex(n2));
		assert(checkEdgeIntegrity(n1,n2));
		if(!neighbours[n1].contains(n2)){
			neighbours[n1].add(n2);
			neighbours[n2].add(n1);
			++nEdges;
		}
	}
	
	public boolean containsEdge(int n1, int n2){
		assert(checkNodeIndex(n1));
		assert(checkNodeIndex(n2));
		assert(checkEdgeIntegrity(n1, n2));
		return neighbours[n1].contains(n2);
	}
	
	public void removeEdge(int n1, int n2){
		assert(checkNodeIndex(n1));
		assert(checkNodeIndex(n2));
		assert(checkEdgeIntegrity(n1, n2));
		neighbours[n1].remove(n2);
		neighbours[n2].remove(n1);
		--nEdges;
	}
	
	public int getDegree(int n){
		assert(checkNodeIndex(n));
		return neighbours[n].size();
	}
	
	public int firstNeighbour(int n){
		assert(checkNodeIndex(n));
		return neighbours[n].first();
	}
	
	public int nextNeighbour(int n, int neighbour){
		assert(containsEdge(n, neighbour));
		return neighbours[n].next(neighbour);
	}
	
	/**
	 * Iterator on neighbours. Direct access with the methods
	 * {@link #firstNeighbour(int)} and {@link #nextNeighbour(int, int)}.<br>
	 * Ex, if <code>n</code> is a node:<br>
	 * <code>
	 * for(int j=firstNeighbour(n);j>-1;j=nextNeighbour(n, j)){
	 * 	//j is a neighbour of n.
	 * }
	 * </code>
	 */
	public Iterable<Integer> neighbours(int n){
		assert(checkNodeIndex(n));
		return neighbours[n];//We just mask the type, which is unsafe.
	}
	
	public void writeAsDotty(PrintStream out){
		
		out.printf(
			"graph{\n"+
			"\tnode [shape=point];\n"
		);
		//Write nodes (in case they're unconnected).
		out.printf("\t");
		for(int i=0;i<getNNodes();i++)
			out.printf("n%d;", i);
		out.println();
		//Write edges
		for(int i=0;i<getNNodes();i++)
			for(int j=firstNeighbour(i);j>-1;j=nextNeighbour(i, j))
				if(j>i) out.printf("\tn%d -- n%d;%n",i,j);		
		out.println();
				
		out.printf("}\n");
		
	}
	
	public static void main(String[] args) throws Exception {
		ConflictGraph graph = new ConflictGraph(10);
		
		graph.addEdge(2, 0);		
		graph.addEdge(2, 1);
		
		System.out.println(graph.containsEdge(0, 2));
		graph.addEdge(2, 8);
		graph.removeEdge(0, 2);
		System.out.println(graph.containsEdge(2, 0));
		
		graph.writeAsDotty(System.out);
		
	}
		

}
