/**
 *
 */
package instance;

import java.io.FileInputStream;
import java.io.PrintStream;
import java.util.*;

import tools.InstanceReader;
import instance.clique.MaxCliqueFinder;
import instance.clique.CompoundClique;


/**
 * The description of an instance of the
 * timetabling problem.
 *
 * @author alexp
 */
public class TimetablingInstance {

	/**
	 * The number of timeslots in a week.
	 */
	public final static int N_TIMESLOTS = 45;

	/**
	 * /**
	 * name of the file
	 */
	protected String name;

	/**
	 * The parameters of the instance.
	 */
	protected int nEvents, nRooms, nFeatures, nStudents;

	/**
	 * The sizes of the rooms.
	 */
	protected int[] roomSizes;

	/**
	 * The number of precedences for each event i.e
	 * the number of event e' such that e < e'
	 */
	protected int[] precSizes;

	/**
	 * The number of reverse precedences for each event i.e
	 * the number of event e' such that e' < e
	 */
	protected int[] invPrecSizes;

	/**
	 * The four matrices.
	 */
	protected BitSet[] studentEvent, roomFeature, eventFeature, eventTimeslot, eventRoom;
	protected BitSet[] eventStudent, featureRoom, featureEvent, timeslotEvent, roomEvent;

	protected int[] evtSize;

	/**
	 * The list of event precedences.
	 */
	protected LinkedList<Precedence> eventPrecedences;

	/**
	 * Initialises a new instance with the provided
	 * parameters.
	 *
	 * @param nEvents   number of events
	 * @param nRooms    number of rooms
	 * @param nFeatures number of features
	 * @param nStudents number of students
	 */
	public TimetablingInstance(int nEvents, int nRooms, int nFeatures, int nStudents) {
		this.nEvents = nEvents;
		this.nRooms = nRooms;
		this.nFeatures = nFeatures;
		this.nStudents = nStudents;

		roomSizes = new int[nRooms];
		evtSize = new int[nEvents];
		Arrays.fill(evtSize, -1);

		studentEvent = new BitSet[nStudents];
		for (int i = 0; i < studentEvent.length; i++)
			studentEvent[i] = new BitSet(nEvents);
		eventStudent = new BitSet[nEvents];
		for (int i = 0; i < eventStudent.length; i++)
			eventStudent[i] = new BitSet(nStudents);

		roomFeature = new BitSet[nRooms];
		for (int i = 0; i < roomFeature.length; i++)
			roomFeature[i] = new BitSet(nFeatures);
		featureRoom = new BitSet[nFeatures];
		for (int i = 0; i < featureRoom.length; i++)
			featureRoom[i] = new BitSet(nRooms);

		eventFeature = new BitSet[nEvents];
		for (int i = 0; i < eventFeature.length; i++)
			eventFeature[i] = new BitSet(nFeatures);
		featureEvent = new BitSet[nFeatures];
		for (int i = 0; i < featureEvent.length; i++)
			featureEvent[i] = new BitSet(nEvents);

		eventTimeslot = new BitSet[nEvents];
		for (int i = 0; i < eventTimeslot.length; i++)
			eventTimeslot[i] = new BitSet(N_TIMESLOTS);
		timeslotEvent = new BitSet[N_TIMESLOTS];
		for (int i = 0; i < timeslotEvent.length; i++)
			timeslotEvent[i] = new BitSet(nEvents);

		eventRoom = new BitSet[nEvents];
		for (int i = 0; i < eventRoom.length; i++)
			eventRoom[i] = new BitSet(nRooms);
		roomEvent = new BitSet[nRooms];
		for (int i = 0; i < roomEvent.length; i++)
			roomEvent[i] = new BitSet(nEvents);

		eventPrecedences = new LinkedList<Precedence>();
		precSizes = new int[nEvents];
		invPrecSizes = new int[nEvents];
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	/**
	 * @return the nEvents
	 */
	public int getNEvents() {
		return nEvents;
	}

	/**
	 * @return the nRooms
	 */
	public int getNRooms() {
		return nRooms;
	}

	/**
	 * @return the nFeatures
	 */
	public int getNFeatures() {
		return nFeatures;
	}

	/**
	 * @return the nStudents
	 */
	public int getNStudents() {
		return nStudents;
	}

	/**
	 * @return the number of prec for event e
	 */
	public int getNPrec(int e) {
		assert (checkEventIndex(e));
		return precSizes[e];
	}

	/**
	 * @return the number of prec for event e
	 */
	public int getNInvprec(int e) {
		assert (checkEventIndex(e));
		return invPrecSizes[e];
	}


	/**
	 * Returns the size of the specified room.
	 *
	 * @param room a valid room index
	 * @return its size
	 */
	public int getRoomSize(int room) {
		assert (checkRoomIndex(room));

		return roomSizes[room];
	}

	/**
	 * Sets the size of the specified room.
	 *
	 * @param room a valid room index
	 * @param size its size
	 */
	public void setRoomSize(int room, int size) {
		assert (checkRoomIndex(room));

		roomSizes[room] = size;
	}

	/**
	 * Specify that the given student should attend
	 * the given event.
	 *
	 * @param student a valid index for a student
	 * @param event   a valid index for an event
	 */
	public void setStudentForEvent(int student, int event) {
		assert (checkStudentIndex(student));
		assert (checkEventIndex(event));

		studentEvent[student].set(event);
		eventStudent[event].set(student);
	}

	/**
	 * Specify that the given room satisfies the
	 * given feature.
	 *
	 * @param room    a valid index for a room
	 * @param feature a valid index for a feature
	 */
	public void setRoomForFeature(int room, int feature) {
		assert (checkRoomIndex(room));
		assert (checkFeatureIndex(feature));

		roomFeature[room].set(feature);
		featureRoom[feature].set(room);
	}

	/**
	 * Specify that the given event requires the
	 * given feature.
	 *
	 * @param event   a valid index for an event
	 * @param feature a valid index for a feature
	 */
	public void setEventForFeature(int event, int feature) {
		assert (checkEventIndex(event));
		assert (checkTimeslotIndex(feature));

		eventFeature[event].set(feature);
		featureEvent[feature].set(event);
	}

	/**
	 * Specify that the given event can be
	 * assigned to the given timeslot.
	 *
	 * @param event    a valid index for an event
	 * @param timeslot a valid index for a timeslot
	 */
	public void setEventForTimeslot(int event, int timeslot) {
		assert (checkEventIndex(event));
		assert (checkTimeslotIndex(timeslot));

		eventTimeslot[event].set(timeslot);
		timeslotEvent[timeslot].set(event);
	}

	/**
	 * Computes the matrix that associates compatible
	 * event and rooms, as detected by the
	 * {@link #isRoomAvailableForEvent(int,int)}
	 * method. This method should be called after
	 * all the basic matrices have been fully defined.
	 *
	 * @see #isRoomAvailableForEvent(int,int)
	 */
	public void fillEventRoomMatrix() {
		for (int e = 0; e < nEvents; e++) {
			for (int r = 0; r < nRooms; r++) {
				if (isRoomAvailableForEvent(e, r)) {
					eventRoom[e].set(r);
					roomEvent[r].set(e);
				}
			}
		}
	}

	/**
	 * Specify that a precedences constraint holds
	 * between the specified events.
	 *
	 * @param firstEvent
	 * @param secondEvent
	 */
	public void addPrecedence(int firstEvent, int secondEvent) {
		assert (checkEventIndex(firstEvent));
		assert (checkEventIndex(secondEvent));
		eventPrecedences.add(new Precedence(firstEvent, secondEvent));
		precSizes[firstEvent]++;
		invPrecSizes[secondEvent]++;
	}

	public BitSet eventsAttendedByStudent(int student) {
		assert (checkStudentIndex(student));

		return studentEvent[student];
	}

	public BitSet studentsAttendingEvent(int event) {
		assert (checkEventIndex(event));

		return eventStudent[event];
	}

	public BitSet featuresSatisfiedByRoom(int room) {
		assert (checkRoomIndex(room));

		return roomFeature[room];
	}

	public BitSet roomsSatisfyingFeature(int feature) {
		assert (checkFeatureIndex(feature));

		return featureRoom[feature];
	}

	public BitSet featuresRequiredByEvent(int event) {
		assert (checkEventIndex(event));

		return eventFeature[event];
	}

	public BitSet eventsRequiringFeature(int feature) {
		assert (checkFeatureIndex(feature));

		return featureEvent[feature];
	}

	public BitSet timeslotsAssignableToEvent(int event) {
		//TODO Find a better name
		assert (checkEventIndex(event));
		return eventTimeslot[event];
	}

	public BitSet eventsAvailableForTimeslot(int timeslot) {
		assert (checkTimeslotIndex(timeslot));

		return timeslotEvent[timeslot];
	}

	public BitSet roomsAvailableForEvent(int event) {
		assert (checkEventIndex(event));

		return eventRoom[event];
	}

	public BitSet eventsFittingToRoom(int room) {
		assert (checkRoomIndex(room));

		return roomEvent[room];
	}

	/**
	 * Specify that the given student should attend
	 * the given event.
	 *
	 * @param student a valid index for a student
	 * @param event   a valid index for an event
	 */
	public boolean studentAttendsEvent(int student, int event) {
		assert (checkStudentIndex(student));
		assert (checkEventIndex(event));

		assert (studentEvent[student].get(event) == eventStudent[event].get(student));

		return studentEvent[student].get(event);
	}

	/**
	 * Specify that the given room satisfies the
	 * given feature.
	 *
	 * @param room    a valid index for a room
	 * @param feature a valid index for a feature
	 */
	public boolean roomSatisfiesFeature(int room, int feature) {
		assert (checkRoomIndex(room));
		assert (checkFeatureIndex(feature));

		assert (roomFeature[room].get(feature) == featureRoom[feature].get(room));

		return roomFeature[room].get(feature);
	}

	/**
	 * Specify that the given event requires the
	 * given feature.
	 *
	 * @param event   a valid index for an event
	 * @param feature a valid index for a feature
	 */
	public boolean eventRequiresFeature(int event, int feature) {
		assert (checkEventIndex(event));
		assert (checkFeatureIndex(feature));

		assert (eventFeature[event].get(feature) == featureEvent[feature].get(event));

		return eventFeature[event].get(feature);
	}

	/**
	 * Specify that the given event can be assigned
	 * to the given timeslot.
	 *
	 * @param event    a valid index for an even
	 * @param timeslot a valid index for a timeslot
	 */
	public boolean eventIsCompatibleWithTimeslot(int event, int timeslot) {
		assert (checkEventIndex(event));
		assert (checkTimeslotIndex(timeslot));

		assert (eventTimeslot[event].get(timeslot) == timeslotEvent[timeslot].get(event));

		return eventTimeslot[event].get(timeslot);
	}

	public boolean eventFitsInRoom(int event, int room) {
		assert (checkEventIndex(event));
		assert (checkRoomIndex(room));

		assert (eventRoom[event].get(room) == roomEvent[room].get(event));

		return eventRoom[event].get(room);
	}

	/**
	 * Return the number of students following event
	 *
	 * @param event a valid index for an event
	 */
	public int getNbStudentInEvent(int event) {
		assert (checkEventIndex(event));
		if (evtSize[event] == -1) {
			evtSize[event] = studentsAttendingEvent(event).cardinality();
		}
		return evtSize[event];
	}

	/**
	 * Return the number feasible timeslot for event
	 *
	 * @param event a valid index for an event
	 */
	public int getNbFeasTimeslotForEvent(int event) {
		assert (checkEventIndex(event));
		return eventTimeslot[event].cardinality();
	}

	/**
	 * Specify if the two given events share a student or not
	 * (meaning a conflict between them or not)
	 *
	 * @param event1 a valid event
	 * @param event2 another valid event
	 */
	public boolean eventsShareStudents(int event1, int event2) {
		assert (event1 != event2);
		assert (checkEventIndex(event1));
		assert (checkEventIndex(event2));
		boolean intersect = false;
		for (int k = 0; k < getNStudents() && !intersect; k++) {
			intersect = (studentAttendsEvent(k, event1) &&
					studentAttendsEvent(k, event2));
		}
		return intersect;
	}


	/**
	 * Specify if the two given events can only use a single same room.
	 * if yes there is a conflict between them in the sense that they cannot
	 * be in the same timeslot.
	 *
	 * @param event1 a valid event
	 * @param event2 another valid event
	 * @return true if the two events are different and the can use
	 *         only a single same room.
	 */
	public boolean isRoomConflictBetweenEvents(int event1, int event2) {
		assert (checkEventIndex(event1));
		assert (checkEventIndex(event2));

		if (event1 == event2) return false;

		return roomsAvailableForEvent(event1).equals(roomsAvailableForEvent(event2))
				&& roomsAvailableForEvent(event1).cardinality() == 1
				&& roomsAvailableForEvent(event2).cardinality() == 1;
	}

	/**
	 * Indicates if the available timeslots for the specified events
	 * are disjoint, thus preventing these events from taking place
	 * at the same timeslot.
	 *
	 * @param event1 a valid event
	 * @param event2 a valid event
	 * @return <code>true</code> if the two events have a timeslot
	 *         availability conflict.
	 */
	public boolean isTimeslotConflictBetweenEvents(int event1, int event2) {
		return !eventTimeslot[event1].intersects(eventTimeslot[event2]);
	}

	/**
	 * @param event a valid event
	 * @param room  a valid room
	 */
	public boolean isRoomAvailableForEvent(int event, int room) {
		assert (checkEventIndex(event));
		assert (checkRoomIndex(room));
		for (int k = 0; k < getNFeatures(); k++) {
			if (eventRequiresFeature(event, k) &&
					!roomSatisfiesFeature(room, k)) return false;
		}
		if (getNbStudentInEvent(event) > getRoomSize(room))
			return false;
		return true;
	}

	public double getAverageAvailableTimeslotsPerEvent() {
		double avg = 0.0;
		for (int i = 0; i < nEvents; i++) {
			avg += (double) timeslotsAssignableToEvent(i).cardinality();
		}
		return avg / (double) nEvents;
	}

	public double getAverageCompatibleRoomsPerEvent() {
		double avg = 0.0;
		for (int i = 0; i < nEvents; i++)
			avg += roomsAvailableForEvent(i).cardinality();

		return avg / nEvents;
	}

	public double getAverageEventsPerStudent() {
		double avg = 0.0;
		for (int i = 0; i < nStudents; i++) {
			avg += eventsAttendedByStudent(i).cardinality();
		}

		return avg / nStudents;
	}

	public double getAverageStudentsPerEvent() {
		double avg = 0.0;
		for (int i = 0; i < nEvents; i++) {
			avg += getNbStudentInEvent(i);
		}

		return avg / nEvents;
	}

	/* Index validity checks */
	private boolean checkEventIndex(int event) {
		return (event >= 0)
				&& (event < nEvents);
	}

	private boolean checkRoomIndex(int room) {
		return (room >= 0)
				&& (room < nRooms);
	}

	private boolean checkFeatureIndex(int feature) {
		return (feature >= 0)
				&& (feature < nFeatures);
	}

	private boolean checkStudentIndex(int student) {
		return (student >= 0)
				&& (student < nStudents);
	}

	private boolean checkTimeslotIndex(int timeslot) {
		return (timeslot >= 0)
				&& (timeslot < N_TIMESLOTS);
	}

	/**
	 * Returns the number of precedences.
	 *
	 * @return the number of precedences
	 */
	public int getNPrecedences() {
		return eventPrecedences.size();
	}

	/**
	 * Returns an iterable view of the set of
	 * precedences between events, for use
	 * in a foreach loop.
	 *
	 * @return
	 */
	public Iterable<Precedence> eventPrecedences() {

		if (evtPrec == null)
			evtPrec = new Iterable<Precedence>() {

				public Iterator<Precedence> iterator() {
					return eventPrecedences.iterator();
				}

				public String toString() {
					return basicToString();
				}

				private String basicToString() {
					return eventPrecedences.toString();
				}

				@SuppressWarnings("unused")
				private String matrixToString() {
					return null;//TODO Print the precedences matrix as it is in the instance file
				}

			};

		return evtPrec;
	}

	private Iterable<Precedence> evtPrec;

	/**
	 * Indicates whether the first specified event
	 * must precede the second one.<br>
	 * Makes the test by traversing the list of precedences.
	 *
	 * @param firstEvent  a valid event index
	 * @param secondEvent a valid event index
	 */
	public boolean precedes(int firstEvent, int secondEvent) {
		assert (checkEventIndex(firstEvent));
		assert (checkEventIndex(secondEvent));

		for (Precedence p : eventPrecedences())
			if (p.firstEvent == firstEvent
					&& p.secondEvent == secondEvent)
				return true;
		return false;
	}

	/**
	 * Indicates whether the two events are constrained
	 * by a precedence constraint, i.e. there is one
	 * that must take place before the other.
	 */
	public boolean precedenceConstraint(int event1, int event2) {
		assert (checkEventIndex(event1));
		assert (checkEventIndex(event2));
		for (Precedence p : eventPrecedences()) {
			if (p.firstEvent == event1
					&& p.secondEvent == event2)
				return true;
			if (p.firstEvent == event2
					&& p.secondEvent == event1)
				return true;
		}
		return false;
	}

	/**
	 * Returns a string representation of the sizes
	 * of the rooms.
	 *
	 * @return a string representation of the array of sizes
	 *         of the rooms
	 */
	public String roomSizes() {
		return Arrays.toString(roomSizes);
	}

	/**
	 * Indicate that the two events cannot be assigned to the same
	 * timeslot.
	 *
	 * @param event1 a valid event
	 * @param event2 a valid event
	 * @return <code>true</code> only if the events cannot be assigned
	 *         to the same timeslot (but not necessarily a complete test).
	 */
	public boolean isConflictBetweenEvents(int event1, int event2) {
		return simpleConflictsBetweenEvents(event1, event2);
	}

	protected boolean simpleConflictsBetweenEvents(int event1, int event2) {
		return eventsShareStudents(event1, event2)
				|| precedenceConstraint(event1, event2);
	}

	protected boolean allConflictsBetweenEvents(int event1, int event2) {
		return simpleConflictsBetweenEvents(event1, event2)
				|| isRoomConflictBetweenEvents(event1, event2)
				|| isTimeslotConflictBetweenEvents(event1, event2);
	}

	/**
	 * Builds (upon first call) and returns the graph of the conflicts
	 * between the events. The conflicts are detected according to the method
	 * {@link #isConflictBetweenEvents(int,int)}.
	 *
	 * @return the conflict graph
	 * @see ColourableConflictGraph
	 */
	public ColourableConflictGraph getEventConflictGraph() {
		if (colouredGraph == null) {
			colouredGraph = new ColourableConflictGraph(nEvents, N_TIMESLOTS, nRooms);

			for (int i = 0; i < nEvents; i++) {
				BitSet tsl = timeslotsAssignableToEvent(i);
				for (int c = tsl.nextSetBit(0); c > -1; c = tsl.nextSetBit(c + 1))
					colouredGraph.colourList(i).add(c);
			}

			for (int i = 0; i < nEvents; i++)
				for (int j = 0; j < nEvents; j++)
					if (i < j
							&& isConflictBetweenEvents(i, j))
						colouredGraph.addEdge(i, j);
		}

		return colouredGraph;
	}

	protected ColourableConflictGraph colouredGraph;

	/**
	 * Builds (upon first call) and returns the graph of all
	 * the conflicts between the events, detected with the method
	 * {@link #allConflictsBetweenEvents(int,int)}.
	 *
	 * @return the conflict graph
	 * @see ColourableConflictGraph
	 */
	public ColourableConflictGraph getFullConflictGraph() {
		if (colouredFullConflictGraph == null) {
			resetFullConflictGraph();
		}

		return colouredFullConflictGraph;
	}

	protected ColourableConflictGraph colouredFullConflictGraph;

	/**
	 * Rebuilds the graph of all the conflicts between the events.
	 *
	 * @return
	 * @see #getFullConflictGraph()
	 */
	public ColourableConflictGraph resetFullConflictGraph() {
		colouredFullConflictGraph = new ColourableConflictGraph(nEvents, N_TIMESLOTS, nRooms);

		for (int i = 0; i < nEvents; i++) {
			BitSet tsl = timeslotsAssignableToEvent(i);
			for (int c = tsl.nextSetBit(0); c > -1; c = tsl.nextSetBit(c + 1))
				colouredFullConflictGraph.colourList(i).add(c);
		}

		for (int i = 0; i < nEvents; i++)
			for (int j = 0; j < nEvents; j++)
				if (i < j
						&& allConflictsBetweenEvents(i, j))
					colouredFullConflictGraph.addEdge(i, j);

		return colouredFullConflictGraph;
	}

	/**
	 * Returns the density of the conflict graph.
	 *
	 * @return the "conflict density" of this instance.
	 */
	public double getConflictDensity() {
		return getEventConflictGraph().getDensity();
	}

	public double getFullConflictDensity() {
		return getFullConflictGraph().getDensity();
	}

	public double getRoomConflictDensity() {
		ConflictGraph graph = new ConflictGraph(nEvents);

		for (int i = 0; i < nEvents; i++)
			for (int j = 0; j < nEvents; j++)
				if (i < j
						&& (simpleConflictsBetweenEvents(i, j)
						|| isRoomConflictBetweenEvents(i, j)))
					graph.addEdge(i, j);
		return graph.getDensity();
	}

	public double getTimeslotConflictDensity() {
		ConflictGraph graph = new ConflictGraph(nEvents);

		for (int i = 0; i < nEvents; i++)
			for (int j = 0; j < nEvents; j++)
				if (i < j
						&& (simpleConflictsBetweenEvents(i, j)
						|| isTimeslotConflictBetweenEvents(i, j)))
					graph.addEdge(i, j);
		return graph.getDensity();
	}

	/**
	 * Returns a string representation of the
	 * matrix of student attendance to events.
	 *
	 * @return a string representation of
	 *         the student/event matrix
	 */
	public String studentEventMatrix() {
		return bitsetToMatrix(studentEvent, nEvents);
	}

	/**
	 * Returns a string representation of the
	 * matrix of room satisfaction of features.
	 *
	 * @return a string representation of
	 *         the room/feature matrix
	 */
	public String roomFeatureMatrix() {
		return bitsetToMatrix(roomFeature, nFeatures);
	}

	/**
	 * Returns a string representation of the
	 * matrix of event requirement of features.
	 *
	 * @return a string representation of
	 *         the event/feature matrix
	 */
	public String eventFeatureMatrix() {
		return bitsetToMatrix(eventFeature, nFeatures);
	}

	/**
	 * Returns a string representation of the
	 * matrix event timeslot compatibility.
	 *
	 * @return a string representation of
	 *         the event/feature matrix
	 */
	public String eventTimeslotMatrix() {
		return bitsetToMatrix(eventTimeslot, N_TIMESLOTS);
	}

	public String eventRoomMatrix() {
		return bitsetToMatrix(eventRoom, nRooms);
	}

	private String bitsetToMatrix(BitSet[] bitset, int nCols) {
		StringBuffer matrix = new StringBuffer(bitset.length * (nCols * 2 + 1));
		for (int row = 0; row < bitset.length; row++) {
			String sep = "";
			for (int col = 0; col < nCols; col++) {
				matrix.append(sep);
				sep = "\t";
				if (bitset[row].get(col))
					matrix.append(1);
				else
					matrix.append(0);
			}
			matrix.append("\n");
		}
		return matrix.toString();
	}

	/**
	 * @return true if the instance is considered as a hard one from the list coloring
	 *         point of view and an easy one from the matching point of view.
	 */
	public boolean is_hardcolorANDeasymatch() {
		double crit = (double) nRooms / (double) nEvents;
		return crit <= 0.03;
	}

	public static double[] speed = new double[]{0.07, 0.12, 0.15};

	public double getSoftSpeed() {
		// nevents
		double score = (((double) nEvents - 200d) / 200d);
		//nstudents
		score = score + (((double) nStudents - 300d) / 700d);
		//nrooms
		score = score + 1 - (((double) nRooms - 10d) / 10d);
		int res = 1;
		if (score <= 1.5)
			res = 0;
		if (score > 2)
			res = 2;
		return speed[res];
	}

	private static void writeGraphs() throws Exception {
		for (int i = 1; i <= 8; i++) {
			String instance = "comp-2007-2-" + i;
			TimetablingInstance ti = new InstanceReader(new FileInputStream("./data/Track2/" + instance + ".tim")).getInstance();
			ColourableConflictGraph graph = ti.getEventConflictGraph();
			ColourableConflictGraph completeGraph = ti.getFullConflictGraph();

			graph.writeRaw(new PrintStream("./output/graphs/light/" + instance + ".graph"));
			completeGraph.writeRaw(new PrintStream("./output/graphs/heavy/" + instance + ".graph"));

			System.out.println("Graph " + instance + " written");
		}
	}

	private static void printParameters() throws Exception {

		//System.out.println("inst & nevt & nstn & nrms & rms/event & rm ratio & evt/stn & stn/evt & tsl/evt  & basic dens & tsl dens & rm dens & comp dens \\\\");
		for (int i = 1; i <= 16; i++) {
			TimetablingInstance ti = new InstanceReader(new FileInputStream("./data/Track2/comp-2007-2-" + i + ".tim")).getInstance();

			double avgRoomsPerEvent = ti.getAverageCompatibleRoomsPerEvent();

			System.out.printf("%d & %d & %d & %d & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f\\\\%n",
					i,
					ti.getNEvents(),
					ti.getNStudents(),
					ti.getNRooms(),
					avgRoomsPerEvent,
					avgRoomsPerEvent / ti.getNRooms(),
					ti.getAverageEventsPerStudent(),
					ti.getAverageStudentsPerEvent(),
					ti.getAverageAvailableTimeslotsPerEvent(),
					ti.getConflictDensity(),
					ti.getTimeslotConflictDensity(),
					ti.getRoomConflictDensity(),
					ti.getFullConflictDensity());

			//following code was to answer diarmud requests :
			/*ConflictGraph g = ti.getFullConflictGraph();
			ConflictGraph g2 = ti.getEventConflictGraph();
			System.out.println(g2.getNEdges() + " " + g.getNEdges());
			for (int j = 0; j < ti.getNEvents(); j++) {
				System.out.println("Event " + j + ": " + ti.timeslotsAssignableToEvent(j).cardinality()+ " - " + g2.getDegree(j) + " - " + g.getDegree(j));
			}*/

		}

	}

    /**
     * A simple example of how to call the clique finder
     * @param args
     * @throws Exception
     */
    public static void main(String[] args) throws Exception {
        TimetablingInstance ins = new InstanceReader(new FileInputStream("./data/Track2/comp-2007-2-1.tim")).getInstance();
        MaxCliqueFinder mc = new MaxCliqueFinder(ins);
		mc.findCliques(true);
		CompoundClique[] cc = mc.getCliques();
		for (CompoundClique clique : cc) {
            System.out.println(clique);
		}
    }

}
