package instance.clique;

import instance.ConflictGraph;

import java.util.BitSet;
import java.util.Comparator;

/**
 * The class representing a clique.
 *
 */
public abstract class Clique{
	
	protected static final Comparator<Clique> DECREASING_ORDER = 
		new Comparator<Clique>(){
			public int compare(Clique c1, Clique c2) {
				return Integer.signum(c2.getSize() - c1.getSize());
			}
		};
		
	protected static final Comparator<Clique> INCREASING_ORDER = 
		new Comparator<Clique>(){
			public int compare(Clique c1, Clique c2) {
				return Integer.signum(c1.getSize() - c2.getSize());
			}
		};
	
	public abstract int getSize();
	
	/**
	 * Computes and returns the set of nodes of this clique.
	 * Each instantiation must return a newly allocated bitset.
	 * @return
	 */
	public abstract BitSet retrieveNodes();
	
	/**
	 * Returns the graph in which the nodes of this clique belong.
	 * @return
	 */
	public abstract ConflictGraph getGraph();
	
	public boolean isClique(){
		ConflictGraph graph = getGraph();
		BitSet nodes = retrieveNodes();
		
		for(int n1 = nodes.nextSetBit(0);n1 > -1; n1 = nodes.nextSetBit(n1+1)){
			for(int n2 = nodes.nextSetBit(0); n2 > -1; n2 = nodes.nextSetBit(n2+1)){
				if(n1 != n2 && !graph.containsEdge(n1, n2)) return false;
			}
		}
		
		return true;
	}
	
	/**
	 * Returns the intersection of the neighbours
	 * of all the nodes in this clique.
	 * 
	 * @return
	 */
	public BitSet globalNeighbourhood(){
		BitSet nodes = retrieveNodes();
		ConflictGraph graph = getGraph();
		BitSet globalNeighbours = new BitSet(graph.getNNodes());
		
		globalNeighbours.set(0, globalNeighbours.size());//Full of one's.
		
		for(int node=nodes.nextSetBit(0);node>-1;node=nodes.nextSetBit(node+1)){
			BitSet neighbours = new BitSet(graph.getNNodes());
			for(int n1=graph.firstNeighbour(node);n1>-1;n1=graph.nextNeighbour(node, n1))
				neighbours.set(n1);
			globalNeighbours.and(neighbours);
		}
		return globalNeighbours;		
	}
	
	public String toString(){
		return Integer.toString(getSize()) + " " + retrieveNodes();
	}
	
	
}