/**
 * 
 */
package instance.clique;

import instance.ConflictGraph;
import instance.TimetablingInstance;

import java.io.FileInputStream;
import java.util.BitSet;

import tools.InstanceReader;
import util.IntSet;

/**
 * @author alexp
 *
 */
public class JointGraph extends ConflictGraph {

	/**
	 * The original graph (event conflicts).
	 */
	protected ConflictGraph graph;
	
	/**
	 * And the cliques it contains.
	 */
	protected CompoundClique[] cliques;
	
	/**
	 * <code>cliquesOfEvent[i]</code> contains the cliques
	 * containing the event <code>i</code>.
	 */
	protected IntSet[] cliquesOfEvent;
	
	/**
	 * <code>cliqueIntsections[i][j]</code> is a
	 * BitSet of the intersection between the clique
	 * i and the clique j. 
	 */
	protected BitSet[][] cliqueIntersections;
	
	/**
	 * currentDegree[i] contains the number of neighbours
	 * with which the intersection is not empty.
	 */
	protected int[] currentDegree;
		
	/**
	 * Builds a new joint graph for a given conflict graph
	 * whose cliques have been determined by the
	 * specified clique finder.
	 * @param mc a clique finder, on which a procedure
	 * that finds clique must have already been
	 * called.
	 */
	public JointGraph(MaxCliqueFinder mc) {
		super(mc.getCliques().length);
		cliques = new CompoundClique[mc.getCliques().length];
		System.arraycopy(mc.getCliques(), 0, cliques, 0, cliques.length);
		graph = mc.getGraph();
		currentDegree = new int[getNNodes()];
		initCliqueStructures();		
		computeJointGraph();
	}
	
	protected void initCliqueStructures(){
		cliquesOfEvent = new IntSet[graph.getNNodes()];
		for(int i=0;i<cliquesOfEvent.length;i++){
			cliquesOfEvent[i] = new IntSet(cliques.length);
		}		
		for(int c=0;c<cliques.length;c++){
			BitSet events = getEvents(c);
			for(int e = events.nextSetBit(0); e > -1; e = events.nextSetBit(e+1)){
				cliquesOfEvent[e].add(c);
			}
		}
		cliqueIntersections = new BitSet[getNNodes()][];
		for(int i=0;i<cliqueIntersections.length;i++){
			cliqueIntersections[i] = new BitSet[getNNodes()];
		}
	}
	
	/**
	 * Computes the "complete" joint graph,
	 * i.e. the clique intersection graph.
	 */
	protected void computeJointGraph(){
		for(int c1 = 0; c1 < cliques.length; c1++){
			BitSet events1 = getEvents(c1);
			for(int c2 = 0; c2 < cliques.length; c2++){
				if(c1 < c2){
					BitSet events2 = getEvents(c2);
					events2.and(events1);
					if(events2.cardinality() > 0){
						addEdge(c1, c2);
						cliqueIntersections[c1][c2] =
						cliqueIntersections[c2][c1] = events2;
						++currentDegree[c1];
						++currentDegree[c2];
					}
				}
			}
		}
	}
	
	public CompoundClique[] getCliques(){
		return cliques;
	}
	
	/**
	 * Returns the events contained in the
	 * specified clique
	 * @param clique a valid clique index
	 * @return a newly allocated and computed
	 * bitset.
	 */
	public BitSet getEvents(int clique){
		assert(checkNodeIndex(clique));
		return cliques[clique].retrieveNodes();
	}
	
	public IntSet getCliquesOfEvent(int event){
		assert(checkEventIndex(event));
		return cliquesOfEvent[event];
	}
	
	public BitSet getCliqueIntersection(int clique1, int clique2){
		assert(checkNodeIndex(clique1));
		assert(checkNodeIndex(clique2));
		assert(checkEdgeIntegrity(clique1, clique2));
		assert((cliqueIntersections[clique1][clique2] == null && !containsEdge(clique1, clique2))
			||(cliqueIntersections[clique1][clique2] != null && containsEdge(clique1, clique2)));
		
		return cliqueIntersections[clique1][clique2];
	}
	
	
	protected boolean checkEventIndex(int event){
		return event >= 0 && event < cliquesOfEvent.length;
	}
	
	public void removeEvent(int event){
		//TODO OK for the moment because the graph is complete,
		//but we should rather explore the subgraph induced by 'cliques'
		IntSet cliques = getCliquesOfEvent(event);		
		for(int c1 = cliques.first(); c1 > -1; c1 = cliques.next(c1)){
			for(int c2 = cliques.first(); c2 > -1; c2 = cliques.next(c2)){
				if(c1 < c2){
					BitSet intersection = getCliqueIntersection(c1, c2);
					if(intersection != null){//For the moment it should always be the case as we build the complete graph
						intersection.clear(event);
						if(intersection.cardinality() == 0){
							--currentDegree[c1];
							--currentDegree[c2];
						}
					}
				}
			}
		}
	}
	
	public void restoreEvent(int event){
		//TODO Works only for complete graph because
		//we add back all intersections. Otherwise, we should
		//the original state (that is the list of edges kept
		//for a given event).
		IntSet cliques = getCliquesOfEvent(event);
		for(int c1 = cliques.first(); c1 > -1; c1 = cliques.next(c1)){
			for(int c2 = cliques.first(); c2 > -1; c2 = cliques.next(c2)){
				if(c1 < c2){
					BitSet intersection = getCliqueIntersection(c1, c2);
					if(intersection != null){						
						if(intersection.cardinality() == 0){
							++currentDegree[c1];
							++currentDegree[c2];
						}
						intersection.set(event);
					}
				}
			}
		}
	}
	
	public int getCurrentDegree(int clique){
		assert(checkNodeIndex(clique));
		
		return currentDegree[clique];
	}

	public static void main(String[] args) throws Exception {
		for(int i=1;i<=8;i++){
			TimetablingInstance ti = new InstanceReader(new FileInputStream("./data/Track2/comp-2007-2-"+i+".tim")).getInstance();
			MaxCliqueFinder mc = new MaxCliqueFinder(ti);
			mc.findCliques(true);
			JointGraph jGraph = new JointGraph(mc);
			
			//System.out.println(jGraph.getCliqueIntersection(10, 21));
			int[] events = {7, 19, 31, 32, 59, 88, 93, 113, 126, 146, 196, 251, 255, 265, 267, 272, 279, 289, 382};
			System.out.println(jGraph.getCurrentDegree(10));
			for(int neighbour : jGraph.neighbours(10))
				System.out.println(jGraph.getCliqueIntersection(10, neighbour));
			jGraph.removeEvent(7);
			System.out.println(jGraph.getCurrentDegree(10));
			
			break;
		}
	}
}
