/**
 * 
 */
package instance.clique;

import instance.ConflictGraph;
import instance.TimetablingInstance;

import java.io.FileInputStream;
import java.util.Arrays;
import java.util.BitSet;

import tools.InstanceReader;
import util.IntSet;

/**
 * @author alexp
 *
 */
public class MaxCliqueFinder {

	protected ConflictGraph graph;
	
	protected TimetablingInstance ti;
	
	/**
	 * The elementary cliques in decreasing size order.
	 */
	protected CompoundClique[] cliques;
	
	/**
	 * The dual cliques in decreasing size order.
	 * one clique per event, t
	 */
	protected CompoundClique[] dualCliques;
	
	/**
	 * The association between critical rooms and events.
	 * A room is critical if it is the only room available
	 * for an event.
	 * If <code>criticalRooms.contains(i)</code>, then
	 * <code>eventsOfCriticalRooms[i]</code> is the set
	 * of events that have this single possible room.
	 */
	protected IntSet criticalRooms;
	protected IntSet[] eventsOfCriticalRooms;
	
	/**
	 * 
	 */
	public MaxCliqueFinder(TimetablingInstance ti) {
		this.ti = ti;
		this.graph = ti.getFullConflictGraph();
		initCriticalRooms();
	}
	
	protected void initCriticalRooms(){
		criticalRooms = new IntSet(ti.getNRooms());
		eventsOfCriticalRooms = new IntSet[ti.getNRooms()];
		for(int i=0;i<eventsOfCriticalRooms.length;i++)
			eventsOfCriticalRooms[i] = new IntSet(ti.getNEvents());
		
		for(int e=0; e < ti.getNEvents(); e++){
			BitSet rooms = ti.roomsAvailableForEvent(e);
			if(rooms.cardinality() == 1){
				int room = rooms.nextSetBit(0);
				assert(rooms.nextSetBit(room+1) == -1);//It's the only element
				criticalRooms.add(room);
				eventsOfCriticalRooms[room].add(e);
			}
		}
		
	}
	
	public TimetablingInstance getProblem(){
		return ti;
	}
	
	public ConflictGraph getGraph(){
		return graph;
	}
	
	/**
	 * Returns the cliques as they have been computed.
	 * The cliques are computed after a call to
	 * {@link #findCliques(boolean)}, or secondarily to
	 * {@link #initElementaryCliques(boolean)}, {@link #filterDominated()}
	 * {@link #oneCliqueFirst(boolean, instance.clique.MaxCliqueFinder.Filter)} or
	 * {@link #allCliquesFirst(boolean)}; 
	 * @return an array of cliques of the graph, which is complete in the sense
	 * that every edge of the graph is included in at least one clique.
	 */
	public CompoundClique[] getCliques(){
		return cliques;
	}
	
	public boolean isCritical(int room){
		return criticalRooms.contains(room);
	}
	
	public IntSet getCriticalRooms(){
		return criticalRooms;
	}
	
	public IntSet getEvents(int room){
		assert(isCritical(room));
		
		return eventsOfCriticalRooms[room];
	}
	
	/**
	 * Removes from the current set of cliques those that are
	 * a included in another.
	 */
	public void filterDominated(){
		int nCliques = 0;
		
		for(int c1=0;c1<cliques.length;c1++){
			if(cliques[c1] == null) continue;
			BitSet nodes1 = cliques[c1].retrieveNodes();
			++nCliques;
			for(int c2=c1+1;c2<cliques.length;c2++){
				if(cliques[c2] == null) continue;
				BitSet nodes2 = cliques[c2].retrieveNodes();
				nodes2.andNot(nodes1);
				if(nodes2.cardinality() == 0)
					cliques[c2] = null;
			}
		}
		
		CompoundClique[] filteredCliques = new CompoundClique[nCliques];
		int cf=0;
		for(int c=0;c<cliques.length;c++){
			if(cliques[c] == null) continue;
			filteredCliques[cf++] = cliques[c];
		}
		cliques = filteredCliques;
	}
	
	public void printIntersections(){
		for(int c1=0;c1<cliques.length;c1++){
			BitSet nodes1 = cliques[c1].retrieveNodes();
			for(int c2=0;c2<cliques.length;c2++){
				BitSet nodes2 = cliques[c2].retrieveNodes();
				int s = nodes2.cardinality();
				nodes2.and(nodes1);
				s = s - nodes2.cardinality();
				if(s < 3) System.out.print(s+"\t");
			}
			System.out.println();
		}
	}
	
	public void initElementaryCliques(boolean withRooms){
		int nStudents = ti.getNStudents();
		IntSet cr = getCriticalRooms();
		
		if(withRooms)
			cliques = new CompoundClique[nStudents+getCriticalRooms().size()];
		else
			cliques = new CompoundClique[nStudents];
		
		int i=0;
		for(; i<ti.getNStudents();i++){
			cliques[i] = new CompoundClique(this, new StudentClique(this, i), null);
		}
		
		if(withRooms){
			for(int r=cr.first();r>-1;r=cr.next(r)){
				cliques[i++] = new CompoundClique(this, new RoomClique(this, r), null);				
			}
		}
		
		
		Arrays.sort(cliques, Clique.DECREASING_ORDER);
	}
	
	
	public void findCliques(boolean withRooms){
		oneCliqueFirst(withRooms, Filter.PRE);
		//allCliquesFirst(withRooms);
	}
	
	/**
	 * Specify whether the filtering should
	 * be done before clique computation,
	 * after, or never.
	 */
	public enum Filter {PRE, POST, NONE};
	
	public void allCliquesFirst(boolean withRooms, Filter filter){
		initElementaryCliques(withRooms);
		if(filter == Filter.PRE) filterDominated();
		
		boolean augmented = true;
		while(augmented){
			augmented = false;
			CompoundClique[] newCliques = new CompoundClique[cliques.length];
			for(int c = 0; c < cliques.length; c++) {
				BitSet neighbours = cliques[c].globalNeighbourhood(),
					newClique = null;//The set of nodes to add to augment the clique.
				
				int max = 0;
				for(int cp = 0; cp < cliques.length; cp++){
					if(c != cp){
						BitSet nodesp = cliques[cp].retrieveNodes();
						nodesp.and(neighbours);
						int card;
						if((card=nodesp.cardinality()) > max){
							max = card;
							newClique = nodesp;
						}
					}
				}
				if(max > 0){
					augmented = true;
					ExtensionClique ext = new ExtensionClique(this, newClique);
					newCliques[c] = new CompoundClique(this,
							ext,
							cliques[c]);
					assert((cliques[c].getSize()+max) == newCliques[c].getSize());
					assert(newCliques[c].isClique());
				}
				else
					newCliques[c] = cliques[c];
			}
			cliques = newCliques;
			Arrays.sort(cliques, Clique.DECREASING_ORDER);
			if(filter == Filter.POST) filterDominated();
		}
	}
	
	public void oneCliqueFirst(boolean withRooms, Filter filter){
		initElementaryCliques(withRooms);
		
		if(filter == Filter.PRE) filterDominated();
		
		for(int c = 0; c < cliques.length; c++) {
			boolean augmented = true;
			while(augmented){
				augmented = false;

				BitSet neighbours = cliques[c].globalNeighbourhood(),
					newClique = null;//The set of nodes to add to augment the clique.

				int max = -1;
				for(int cp = 0; cp < cliques.length; cp++){
					if(c != cp){
						BitSet nodesp = cliques[cp].retrieveNodes();
						nodesp.and(neighbours);
						int card;
						if((card=nodesp.cardinality()) > max){
							max = card;
							newClique = nodesp;
						}
					}
				}
				if(max > 0){
					augmented = true;
					ExtensionClique ext = new ExtensionClique(this, newClique);
					CompoundClique augmentedClique = new CompoundClique(this,
							ext,
							cliques[c]);
					cliques[c] = augmentedClique;
					assert(cliques[c].isClique());					
				}
			}
		}
		Arrays.sort(cliques, Clique.DECREASING_ORDER);
		if(filter == Filter.POST) filterDominated();
	}
	
	public void printCliques(){
		int sum = 0;for(CompoundClique clique:cliques)sum += clique.getSize();
		System.out.println(sum+" "+Arrays.toString(cliques));
	}
	
	public static void main(String[] args) throws Exception{
		for(int i=1;i<=8;i++){
			TimetablingInstance ti = new InstanceReader(new FileInputStream("./data/Track2/comp-2007-2-"+i+".tim")).getInstance();
			MaxCliqueFinder mc = new MaxCliqueFinder(ti);
			mc.findCliques(true);
//			CompoundClique[] cliques = mc.getCliques();
//			for(CompoundClique clique : cliques){
//				BitSet nodes = clique.retrieveNodes();
//				for(int n=nodes.nextSetBit(0);n>-1;n=nodes.nextSetBit(n+1)){
//					
//				}
//			}
			mc.printCliques();
			System.out.println();
		}
	}
	
	
}
