package instance.colouring;

import java.util.Random;

import instance.ColourableConflictGraph;
import util.IntSet;

public abstract class GraphColouring {
	
	protected ColourableConflictGraph graph;
	
	protected IntSet unassignedNodes;
	
	protected Random random = new Random();
	
	/**
	 * The number of times each colour has already
	 * been assigned to a node.
	 */
	protected int[] occurrences;
	
	public GraphColouring(ColourableConflictGraph graph){
		this.graph = graph;
		unassignedNodes = graph.getUnassignedNodes();
		occurrences = new int[graph.colourList(0).capacity()];
	}
	
	public void setRandom(Random r){
		random = r;
	}
	
	/**
	 * Tries to find a valid colouration of the colourable graph
	 * or returns false if it couldn't.
	 * 
	 * @return <code>false</code> if failed: no compatible colour
	 * will number of occurrences less than max found.
	 */
	public boolean greedyList(){
		
		while(!unassignedNodes.isEmpty()){
			int node = nextNode();
			if(graph.colourList(node).size() > 0){
				int colour = selectColour(node);
				graph.setToColour(node, colour);
				occurrences[colour]++;
				pruneNeighbours(node);
				unassignedNodes.remove(node);
			}	
			else
				return false;
		}
		return true;
	}
	
	protected void pruneNeighbours(int node){
		assert(graph.colourList(node).size() == 1);
		int colour = graph.colourList(node).first();
		for(int u=graph.firstNeighbour(node);u>-1;u=graph.nextNeighbour(node, u))
			graph.colourList(u).remove(colour);
		
	}
	
	protected abstract int nextNode();
	
	protected int selectColour(int node){
		IntSet colours = graph.colourList(node);
		int min = Integer.MAX_VALUE;
		IntSet minColours = null;
		for(int c=colours.first();c>-1;c=colours.next(c)){
			int occ = occurrences[c];
			if(occ < min){
				min = occ;
				minColours = new IntSet(occurrences.length);
				minColours.setRandom(random);
			}
			if(occ == min)
				minColours.add(c);
		}
		assert(minColours != null);
		//Fail if min > max occurrences?
		return minColours.randomValue();
	}
}
