package localsearch;

import tools.InstanceReader;

import java.util.*;
import java.io.*;

import instance.TimetablingInstance;
import instance.Precedence;
import choco.integer.IntDomainVar;

public class HardSolutionSpace {

	//random generator
	public Random rand;

	//******** Instance data ************//

	public TimetablingInstance ins;

	//number of TimeSlot, Days, Hours, Rooms, Students, Events
	public int T, D, H, R, S, E;

	// Dim: Events * Rooms
	// indicate whether a room is feasible for an event (0 ou 1)
	// this is not a boolean to perform cost operations
	public int[][] roomFeas;

	// Dim: Events * Timeslot
	// indicate whether a timeslot is feasible for an event (0 ou 1)
	// this is not a boolean to perform cost operations
	public int[][] timeslotFeas;

	// Dim: Events * Events
	// indicate whether two events share a student
	public int[][] conflicts;

	// Dim: Events * (at most 45)
	// greaterthan for each event e it gives the list of events e' such that e > e'
	// lessthan for each event e it gives the list of events e' such that e < e'
	public int[][] greaterthan;
	public int[][] lessthan;

	// Dim: Events * Events
	// prel[e1][e2] is true e1 < e2
	public boolean[][] before;

	//******** Solution representation ************//
	// Dim: TimeSlot * Rooms
	// The assignment matrix encoding the current solution
	public int[][] matrix;

	//the number of events in each timeslot
	public int[] nbept;

	//Dim: Events
	//the current room of each event
	public int[] rooms;

	//Dim: Events
	//the current timeslot of each event
	public int[] timeslot;

	// Dim: TimeSlot * Rooms
	// The "list" of events per timeslots (no holes)
	public int[][] listevts;


	//The list of free slots in the timetabling
	public List<Place> freeSlots;
	public Place[][] freeMatrixSlots;

	//******** Cost computation ************//

	//Global violation of the hard constraints
	public int hardObj;

	//Dim Events
	//the cost of an event in its current timeslot regarding hard constraints:
	//nbconflict + (isTimeSlotAvailable: 0 or 1) + (is the room big enough: 0 or 1)
	public int[] eventcost;

	//Dim Events
	//the cost of an event with its neigbourhs
	public int[] conflictcost;


    //default constructor
    public HardSolutionSpace() {
        
    }

    public HardSolutionSpace(TimetablingInstance ins, int seed) {
		this.ins = ins;
		this.rand = new Random(seed);
		this.T = 45;
		this.D = 5;
		this.H = 9;
		this.R = ins.getNRooms();
		this.S = ins.getNStudents();
		this.E = ins.getNEvents();
		initInstanceData();
	}

	/**
	 * return true if there is a precedence relationship between ev1 and
	 * ev2
	 *
	 * @param ev1
	 * @param ev2
	 */
	public boolean isThereAPrecRelationShip(int ev1, int ev2) {
		return before[ev1][ev2] || before[ev2][ev1];
	}

	/**
	 * @param ev the event
	 * @return true if ev is not involved in any precedences
	 */
	public boolean isPrecedencFree(int ev) {
		return greaterthan[ev].length == 0 && lessthan[ev].length == 0;
	}

	/**
	 * @param t the timeslot
	 * @return a bitset representing the students involved in timeslot
	 */
	public BitSet studentsAttendingTimeslot(int t) {
		BitSet b = new BitSet(S);
		for (int i = 0; i < nbept[t]; i++) {
			int evt = listevts[t][i];
			b.or(ins.studentsAttendingEvent(evt));
		}
		return b;
	}

	/**
	 * @param t1 the timeslot
	 * @param t2 the timeslot
	 * @return a bitset representing the students attending t1 but not t2
	 */
	public BitSet studentsAttendingB1butB2(BitSet t1, BitSet t2) {
		BitSet b = (BitSet) t1.clone();
		b.or(t1);
		b.xor(t2);
		b.and(t1);
		return b;
	}

	/**
	 * @param s the student
	 * @param t the timeslot
	 * @return the event attending by s in timeslot t
	 */
	public int eventOfStudentInTimeslot(int s, int t) {
		for (int i = 0; i < nbept[t]; i++) {
			int evt = listevts[t][i];
			if (ins.studentAttendsEvent(s, evt)) return evt;
		}
		assert (false);
		return -1;
	}

	//******** initialization of the data structure ************//

	public void initSolutionSpace() {
		initSolutionRepresentation();
	}

	public void initInstanceData() {
		roomFeas = new int[E][R];
		timeslotFeas = new int[E][T];
		conflicts = new int[E][E];
		before = new boolean[E][E];
		greaterthan = new int[E][]; //add a fake -1 at the end to avoid testing the end of the table
		lessthan = new int[E][]; //add a fake -1 at the end to avoid testing the end of the table
		for (int i = 0; i < E; i++) {
			greaterthan[i] = new int[ins.getNInvprec(i)];
			lessthan[i] = new int[ins.getNPrec(i)];
			for (int j = 0; j < R; j++) {
				roomFeas[i][j] = (ins.isRoomAvailableForEvent(i, j)) ? 0 : 1;
			}
			for (int j = 0; j < T; j++) {
				timeslotFeas[i][j] = ins.eventIsCompatibleWithTimeslot(i, j) ? 0 : 1;
			}
			for (int j = i + 1; j < E; j++) {
				conflicts[i][j] = (ins.eventsShareStudents(i, j)) ? 1 : 0;
				conflicts[j][i] = conflicts[i][j];
			}
			Arrays.fill(greaterthan[i], -1);
			Arrays.fill(lessthan[i], -1);
		}
		for (Precedence p : ins.eventPrecedences()) {
			//e1 < e2
			int e1 = p.firstEvent;
			int e2 = p.secondEvent;
			before[e1][e2] = true;
			addEvtInPrecTable(greaterthan[e2], e1);
			addEvtInPrecTable(lessthan[e1], e2);
		}
	}

	public void addEvtInPrecTable(int[] ptable, int evt) {
		int i = 0;
		while (ptable[i] != -1) {
			i++;
		}
		ptable[i] = evt;
	}

	public void initSolutionRepresentation() {
		matrix = new int[T][R];
		listevts = new int[T][R];
		freeMatrixSlots = new Place[T][R];
		rooms = new int[E];
		timeslot = new int[E];
		hardObj = 0;
		eventcost = new int[E];
		conflictcost = new int[E];
		nbept = new int[T];
		freeSlots = new LinkedList<Place>();
		for (int i = 0; i < T; i++) {
			Arrays.fill(listevts[i], -1);
		}
	}

	//******** Basic operation on events *********************//

	/**
	 * Remove event evt from its current location while
	 * maintaining all data structures and costs (except precedence cost)
	 *
	 * @param evt the event removed
	 */
	public void removeEvent(int evt) {
		int e1t = timeslot[evt];
		int e1r = rooms[evt];
		matrix[e1t][e1r] = -1;
		rooms[evt] = -1;
		timeslot[evt] = -1;
		hardObj -= eventcost[evt];
		int pevt = -1;
		for (int i = 0; i < R; i++) { //reduce the cost of the neighbours
			int ei = matrix[e1t][i];
			if (ei != -1) {
				eventcost[ei] -= conflicts[ei][evt];
				conflictcost[ei] -= conflicts[ei][evt];
			}
			if (listevts[e1t][i] == evt) pevt = i;
		}
		eventcost[evt] = 0;
		conflictcost[evt] = 0;
		maintainListEvt(evt, e1t, pevt);
		nbept[e1t]--;
	}

	/**
	 * maintain the listevts data structure due to removal of evt
	 *
	 * @param evt  the event to be removed
	 * @param e1t  its timeslot
	 * @param pevt its position in the list of events data structure
	 */
	public void maintainListEvt(int evt, int e1t, int pevt) {
		int evttoswap = listevts[e1t][nbept[e1t] - 1];
		listevts[e1t][nbept[e1t] - 1] = -1;
		if (evttoswap != evt)
			listevts[e1t][pevt] = evttoswap;
	}

	/**
	 * Place event evt in the EMPTY place corresponding to the given
	 * timeslot and room while maintaining all data structures and
	 * cost (except precedence cost).
	 *
	 * @param evt     the event to add
	 * @param timeslo the timeslot where we want to put the event
	 * @param room    the room where the event has to go
	 */
	public void addEvent(int evt, int timeslo, int room) {
		assert (matrix[timeslo][room] == -1);
		int evtcost = roomFeas[evt][room] + timeslotFeas[evt][timeslo];
		int conflict = 0;
		for (int i = 0; i < nbept[timeslo]; i++) {
			int ei = listevts[timeslo][i];
			conflict += conflicts[ei][evt];
			conflictcost[ei] += conflicts[ei][evt];
			eventcost[ei] += conflicts[ei][evt];
		}
		eventcost[evt] = evtcost + conflict;
		conflictcost[evt] = conflict;
		hardObj += eventcost[evt];
		matrix[timeslo][room] = evt;
		timeslot[evt] = timeslo;
		rooms[evt] = room;
		listevts[timeslo][nbept[timeslo]] = evt;
		nbept[timeslo]++;
		//System.out.println("add event " + evt + " in timeslot " + timeslo + " and room " + room);
	}

	/**
	 * Remove an event knowing the fact that it will be added back into
	 * the same timeslot !!
	 *
	 * @param evt the event removed
	 */
	public void fastRemoveEvent(int evt) {
		int e1t = timeslot[evt];
		int e1r = rooms[evt];
		matrix[e1t][e1r] = -1;
		rooms[evt] = -1;
		hardObj -= roomFeas[evt][e1r];
		eventcost[evt] -= roomFeas[evt][e1r];
	}


	/**
	 * add an event into the same timeslot that he was before
	 * but a different room
	 *
	 * @param evt  the event added
	 * @param room the new room for the evt
	 */
	public void fastAddEvent(int evt, int room) {
		int e1t = timeslot[evt];
		matrix[e1t][room] = evt;
		rooms[evt] = room;
		hardObj += roomFeas[evt][room];
		eventcost[evt] += roomFeas[evt][room];
	}

	/**
	 * move an event without maintaining the precedence cost
	 *
	 * @param evt1 the two event to move with its corresponding previous and
	 *             new precedence cost.
	 */
	public void fastMoveTimeslot(int evt1, int e2t) {
		int e1t = timeslot[evt1];
		timeslot[evt1] = e2t;
		matrix[e1t][rooms[evt1]] = -1;
		matrix[e2t][rooms[evt1]] = evt1;
		eventcost[evt1] -= timeslotFeas[evt1][e1t];
		eventcost[evt1] += timeslotFeas[evt1][e2t];
		hardObj -= timeslotFeas[evt1][e1t];
		hardObj += timeslotFeas[evt1][e2t];
	}

	/**
	 * swap two evts without maintaining the precedence cost
	 *
	 * @param evt1,evt2 the two events to swap.
	 */
	public void fastSwitchTimeslot(int evt1, int evt2) {
		int e1t = timeslot[evt1];
		int e2t = timeslot[evt2];
		timeslot[evt2] = e1t;
		timeslot[evt1] = e2t;
		matrix[e1t][rooms[evt2]] = evt2;
		matrix[e2t][rooms[evt1]] = evt1;
		eventcost[evt1] -= timeslotFeas[evt1][e1t];
		eventcost[evt1] += timeslotFeas[evt1][e2t];
		eventcost[evt2] -= timeslotFeas[evt2][e2t];
		eventcost[evt2] += timeslotFeas[evt2][e1t];
		hardObj += timeslotFeas[evt1][e2t] - timeslotFeas[evt1][e1t];
		hardObj += timeslotFeas[evt2][e1t] - timeslotFeas[evt2][e2t];
	}

	//******** Precedence managment *********************//

	/**
	 * maintain the precedence cost of moving evt from timeslot t1 to t2
	 * evt2 is an event such that we don't want to modify its cost during the process.
	 * (because it will be updated later on. That's a special condition for swap)
	 */
	public void maintainPrecedenceCostFromMove(int evt, int evt2, int t1, int t2) {
		for (int i = 0; i < greaterthan[evt].length; i++) {
			if (timeslot[greaterthan[evt][i]] >= t2) {
				eventcost[evt]++;
				hardObj++;
			}
		}
		for (int i = 0; i < lessthan[evt].length; i++) {
			int e2 = lessthan[evt][i];// we should have e2 > evt
			if (e2 != evt2) {
				//if the precedences was satisfied before but is violated now
				if (timeslot[e2] > t1 && timeslot[e2] <= t2) {
					eventcost[e2]++;
					hardObj++;
				}
				//if the precedences was violated before but satisfied now
				if (timeslot[e2] <= t1 && timeslot[e2] > t2) {
					eventcost[e2]--;
					hardObj--;
				}

			}
		}
	}

	/**
	 * maintain the precedence cost of moving evt from timeslot t1 to t2 without
	 * changing the cost of events placed on t1 or t2 (that's a special condition
	 * for swapping timeslots)
	 */
	public void maintainPrecedenceCostFromMove(int evt, int t1, int t2) {
		for (int i = 0; i < greaterthan[evt].length; i++) {
			int e2 = greaterthan[evt][i]; //evt > e2
			int ts = timeslot[e2];
			int ts2 = timeslot[e2];
			if (ts == t1) {
				ts = t2;
			} else if (ts == t2) {
				ts = t1;
			}
			if (ts < t1 && ts2 >= t2) {
				eventcost[evt]++;
				hardObj++;
			}
			if (ts >= t1 && ts2 < t2) {
				eventcost[evt]--;
				hardObj--;
			}
		}
		for (int i = 0; i < lessthan[evt].length; i++) {
			int e2 = lessthan[evt][i];// we should have e2 > evt
			int ts2 = timeslot[e2];
			if (ts2 != t2 && ts2 != t1) {
				//if the precedences was satisfied before but is violated now
				if (timeslot[e2] > t1 && timeslot[e2] <= t2) {
					eventcost[e2]++;
					hardObj++;
				}
				//if the precedences was violated before but satisfied now
				if (timeslot[e2] <= t1 && timeslot[e2] > t2) {
					eventcost[e2]--;
					hardObj--;
				}
			}
			//otherwise nothing to change
		}
	}

	/**
	 * evaluate the precedence cost of moving evt from timeslot t1 to t2
	 * It returns the new cost projection of evt regarding precedences AND
	 * the NEW cost of other violated precedences of other events with evt
	 */
	public int evalPrecedenceCostFromMove(int evt, int evt2, int t1, int t2) {
		int preccost = 0;
		for (int i = 0; i < greaterthan[evt].length; i++) {
			if (timeslot[greaterthan[evt][i]] >= t2) {
				preccost++;
			}
		}
		for (int i = 0; i < lessthan[evt].length; i++) {
			int e2 = lessthan[evt][i];// we should have e2 > evt
			if (evt2 != e2) {
				//if the precedences was satisfied before but is violated now
				if (timeslot[e2] > t1 && timeslot[e2] <= t2) {
					preccost++;
				}
				//if the precedences was violated before but satisfied now
				if (timeslot[e2] <= t1 && timeslot[e2] > t2) {
					preccost--;
				}
			}
		}
		return preccost;
	}

	/**
	 * Return the number of precedences violated and linked to evt (without counting
	 * the precedences violated from events placed on timeslot t1 and t2)
	 */
	public int evalPrecCost(int evt, int ts1, int ts2) {
		int preccost = 0;
		int t1 = timeslot[evt];
		for (int i = 0; i < greaterthan[evt].length; i++) { //we should evt > greaterthan[evt][cpt]
			if (timeslot[greaterthan[evt][i]] >= t1) {
				preccost++;
			}
		}
		for (int i = 0; i < lessthan[evt].length; i++) {
			int e2 = lessthan[evt][i];// we should have e2 > evt
			int t2 = timeslot[e2];
			if (t2 != ts1 && t2 != ts2 && timeslot[e2] <= t1)
				preccost++;
		}
		return preccost;
	}

	//******** Maintain free place structure ******************//

	/**
	 * t1,r1 is a free place, t2,r2 is occupied by evt. evt is an
	 * event moving from t2,r2 to r1,t1
	 */
	public void switchFreePlace(int t1, int r1, int t2, int r2) {
		assert (freeMatrixSlots[t1][r1] != null);
		Place p = freeMatrixSlots[t1][r1];
		freeMatrixSlots[t1][r1] = null;
		p.t = t2;
		p.r = r2;
		freeMatrixSlots[t2][r2] = p;
	}

	public void addFreePlace(int t1, int r1) {
		assert (matrix[t1][r1] == -1);
		assert (freeMatrixSlots[t1][r1] == null);
		Place p = new Place(t1, r1);
		freeMatrixSlots[t1][r1] = p;
		freeSlots.add(p);
	}

	public void removeFreePlace(int t1, int r1) {
		assert (matrix[t1][r1] != -1);
		assert (freeMatrixSlots[t1][r1] != null);
		Place p = freeMatrixSlots[t1][r1];
		freeSlots.remove(p);
		freeMatrixSlots[t1][r1] = null;
	}

	/**
	 * repare the free Place data structure on a given timeslot when the rooms
	 * of events already present on this timeslot have changed.
	 *
	 * @param timeslot the timeslot where free places have to be repared
	 *                 todo : to improve !
	 */
	public void maintainFreePlacesOnTimeslot(int timeslot) {
		for (int i = 0; i < R; i++) {
			if (matrix[timeslot][i] == -1 &&
					freeMatrixSlots[timeslot][i] == null) {
				addFreePlace(timeslot, i);
			} else if (matrix[timeslot][i] != -1 &&
					freeMatrixSlots[timeslot][i] != null) {
				removeFreePlace(timeslot, i);
			}
		}
	}

	//******** initialization of the solution ****************//

	public void restart() {
		for (int i = 0; i < T; i++) {
			Arrays.fill(matrix[i], -1);
			Arrays.fill(listevts[i], -1);
		}
		Arrays.fill(timeslot, -1);
		Arrays.fill(rooms, -1);
		Arrays.fill(nbept, 0);
		Arrays.fill(eventcost, 0);
		Arrays.fill(conflictcost, 0);
		freeSlots.clear();
		hardObj = 0;
	}

	/**
	 * Randomly place each event in the assignment matrix
	 */
	public void randomInit() {
		restart();
		for (int i = 0; i < E; i++) {
			int position = rand.nextInt((T * R - 1) - i);
			int cpt = 0;
			boolean mark = false;
			for (int j = 0; j < T && !mark; j++) {
				for (int k = 0; k < R && !mark; k++) {
					if (matrix[j][k] == -1) {
						if (cpt == position) {
							addEvent(i, j, k);
							mark = true;
						} else cpt++;
					}
				}
			}
		}
		initMissingCosts();
	}

	/**
	 * Place each event in the assignment matrix according to a choco
	 * solution
	 */
	public void initFromChocoSol(IntDomainVar[] chocoevents) {
		restart();
		int[] roomcpt = new int[T];
		for (int i = 0; i < E; i++) {
			assert (chocoevents[i].isInstantiated());
			int timeslot = chocoevents[i].getVal() - 1;
			int room = roomcpt[timeslot];
			roomcpt[timeslot]++;
			assert (matrix[timeslot][room] == -1);
			addEvent(i, timeslot, room);
		}
		initMissingCosts();
	}

    /**
	 * Place each event in the assignment matrix according to a timeslot
	 * solution
	 */
	public void initFromTimeslotSol(int[] timeslots) {
		restart();
		int[] roomcpt = new int[T];
		for (int i = 0; i < E; i++) {
			int timeslot = timeslots[i];
			int room = roomcpt[timeslot];
			roomcpt[timeslot]++;
			assert (matrix[timeslot][room] == -1);
			addEvent(i, timeslot, room);
		}
		initMissingCosts();
	}

    /**
	 * Place each event in the assignment matrix according to a choco
	 * solution
	 */
	public void initFromGivenSol(int[][] mat) {
		restart();
		for (int i = 0; i < T; i++) {
			for (int j = 0; j < R; j++) {
				if (mat[i][j] != -1) {
					addEvent(mat[i][j], i, j);
				}
			}
		}
		initMissingCosts();
	}


	/**
	 * Costs involving greaterthan have to be computed when all position of the events
	 * are known ! Thus a separate method to be called after everything has been
	 * reassigned
	 */
	public void addGreaterInCost(int evt) {
		for (int i = 0; i < greaterthan[evt].length; i++) {
			if (timeslot[greaterthan[evt][i]] >= timeslot[evt]) {
				eventcost[evt]++;
				hardObj++;
			}
		}
	}

	/**
	 * init the free slots data structure
	 */
	public void initMissingCosts() {
		for (int i = 0; i < E; i++) {
			addGreaterInCost(i);
		}
		for (int i = 0; i < T; i++) {
			for (int j = 0; j < R; j++) {
				if (matrix[i][j] == -1) {
					Place p = new Place(i, j);
					freeSlots.add(p);
					freeMatrixSlots[i][j] = p;
				} else {
					freeMatrixSlots[i][j] = null;
				}
			}
		}
	}

	//******** print solution ********************************//

	public void printSolution() {
		for (int i = 0; i < T; i++) {
			System.out.println(Arrays.toString(matrix[i]));
		}
		System.out.println("violation " + hardObj);
		System.out.println("details of violations ");
		System.out.println(Arrays.toString(eventcost));
		System.out.println(Arrays.toString(conflictcost));
	}

	public void printCost() {
		int c = 0;
		for (int i = 0; i < E; i++) {
			c += conflictcost[i];
		}
		System.out.println("violation " + hardObj);
		System.out.println("cost conflict: " + c / 2);
		System.out.println("cost room: " + (hardObj - c / 2));
	}

	//******** simple tests **********************************//

	public static void main(String[] args) {
		InstanceReader reader = null;
		try {
			reader = new InstanceReader(new FileInputStream("data/Track2/comp-2007-2-2.tim.txt"));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		TimetablingInstance ins = reader.getInstance();
		HardSolutionSpace sp = new HardSolutionSpace(ins, 1);
		sp.randomInit();
		sp.printSolution();
		System.out.println("fin");
	}

	//*************** Assert current data struture*************//

	public boolean assertAllSoftCosts() {
		return true;
	}


	public boolean assertFeasible() {
		return true;
	}

	public boolean assertHardCostsWithoutFreeSpaces() {
		return assertHardCosts(false);
	}

	public boolean assertAllHardCosts() {
		return assertHardCosts(true);
	}


	public boolean assertHardCosts(boolean freeSpaceData) {
		int asserthardObj = 0;
		int totalconfcost = 0;
		int[] cardEvtsPerTimeslot = new int[T];
		for (int e = 0; e < E; e++) {
			int er = rooms[e];
			int et = timeslot[e];
			cardEvtsPerTimeslot[et]++;
			assert (er != -1);
			assert (et != -1);
			int coste = roomFeas[e][er] + timeslotFeas[e][et];
			int conflicte = 0;
			for (int r = 0; r < R; r++) {
				if (er != r) {
					int e2 = matrix[et][r];
					if (e2 != -1) {
						conflicte += conflicts[e][e2];
					}
				}
			}
			for (int i = 0; i < greaterthan[e].length; i++) {
				if (timeslot[greaterthan[e][i]] >= et) {
					coste++;
				}
			}
			totalconfcost += conflicte;
			coste += conflicte;
			if (coste != eventcost[e])
				System.out.println("event " + e + " " + coste + " != " + eventcost[e]);
			assert (coste == eventcost[e]);
			assert (conflicte == conflictcost[e]);
			asserthardObj += coste;
		}
		for (int i = 0; i < T; i++) {
			assert (nbept[i] == cardEvtsPerTimeslot[i]);
		}
		asserthardObj -= totalconfcost / 2;
		//System.out.println(asserthardObj + " " + hardObj);
		assert (asserthardObj == hardObj);
		//System.out.println(" " + costtf);
		for (int i = 0; i < E; i++) {
			int eit = timeslot[i];
			boolean isInList = false;
			for (int j = 0; j < nbept[eit] && !isInList; j++) {
				if (listevts[eit][j] == i) isInList = true;
			}
			if (!isInList) System.out.println("evt " + i + " in " + eit);
			assert (isInList);
		}
		if (freeSpaceData && hardObj != 0) {
			for (Iterator<Place> it = freeSlots.iterator(); it.hasNext();) {
				Place place = it.next();
				assert (matrix[place.t][place.r] == -1);
				assert (freeMatrixSlots[place.t][place.r].equals(place));
			}
			int nbFreeSlot = 0;
			for (int i = 0; i < T; i++) {
				for (int j = 0; j < R; j++) {
					if (matrix[i][j] == -1) nbFreeSlot++;
				}
			}
			assert (nbFreeSlot == freeSlots.size());
		}
		return true;
	}

}
