package localsearch.graphalgos;

import localsearch.SolutionSpace;
import localsearch.graphalgos.matchings.Matching;
import localsearch.graphalgos.matchings.Node;

import java.util.Arrays;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * A class to hide the solving of the matching
 */
public class MatchingInterface {

	public SolutionSpace space;

	//A table of events to reallocate
	public int[] evts;

    //A table of events left outside the matching but that were in
    //the timeslot and need to be arbitrarly put in the rooms 
    public int[] leftEvts;

    //The timeslot to reorganize
	public int timeslot;
	public int nbEvt;

	// the Hungarish algorithm O(n^3)
	public Hongrois HungarishAlgo;

	// A matching algorithm O(nm) and incremental
	public Matching matchingAlgo;

	public MatchingInterface(SolutionSpace space) {
		this.space = space;
		this.rcosts = new int[space.R][space.R];
		this.matchingAlgo = new Matching(space);
    }

	/**
	 * solve and return true if a matching exists
	 */
	public boolean solve() {
		return solveLazyMatching();
	}
	
	/**
	 * @return the matching found
	 */
	public int[] getMatching() {
		return matchingAlgo.getBestMatchingFound();
	}
	
	public int getMatchingCost() {
		return matchingAlgo.getMatchingCost();
	}

	/**
	 * @return the matching found
	 */
	public MatchingSolution getMatchingSolution() {
		return new MatchingSolution(getMatching(), evts);
	}

    protected MatchingSolution[][] matchsol;
    
    /**
	 * @return the matching found
	 */
	public MatchingSolution getMatchingSolutionWithLeftEvts(MatchingSolution msol) {
        int[] newmatch = new int[evts.length + leftEvts.length];
        int[] newevts = new int[newmatch.length];
		int[] match = getMatching();
        //concatenate the evts (match + left evts)
        System.arraycopy(evts, 0, newevts, 0, evts.length);
        System.arraycopy(leftEvts, 0, newevts, evts.length, leftEvts.length);
        // use the known matching
        System.arraycopy(match, 0, newmatch, 0, match.length);
        //complete with arbitrary room
        Node[] right = matchingAlgo.getRight();
        int cpt = match.length;
        for (int i = 0; i < right.length && cpt < newmatch.length; i++) {
			if (right[i].getMatchNode() == -1) {
				newmatch[cpt] = i;
                cpt ++;
            }
		}
        msol.set(newmatch,newevts);
        return msol;//new MatchingSolution(newmatch, newevts);
    }

    //*******************************************************//
	//********** initialize the table of events *************//
	//*******************************************************//

	/**
	 * set timeslot i to be checked by matching
	 */
	public void setTimeSlot(int i) {
		timeslot = i;
		nbEvt = space.nbept[i];
		evts = new int[nbEvt];
		System.arraycopy(space.listevts[i], 0, evts, 0, nbEvt);
	}

	/**
	 * set timeslot i and event evt to be checked by the matching
	 * used by translation
	 */
	public int setTimeSlot(int i, int evt) {
		assert (i <= space.T);
		assert (i >= 0);
		assert (space.timeslot[evt] != i);
		timeslot = i;
		nbEvt = space.nbept[i] + 1;
		evts = new int[nbEvt];
		System.arraycopy(space.listevts[i], 0, evts, 0, nbEvt - 1);
		evts[nbEvt - 1] = evt;
		return nbEvt - 1;
	}

	/**
	 * set timeslot i, replacing event e1 by e2
	 * to be checked by the matching
	 * used by swapping
	 *
	 * @return index of e2 within the matching
	 */
	public int setTimeSlot(int i, int e1, int e2) {
		assert (i <= space.T);
		assert (i >= 0);
		assert (space.timeslot[e2] != i);
		int idxtoret = -1;
		timeslot = i;
		nbEvt = space.nbept[i];
		evts = new int[nbEvt];
		for (int j = 0; j < nbEvt; j++) {
			int evt = space.listevts[i][j];
			if (evt != e1)
				evts[j] = evt;
			else {
				evts[j] = e2;
				idxtoret = j;
			}
		}
		return idxtoret;
	}

    /**
	 * set the events as the ones that have a valid room in timeslot i,
     * - by adding e1 if e1 != -1 and removing e2 if e2 != -1 
	 */
	public void setPartialTimeSlot(int i, int e1, int e2) {
		assert (i <= space.T);
		assert (i >= 0);
		timeslot = i;
        ArrayList<Integer> mevts = new ArrayList<Integer>();
        ArrayList<Integer> levts = new ArrayList<Integer>();
        for (int j = 0; j < space.nbept[i]; j++) {
           int evt = space.listevts[i][j];
           if (evt != e2 && e1 != evt) {
               if (space.roomFeas[evt][space.rooms[evt]] == 0) {
                   mevts.add(evt);
               } else {
                   levts.add(evt);                   
               }
           }
        }
        if (e1 != -1) {
            mevts.add(e1);
        }
        evts = new int[mevts.size()];
        nbEvt = evts.length;
        leftEvts = new int[levts.size()];
        for (int j = 0; j < evts.length; j++)
            evts[j] = mevts.get(j);
        for (int j = 0; j < leftEvts.length; j++)
            leftEvts[j] = levts.get(j);
    }

    //*******************************************//
	//********** solve by matching *************//
	//*******************************************//

	public boolean solveLazyMatching() {
		matchingAlgo.init(evts, evts.length);
		return matchingAlgo.computeLazyPerfectMatching();
	}

	public boolean solveMatching() {
		matchingAlgo.init(evts, evts.length);
		return matchingAlgo.computePerfectMatching();
	}

	//*******************************************//
	//********** solve by hungarian *************//
	//*******************************************//

	//the cost of allocating event i in room j
	public int[][] rcosts;

	/**
	 * Fill the cost matrix
	 */
	public void computeMatrixCost() {
		for (int i = 0; i < nbEvt; i++) {
			for (int j = 0; j < space.R; j++) {
				int evt = evts[i];
				rcosts[i][j] = space.roomFeas[evt][j];
			}
		}
		for (int i = nbEvt; i < space.R; i++) {
			Arrays.fill(rcosts[i], 0);
		}
	}

	public boolean solveHungarian() {
		computeMatrixCost();
		HungarishAlgo = new Hongrois(rcosts,space); //avoid building the hungarian ?
		//printCostMatrix();
		HungarishAlgo.solve();
		int cost = 0;
		for (int i = 0; i < nbEvt; i++) {
			cost += rcosts[i][HungarishAlgo.getAffectedTask(i)];
		}
		return cost == 0;
	}


	//************************************************//
	//*********** fast matching **********************//
	//************************************************//

	public int room;

	public boolean fastSolve() {
		int evt = evts[nbEvt - 1];
		for (int r = 0; r < space.R; r++) {
			if (space.roomFeas[evt][r] == 0 &&
					space.matrix[timeslot][r] == -1) {
				room = r;
				return true;
			}
		}
		return false;
	}

	public int[] getFastMatching() {
		int[] match = new int[nbEvt];
		for (int i = 0; i < nbEvt - 1; i++) {
			match[i] = space.rooms[evts[i]];
		}
		match[nbEvt - 1] = room;
		return match;
	}

}
