package localsearch.graphalgos.matchings;

import localsearch.SolutionSpace;
import localsearch.solvers.complete.HardSolver;

import java.util.*;

/**
 * Implement a simple matching algorithm for the bipartite graph
 * events - rooms in O(nm) based on a dfs for finding augmenting paths.
 */
public class Matching {

	protected SolutionSpace space;

	//the node corresponding to a given event
	protected static Node[] events;

	//Encoding the bipartite graph
	private ArrayList<Node> left;  //events
	private Node[] right; //rooms
	protected int cmatchsize;

	protected Stack<Node> unmatchedNodes;

	public static void initNodeEvents(SolutionSpace space) {
		events = new Node[space.E];
		for (int i = 0; i < space.E; i++) {
			events[i] = new Node(i, -1, space);
			events[i].createLeftSucc(space);
		}
	}


	public Matching(SolutionSpace space) {
		this.space = space;
		setRight(new Node[space.R]);
		for (int i = 0; i < space.R; i++) {
			getRight()[i] = new Node(i, i, space);
		}
		markright = new BitSet(space.R);
		markleft = new BitSet(space.R);
		path = new Stack<Node>();
		unmatchedNodes = new Stack<Node>();
		savedMatching = new int[space.E];
		Arrays.fill(savedMatching, -1);
	}

	public int getMatchingCost() {
		return getLeft().size() - cmatchsize;
	}

	public void resetRooms() {
		for (int i = 0; i < getRight().length; i++) {
			getRight()[i].succ.empty();
			getRight()[i].setMatchNode(-1);
		}
	}


	/**
	 * build the bipartite graph for a given timeslot
	 *
	 * @param timeslot the timeslot to reoptimize
	 */
	public void init(int timeslot) {
		init(space.listevts[timeslot], space.nbept[timeslot]);
	}

	/**
	 * build the bipartite graph for a set of events
	 *
	 * @param evts a table of events to be assigned
	 */
	public void init(int[] evts, int size) {
		reinit();
		for (int i = 0; i < size; i++) {
			Node cnode = events[evts[i]];
			getLeft().add(cnode);
			if (!cnode.initLeftNode(i, space.rooms[evts[i]], getRight())) {
				unmatchedNodes.push(cnode);
			} else cmatchsize++;
		}
	}

	public void reinit() {
		resetRooms();
		cmatchsize = 0;
		unmatchedNodes.clear();
		setLeft(new ArrayList<Node>());
	}

	/**
	 * Compute a maximum matching and returns true if a perfect one has been
	 * found, false otherwise.
	 *
	 * @return true if a perfect matching exists
	 */
	public boolean computeLazyPerfectMatching() {
		while (!unmatchedNodes.empty()) {
			Node v = unmatchedNodes.pop();
			if (findAugmentingPath(v)) {
				improveMatchingAlongPath();
			} else {
				unmatchedNodes.push(v);
				return false;
			}
		}
		return true;
	}


	/**
	 * Compute a maximum matching and returns true if a perfect one has been
	 * found, false otherwise.
	 *
	 * @return true if a perfect matching exists
	 */
	public boolean computePerfectMatching() {
		Stack<Node> difficultNodes = new Stack<Node>();
		while (!unmatchedNodes.empty()) {
			Node v = unmatchedNodes.pop();
			if (findAugmentingPath(v)) {
				improveMatchingAlongPath();
				cmatchsize++;
			} else {
				difficultNodes.push(v);
			}
		}
		if (cmatchsize != getLeft().size())
			completeMatchingWithViolations(difficultNodes);
		return cmatchsize == getLeft().size();
	}

	public void completeMatchingWithViolations(Stack<Node> remainingNodes) {
		for (int i = 0; i < getRight().length && !remainingNodes.isEmpty(); i++) {
			if (getRight()[i].getMatchNode() == -1) {
				remainingNodes.pop().setMatchNode(i);
			}
		}
	}

	/*
	 * @return the matching as an int table giving the room of each event
	 */
	public int[] getBestMatchingFound() {
		int[] match = new int[getLeft().size()];
		for (int i = 0; i < getLeft().size(); i++) {
			match[i] = getLeft().get(i).getMatchNode();
		}
		return match;
	}

	/**
	 * @param i the index of the event in the matching
	 * @return the room assigned to evt number i in the matching
	 */
	public int getMatch(int i) {
		return getLeft().get(i).getMatchNode();
	}

	/**
	 * switch the matching along the augmenting path
	 */
	public void improveMatchingAlongPath() {
		assert (path.size() % 2 == 0);
		while (!path.isEmpty()) {
			Node n1 = path.pop();
			Node n2 = path.pop();
			n1.setMatchNode(n2.idxInGraph);
			n2.setMatchNode(n1.idxInGraph);
		}
	}

	/**
	 * perform a dfs from the unmatched left nodes
	 *
	 * @return true if a path has been found
	 */
	public boolean findAugmentingPath(Node v) {
		path.clear();
		markright.clear();
		markleft.clear();
		return dfsFromLeft(v);
	}


	//mark nodes during exploration
	protected BitSet markright, markleft;
	//if this stack is empty, no augmenting path has been found, otherwise
	//it contains the augmenting path
	protected Stack<Node> path;

	//traverse "free" edges from left tp right
	public boolean dfsFromLeft(Node node) {
		path.push(node);
		markleft.set(node.idxInGraph);
		DoubleLinkedList succ = node.succ;
		succ.restart();
		while (succ.hasNext()) {
			int rn = succ.next();
			if (node.getMatchNode() != rn && !markright.get(rn)) {
				if (dfsFromRight(getRight()[rn]))
					return true;
			}
		}
		path.pop();
		return false;
	}

	//traverse "matched" edges from right to left	
	public boolean dfsFromRight(Node node) {
		markright.set(node.idxInGraph);
		path.push(node);
		if (node.getMatchNode() == -1) {	//an augmenting path has been found
			return true;
		} else {
			int ln = node.getMatchNode();
			if (!markleft.get(ln)) {
				if (dfsFromLeft(getLeft().get(ln)))
					return true;
			}
			path.pop();
			return false;
		}
	}

	//***************************************************//
	//************ Api Emmanuel *************************//
	//***************************************************//

	protected boolean matchingFound = false;
	protected int[] savedMatching;

	/**
	 * @return true if a matching has been found
	 */
	public boolean solve() {
		if (matchingFound || checkSolAlreadyKnown()) {
			matchingFound = true;
			return true;
		}
		assert (assertAll());
		matchingFound = computeLazyPerfectMatching();
		if (matchingFound)
			saveMatching();
		return matchingFound;
	}

	/**
	 * Save the current known matching to avoid to recompute it in case
	 * the solve is called on the same matching than the last time
	 */
	public void saveMatching() {
		Arrays.fill(savedMatching, -1);
		for (Iterator<Node> it = getLeft().iterator(); it.hasNext();) {
			Node node = it.next();
			savedMatching[node.getRef()] = node.getMatchNode();
		}
	}

	/**
	 * Check that the current state is not a known matching
	 */
	public boolean checkSolAlreadyKnown() {
		if (getLeft() != null) {
			for (Iterator<Node> it = getLeft().iterator(); it.hasNext();) {
				Node node = it.next();
				if (savedMatching[node.getRef()] == -1)
					return false;
			}
			for (Iterator<Node> it = getLeft().iterator(); it.hasNext();) {
				Node node = it.next();
				node.setMatchNode(savedMatching[node.getRef()]);
				getRight()[savedMatching[node.getRef()]].setMatchNode(node.idxInGraph);
			}
			reinitUnMatchedNode();
			return true;
		} else return true;
	}

	/**
	 * Add an event in the current matching
	 *
	 * @param evt the event to take into account in the current matching
	 */
	public void addEvent(int evt) {
		matchingFound = false;
		if (getLeft() == null) {
			setLeft(new ArrayList<Node>());
		}
		int freeposition = getLeft().size();
		Node evtnode = events[evt];
		evtnode.setMatchNode(-1);
		getLeft().add(evtnode);
		evtnode.partialInitLeft(freeposition, getRight());
		unmatchedNodes.push(evtnode);
	}

	/**
	 * Remove the last added event
	 */
	public void removeLastEvent() {
		getLeft().get(getLeft().size() - 1).resetRightNodes(getRight());
		getLeft().remove(getLeft().size() - 1);
		reinitUnMatchedNode();
	}

	/**
	 * Remove the interval [idx1,idx2]
	 * (included both events at position idx1, idx2).
	 *
	 * @param idx1
	 * @param idx2
	 */
	public void removeInterval(int idx1, int idx2) {
		int nbToRemove = idx2 - idx1 + 1;
		int nbTcpt = nbToRemove;
		while (nbTcpt > 0) {
			Node event = getLeft().get(idx1);
			event.resetRightNodes(getRight());
			getLeft().remove(idx1);
			nbTcpt--;
		}
		for (int i = idx1; i < left.size(); i++) {
			left.get(i).idxInGraph -= nbToRemove;
		}
		reinitUnMatchedNode();
	}

	/**
	 * reinit unmatched nodes
	 */
	public void reinitUnMatchedNode() {
		unmatchedNodes.clear();
		for (Iterator<Node> it = getLeft().iterator(); it.hasNext();) {
			Node node = it.next();
			if (node.getMatchNode() == -1)
				unmatchedNodes.push(node);
		}
	}

	//***************************************************//
	//************ assert api matching ******************//
	//***************************************************//

	public boolean assertAll() {
		for (Iterator<Node> it = unmatchedNodes.iterator(); it.hasNext();) {
			Node node = it.next();
			assert (node.getMatchNode() == -1);
			assert (left.contains(node));
		}
		if (left != null) {
			for (int i = 0; i < left.size(); i++) {
				Node event = left.get(i);
				assert (event.idxInGraph == i);
				if (event.getMatchNode() == -1) {
					assert (unmatchedNodes.contains(event));
				} else {
					assert (getRight()[event.getMatchNode()].getMatchNode() == i);
				}
			}
			for (int i = 0; i < getRight().length; i++) {
				Node room = getRight()[i];
				assert (room.idxInGraph == i);
				if (room.getMatchNode() != -1) {
					assert (left.get(room.getMatchNode()).getMatchNode() == i);
				}
			}
		}
		return true;
	}

	//***************************************************//
	//************ Tests matching ***********************//
	//***************************************************//

	public static void solveAndPrint(Matching malgo) {
		boolean isMatch = malgo.solve();
		System.out.println(isMatch + " : " + malgo.getLeft() + " " + Arrays.toString(malgo.getBestMatchingFound()));
	}

	public static void testEmmanuel() {
		// to do only once at the beginning :
		HardSolver s = new HardSolver("data/Track2/comp-2007-2-3.tim", 0);

		//then create a matching
		Matching malgo = new Matching(s.space);

		malgo.addEvent(41);
		solveAndPrint(malgo);
		malgo.addEvent(57);
		solveAndPrint(malgo);
		malgo.addEvent(68);
		solveAndPrint(malgo);
		malgo.addEvent(13);
		solveAndPrint(malgo);

		solveAndPrint(malgo);
		malgo.removeInterval(2, 2);
		solveAndPrint(malgo);
	}


	public static void main(String[] args) {
		testEmmanuel();
	}


	public void setLeft(ArrayList<Node> left) {
		this.left = left;
	}


	public ArrayList<Node> getLeft() {
		return left;
	}

	public int getLeftSize() {
		return (left == null) ? 0 : left.size();
	}

	public void setRight(Node[] right) {
		this.right = right;
	}


	public Node[] getRight() {
		return right;
	}

}
