package localsearch.greedy;

import localsearch.SolutionSpace;
import localsearch.solvers.complete.HardSolver;
import localsearch.solvers.coloring.HardColoringSolver;
import localsearch.moves.complete.VHungarianMove;
import instance.TimetablingInstance;
import instance.clique.MaxCliqueFinder;
import instance.clique.CompoundClique;

import java.util.*;

public class HungarianGreedy {

	public SolutionSpace space;
	public HardSolver hardsolver;
	public TimetablingInstance ins;

	public VHungarianMove vhunge;

	public List<List<Integer>> cevts;

	public HungarianGreedy(HardSolver hs) {
		this.space = hs.space;
		this.ins = hs.ins;
		this.vhunge = hs.vhunge;
		this.hardsolver = hs;
		MaxCliqueFinder mc = new MaxCliqueFinder(ins);
		mc.findCliques(true);
		CompoundClique[] cc = mc.getCliques();
		cevts = new LinkedList<List<Integer>>();
		for (CompoundClique clique : cc) {
			cevts.add(getEvts(clique));
		}
	}

	/**
	 * extract the list of events corresponding to the clique
	 */
	public List<Integer> getEvts(CompoundClique clique) {
		List<Integer> levts = new LinkedList<Integer>();
		BitSet nodes = clique.retrieveNodes();
		for (int i = nodes.nextSetBit(0); i > -1; i = nodes.nextSetBit(i + 1)) {
			levts.add(i);
		}
		return levts;
	}


	public int getFreeRoomInMatrix(int timeslot) {
		if (space.nbept[timeslot] < space.R) {
			for (int i = 0; i < space.R; i++) {
				if (space.matrix[timeslot][i] == -1) {
					return i;
				}
			}
		}
		return -1;
	}

	//this parameter should be randomized
	public boolean isAConflictClique(List<Integer> cs) {
		for (int evt : cs) {
			if (space.conflictcost[evt] > 0) return true;
		}
		return space.rand.nextFloat() < 0.1f;
	}

	/**
	 * Build an initial solution where cliques are taken into account
	 * as much as possible but greedily
	 */
	public void fillAssignmentMatrix() {
        space.initSolutionSpace();
        space.restart();
		for (List<Integer> cevt : cevts) {
			Collections.shuffle(cevt, space.rand);
			int nt = 0;
			for (Integer evt : cevt) {
				int t = space.timeslot[evt];
				if (t == -1) {
					int r;
					do {
						r = getFreeRoomInMatrix(nt);
						if (r != -1) space.addEvent(evt, nt, r);
						nt = (nt + 1) % space.T;
					} while (r == -1);
				}
			}
		}
		space.initMissingCosts();
		assert (space.assertAllHardCosts());
	}

	public int nbhc = 0;

	/**
	 * Iterate through the cliques and reoptimize each clique in sequence
	 * return true if an improvement has been made.
	 */
	public boolean mainHLoop() {
		boolean improve = false;
		for (List<Integer> cevt : cevts) {
			if (isAConflictClique(cevt)) {
				Collections.shuffle(cevt, space.rand);
				List<Integer> finalevts = new LinkedList<Integer>();
				BitSet finaltimeslots = new BitSet(space.T);
				for (Integer evt : cevt) {
					int t = space.timeslot[evt];
					if (!finaltimeslots.get(t) &&
						hardsolver.precedenceFree(evt, finalevts)) {
						finalevts.add(evt);
						finaltimeslots.set(t);
						vhunge.setEvt(t, evt);
					}
				}
				for (int i = 0; i < space.T; i++) {
					if (!finaltimeslots.get(i))
						vhunge.setEvt(i, -1);
				}
				int cm = vhunge.evalCostMove();
                //System.out.println("" + cm);
                vhunge.performMove();
				if (cm < 0 && hardsolver.isCostImproved()) {					
                    hardsolver.storeNewImprovingSol();
					improve = true;
					if (hardsolver.trace)
						System.out.println("greedy new best sol of cost " + hardsolver.bestCost);
				}
				nbhc++;
			}
		}
		return improve;
	}

	public static void main(String[] args) {
		HardSolver s = new HardSolver("data/Track2/comp-2007-2-6.tim", 0);
		HardSolver.trace = true;
        HungarianGreedy greedy = new HungarianGreedy(s);
		greedy.fillAssignmentMatrix();
		int cpt = 0;
		int st = (int) System.currentTimeMillis();
		while (cpt < 2000 && greedy.space.hardObj != 0) {
			cpt++;
			greedy.mainHLoop();
		}
		System.out.println(greedy.space.hardObj + " : " + cpt + " (" + ((int) System.currentTimeMillis() - st) + ") " + greedy.nbhc);

        s.outputSolution(false);
		//s.space.printSolution();
	}
}
