package localsearch.greedy.acogreedy;

import localsearch.solvers.complete.HardSolver;
import localsearch.SolutionSpace;
import localsearch.solvers.complete.AbstractSolver;

import java.util.*;

import instance.TimetablingInstance;
import cpmodel.LightModel;
import cpmodel.PrimalModel;
import choco.Problem;
import choco.ContradictionException;
import choco.Solver;
import choco.integer.search.StaticVarOrder;
import choco.integer.IntDomainVar;

/**
 * An approach that combine the local search as a repairing step
 * of solutions built by a ant greedy algorithm
 */
public class AcoGreedy {

    /**
     * external data (instance, repair and cp solvers)
     */
    public TimetablingInstance ins;
    public HardSolver hsolver;
    public LightModel lm;
    public Random rand;

    //parameters of the greedy ant process

    public boolean traceAnts = true;

    /**
     * Number of ants
     */
    public int nbAnt = 10;

    /**
     * w[i][j]: the weight associated to the timeslot of each event
     */
    public float[][] w;

    /**
     * Evaporation
     */
    public float rho = 0.75f;

    /**
     * initial world where initial propagation is stored
     */
    public int initWorld;

    public AcoGreedy(HardSolver hsolve) {
        ins = hsolve.ins;
        rand = hsolve.space.rand;
        hsolver = hsolve;
        lm = new LightModel(ins, hsolve, 0);
        w = new float[hsolver.space.E][hsolver.space.T];
        for (int i = 0; i < w.length; i++) {
            Arrays.fill(w[i], 0.5f);
        }
    }

    /**
     * Main solve method of the GreedyAnt-Repair algorithm
     * to find feasible solutions.
     */
    public int[] solve() {
        int iter = 0;
        initGreedySolver();
        List<int[]> antsols = new LinkedList<int[]>();
        List<int[]> antsolsAfterRepair = new LinkedList<int[]>();

        int[] bestpsol = null;
        int bestpe = 0;

        while (hsolver.getBestCost() != 0) {
            float avgevent = 0f;
            //greedy step to build solutions
            int maxNumberPlacedEvent = -1;
            for (int i = 0; i < nbAnt; i++) {
                int[] sol = solveOneGreedy();
                antsols.add(sol);
                if (traceAnts) {
                    int nbpe = 0;
                    for (int k = 0; k < sol.length; k++) {
                        if (sol[k] != -1) {
                            nbpe++;
                            avgevent += 1f;
                        }
                    }
                    if (nbpe > maxNumberPlacedEvent)
                        maxNumberPlacedEvent = nbpe;
                }
                int nbpe = 0;
                for (int j = 0; j < sol.length; j++) {
                    if (sol[j] != -1) nbpe++;
                }
                if (nbpe == sol.length) {
                    System.out.println("feasible solution found");
                    return sol;
                }
                if (nbpe > bestpe) {
                    bestpsol = sol;
                    bestpe = nbpe;
                }
                if (nbpe > maxNumberPlacedEvent) {
                    maxNumberPlacedEvent = nbpe;
                    antsolsAfterRepair.clear();
                    antsolsAfterRepair.add(sol);
                } else if (nbpe == maxNumberPlacedEvent) {
                    antsolsAfterRepair.add(sol);
                }

            }
            if (traceAnts) {
                System.out.println("[" + iter + "] average events placed " + avgevent / (float) nbAnt + " max " + maxNumberPlacedEvent);
            }

//            if (iter % 300 == 0) {
//                maxNumberPlacedEvent = -1;
//
////                antsols.clear();
////                antsols.addAll(antsolsAfterRepair);
////                antsolsAfterRepair.clear();
//
//                //repair step on each solution provided by the ants
//                //for (int[] sol : antsols) {
//                    completePartialSolRandomly(bestpsol);
//                    hsolver.setGreedyInit(bestpsol);
//                    hsolver.solve(); //add a limit in number of iterations
//                    maxNumberPlacedEvent = getNbPlacedEvent();
//                    if (hsolver.getBestCost() == 0)
//                        break; //feasible solution found
//
////                    else {
////                        int nbpe = getNbPlacedEvent();
////                        //System.out.println("Ant " + nbpe);
////                        if (nbpe > maxNumberPlacedEvent) {
////                            maxNumberPlacedEvent = nbpe;
////                            antsolsAfterRepair.clear();
////                            antsolsAfterRepair.add(getUnconflictedSolution());
////                        } else if (nbpe == maxNumberPlacedEvent) {
////                            antsolsAfterRepair.add(getUnconflictedSolution());
////                        }
////                    }
//                //}
//                if (hsolver.getBestCost() == 0)
//                    break; //feasible solution found
//
//                if (traceAnts)
//                    System.out.println("After Repair " + maxNumberPlacedEvent + " events placed");
//
//            }
            //update of weights
            for (int[] sol : antsolsAfterRepair) {
                intensifyWeights(sol, maxNumberPlacedEvent);
            }
            smoothWeights();

            //reset the solutions
            antsols.clear();
            antsolsAfterRepair.clear();
            iter++;
        }
        return null;
    }

    public void intensifyWeights(int[] sol, int nbPlacedEvts) {
        int unplacedEvts = sol.length - nbPlacedEvts;
        for (int i = 0; i < sol.length; i++) {
            int time = sol[i];
            if (time != -1) {
                w[i][time] += 100f / (float) unplacedEvts;
            }
        }
    }

    public void smoothWeights() {
        for (int i = 0; i < w.length; i++) {
            for (int j = 0; j < w[i].length; j++) {
                w[i][j] = rho * w[i][j];
            }
        }
    }

    //******************************************************************//
    //************** Choco solver as the greedy solver *****************//    
    //******************************************************************//

    /**
     * Build the choco model used as a greedy
     * and store its initial propagation in an initial world
     */
    public void initGreedySolver() {
        lm.buildModel(PrimalModel.CONFLICTS);
        Problem pb = lm.getChocoProblem();

        try {
            pb.propagate();
        } catch (ContradictionException e) {
            e.printStackTrace();
        }
        pb.worldPush();
        initWorld = pb.getWorldIndex();
    }


    /**
     * Light Model of choco search using probabilities
     */
    public int[] solveOneGreedy() {
        //lm = new LightModel(ins,hsolver,0);
        //initGreedySolver();
        int[] partialsol;
        Problem pb = lm.getChocoProblem();

        Solver s = pb.getSolver();
        StaticVarOrder svo = new StaticVarOrder(shuffle(lm.getEventTime()));
        //MinDomain svo = new MinDomain(pb,shuffle(lm.getEventTime()));
        ProbabilisticValueChooser pvc = new ProbabilisticValueChooser(w, rand);
        s.setVarIntSelector(svo);
        s.setValIntSelector(pvc);
        s.generateSearchSolver(pb);

        CustomFailLimit cfl = new CustomFailLimit(s.getSearchSolver(), lm.getEventTime(), 10);
        s.getSearchSolver().limits.add(cfl);
        s.launch();
        if (s.getNbSolutions() > 0) {
            IntDomainVar[] evts = lm.getEventTime();
            partialsol = new int[evts.length];
            for (int i = 0; i < evts.length; i++) {
                partialsol[i] = evts[i].getVal() - 1;                
            }
        } else {
            partialsol = cfl.getPartialSolution();
        }
        //reset the solver
        pb.worldPopUntil(initWorld);
        return partialsol;
    }

    /**
     * return a new table corresponding to evtTime once shuffled
     */
    public IntDomainVar[] shuffle(IntDomainVar[] evtTime) {
        IntDomainVar[] evts = new IntDomainVar[evtTime.length];
        List<IntDomainVar> levts = new ArrayList<IntDomainVar>();
        levts.addAll(Arrays.asList(evtTime));
        Collections.shuffle(levts, rand);
        levts.toArray(evts);
        return evts;
    }

    //******************************************************************//
    //*********** Repair solver ****************************************//
    //******************************************************************//


    /**
     * @return the number of events actually placed in the solution
     */
    public int getNbPlacedEvent() {
        int nbPlacedEvt = 0;
        SolutionSpace space = hsolver.space;
        for (int i = 0; i < space.E; i++) {
            if (space.eventcost[i] == 0) {
                nbPlacedEvt++;
            }
        }
        return nbPlacedEvt;
    }

    /**
     * Examine the current solution and keep only the non conflicting events
     *
     * @return
     */
    public int[] getUnconflictedSolution() {
        int[] eventTimeSol = new int[hsolver.space.E];
        Arrays.fill(eventTimeSol, -1);
        SolutionSpace space = hsolver.space;
        for (int i = 0; i < eventTimeSol.length; i++) {
            if (space.eventcost[i] == 0) {
                eventTimeSol[i] = space.timeslot[i];
            }
        }
        return eventTimeSol;
    }


    /**
     * complete randomly the partial solution found by the greedy
     */
    public void completePartialSolRandomly(int[] sol) {
        SolutionSpace space = hsolver.space;
        BitSet timeslotUnAvailable = new BitSet(space.T);
        int[] nbEvtInTimeSlot = new int[space.T];
        int totTimeSlot = space.T;
        for (int i = 0; i < sol.length; i++) {
            if (sol[i] != -1) {
                nbEvtInTimeSlot[sol[i]]++;
                if (nbEvtInTimeSlot[sol[i]] == space.R) {
                    timeslotUnAvailable.set(sol[i]);
                    totTimeSlot--;
                }
            }
        }
        for (int i = 0; i < sol.length; i++) {
            if (sol[i] == -1) {
                int tidx = rand.nextInt(totTimeSlot);
                int cpt = 0;
                boolean mark = false;
                for (int j = 0; j < space.T && !mark; j++) {
                    if (!timeslotUnAvailable.get(j)) {
                        if (cpt == tidx) {
                            sol[i] = j;
                            mark = true;
                        } else cpt++;
                    }
                }
                nbEvtInTimeSlot[sol[i]]++;
                if (nbEvtInTimeSlot[sol[i]] == space.R) {
                    timeslotUnAvailable.set(sol[i]);
                    totTimeSlot--;
                }
            }
        }
    }

    //******************************************************************//
    //*********** Main *************************************************//
    //******************************************************************//


    public static void main(String[] arg) {
        String name = "data/Track2/comp-2007-2-5.tim";
        long time = System.currentTimeMillis();
        int timelimit = 4200000;
        HardSolver s = new HardSolver(name, 0);
        AbstractSolver.trace = false;
        s.dynamicNeighbourhood();
        s.setGreedy();
        s.setMaxNbIter(20000);
        s.setTabuMetaHeuristic(10);
        //s.setStrictTabuMetaHeuristic(10);
        s.setTimelimit(timelimit);

        AcoGreedy acog = new AcoGreedy(s);
        int[] sol = acog.solve();
        s.setGreedyInit(sol);
        System.out.println("" + (System.currentTimeMillis() - time));
        s.solve();
        System.out.println("" + s.getBestCost() + " " + (System.currentTimeMillis() - time));
    }
}
