package localsearch.metah;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.solvers.complete.SoftSolver;
import localsearch.solvers.complete.SearchLoop;
import localsearch.moves.complete.*;
import localsearch.moves.Move;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.File;
import java.io.IOException;

/**
 * A SimulatedAnnealing controller to guide the local search
 */
public class SimulatedAnnealing extends MetaHeuristic {

	//1- Fixed parameters of the SA
	//cooling rate for geometric cooling
	protected double alpha = 0.95d;//0.95d;
	//cooling rate for inverse cooling
	protected double beta = 0.1d;

	//final temperature level
	public static double tempf = 0.2d;
	//the factor apply to find the initial temperature :
	//initial temperature = average delta * tempIFactor
	public static double tempIFactor = 1d;

	// reheating coefficient : temperature is reheated after a number of
	// non improving moves : reheatingFactor * tempLenght
	protected int reheatingFactor = 5;
	public static boolean reheat = false;

	//2- Parameters initialized online for the SA
	//the number of move at each temperature level
	//initial temperature level
	public double tempI;
	//the size of a temperature level (number of iterations performed at this level)
	protected int tempLenght;
	//the number of temperature steps established by the cooling
	public int nbstep;

	//3- Temporary data structure for the SA
	// current temperature
	public double currentTemp = 0;
	public int currentStep = 0;

	// current data for initializing
	// the other parameters of the SA
	protected int nbdelta = 0;
	protected int sumdelta = 0;
	public boolean initializationPhase;
	protected int startTime = 0;
	protected int nbIterEstim = 0;

	//parameters for reheating
	protected int bestCostSinceLastReheating = Integer.MAX_VALUE;
	protected int nbNonImpIter = 0;

	//control when to include the LNS in the SA
	//When entering the p final pourcentage of the cooling
	protected double p = 0.4d;

	//the SA is used only on a ratio of the allwed 10 minutes
	protected double ratio;

	public SimulatedAnnealing(AbstractSolver abs) {
		super(abs);
	}

	public SimulatedAnnealing(AbstractSolver abs, double ratio) {
		super(abs);
		this.ratio = ratio; 	
	}

	/**
	 * We first apply a search at temperature 1 to get an online estimate of
	 * - the average delta of the moves
	 * - the number of iteration performed in given timelimit
	 */
	public void initMhParameters() {
		tempLenght = 100000000; //abs.space.E * abs.space.T; //* abs.R ?
		currentTemp = 1;      //or 1 if tempf = 0.2
		initializationPhase = true;
		startTime = (int) System.currentTimeMillis();
	}

	public void reinit() {
		currentTemp = tempI;
	}

	public void initIntensification() {
		currentTemp = 1;
	}


	/**
	 * Compute the number of steps (of temperature) needed to reach the
	 * final temperature level of 0.2
	 *
	 * @param tinit
	 * @return
	 */
	public int computeNbStep(double tinit) {
		double temp = tinit;
		double prevtemp = -1;
		int point = 0;
		while (temp > tempf && prevtemp != temp) {
			temp = applyCooling(temp);
			point++;
		}
		return point;
	}

	public double applyCooling(double temp) {
		return applyGeometricCooling(temp);
	}

	//standart geometric cooling
	public double applyGeometricCooling(double temp) {
		return alpha * temp;
	}

	//cooling that allow to start at high temperature while still
	//spending time at low temperature
	public double applyInverseCooling(double temp) {
		return 1d / (1d / temp + beta);
	}

	public boolean isInTheLastTenPourcent() {
		if (currentStep >= (nbstep - nbstep * p))
			return true;
		else return false;
	}

	public boolean genericAccept(Move m, int delta) {
		if (initializationPhase) {
			sumdelta += Math.abs(delta);
			nbdelta++;
		}
		if (delta <= 0) {
			return true;
		} else {
			boolean accept = space.rand.nextDouble() <= computeProbability(delta);
			return accept;
		}
	}

	public double computeProbability(int delta) {
		return exp(((double) -delta) / currentTemp);
	}

	public static double exp(double val) {
		final long tmp = (long) (1512775 * val + (1072693248 - 60801));
		return Double.longBitsToDouble(tmp << 32);
	}

	public boolean accept(SwapEvents swe, int delta) {
		return genericAccept(swe, delta);
	}

	public boolean accept(TranslateEvent tre, int delta) {
		return genericAccept(tre, delta);
	}

	public boolean accept(VHungarianMove vmove, int delta) {
		return genericAccept(vmove, delta);
	}

	public boolean accept(SwapTimeslots swt, int delta) {
		return genericAccept(swt, delta);
	}

	public boolean accept(CPMove cpm, int delta) {
		return genericAccept(cpm, delta);
	}

	public boolean diversification = false;

	public void informOfMove(Move mov) {
		if (reheat && !initializationPhase) {
			//update parameters for reheating
			nbNonImpIter++;
			//update parameters for reheating
			if (abs.getCost() < bestCostSinceLastReheating) {
				bestCostSinceLastReheating = abs.getCost();
				nbNonImpIter = 0;
			} else {
				nbNonImpIter++;
			}
			//perform reheating if needed
			if (nbNonImpIter >= reheatingFactor * tempLenght) {
				currentTemp = tempI;
				bestCostSinceLastReheating = Integer.MAX_VALUE;
				nbNonImpIter = 0;
			}
		}
		//update temperature level
		if (abs.iter != 0 && abs.iter % tempLenght == 0) {
			currentTemp = applyCooling(currentTemp); //alpha * currentTemp;
			currentStep += 1;
			//System.out.println("" + currentTemp + " " + abs.iter + " " + tempLenght);
		}

		//initialize phase might be terminated, set the relevant parameters :
		//initial temperature and lenght of a temperatur level
		if (initializationPhase) {
			int ctime = (int) System.currentTimeMillis() - abs.starttime;
			if (abs.getNbIter() == 20000) {
				//temp code to mesure the mapping
				initializationPhase = false;
				if (nbdelta == 0) {
					tempI = 4d;
					throw new Error("SA: comment peut on avoir nbdelta null");
				} else {
					tempI = roundDouble(1, tempIFactor * ((double) sumdelta / (double) nbdelta));
					nbstep = computeNbStep(tempI);
					currentTemp = tempI;
					if (hardPrediction)
						hardEncodedPrediction(ctime);
					else predictionOnTime(ctime);
				}
			}
		}
	}

	public static boolean hardPrediction = true;

	public double nbHardIterEstim = 3000000;
	public double nbSoftIterEstim;


	public void hardEncodedPrediction(int ctime) {
		int timeLIMIT = (int) (372000 * ratio); //the speed of the soft solver was measured on Barry's computer
		nbSoftIterEstim = timeLIMIT / ((SoftSolver) abs).hsolve.ins.getSoftSpeed();
		int iterfeas = ((SoftSolver) abs).hsolve.iter;
		double percentOfTimeledt = 1d - iterfeas / nbHardIterEstim;
		if (percentOfTimeledt <= 0d) percentOfTimeledt = 0.1d;
		nbIterEstim = (int) Math.round(percentOfTimeledt * nbSoftIterEstim);
		tempLenght = nbIterEstim / nbstep + 1;
		if (SearchLoop.searchLoopTrace)
			System.out.println("FIX: " + abs.getNbIter() + " - TI:" + tempI + " TF:" + tempf + " I:" + nbIterEstim + " TL: " + tempLenght + "[" + nbstep + "] " + (abs.getTotalTimelimit() - ctime) + " " + roundInt(5000, abs.getTotalTimelimit() - ctime) + " " + iterfeas + " " + percentOfTimeledt);

	}

	public void predictionOnTime(int ctime) {
		int timeForInit = roundInt(400, ((int) System.currentTimeMillis()) - startTime);
		double speed = 20000d / (double) timeForInit;
		nbIterEstim = (int) Math.floor(speed * (roundInt(5000, abs.getTotalTimelimit() - ctime)));
		tempLenght = nbIterEstim / nbstep + 1;
		if (SearchLoop.searchLoopTrace)
			System.out.println("TIME:" + abs.getNbIter() + " - TI:" + tempI + " TF:" + tempf + " I:" + nbIterEstim + " TL: " + tempLenght + "[" + nbstep + "] " + timeForInit + " " + roundInt(5000, abs.getTotalTimelimit() - ctime));
	}

    //collect the average time of an iteration made over one million iterations
    //at the temperature of 1 for each instance.
    public static double itersec = 0.2d;    

    public double getIterPerSecond() {
		return (1 / itersec);// * (372000d / (double) abs.timelimit);
	}

	public static double roundDouble(int nbdigit, double d) {
		int fact = (10 * nbdigit);
		return (Math.floor(d * fact)) / fact;
	}

	public static int roundInt(int fact, double d) {
		return ((int) Math.round(d / fact) * fact);
	}


	public static void seeCoolingSchedule() {
		try {
			BufferedWriter bw = new BufferedWriter(new FileWriter(new File("schedul_07.txt")));
			double temp = 53.8;
			int point = 0;
			double prevtemp = 0;
			while (temp > 0.2 && temp != prevtemp) {
				bw.write(point + " " + temp);
				bw.newLine();
				prevtemp = temp;
				temp = 0.95 * temp;
				//temp = 1d / (1d / temp + 0.1d);
				point++;
			}
			bw.flush();
			bw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static void main(String[] args) {
		//System.out.println("" + (1 /0.2d));
		seeCoolingSchedule();
	}
}
