package localsearch.moves.coloring;

import localsearch.moves.complete.VHungarianMove;
import localsearch.SolutionSpace;
import localsearch.solvers.coloring.HardColoringSolutionSpace;
import localsearch.solvers.complete.AbstractSolver;
import localsearch.graphalgos.Hongrois;
import localsearch.graphalgos.MatchingSolution;

import java.util.Arrays;

/**
 *
 * The Hungarian move without considering the rooms
 */
public class ColoringVHungarianMove extends VHungarianMove {

    public ColoringVHungarianMove(SolutionSpace space) {
        this.space = space;
        this.id = AbstractSolver.VERT_HUNGARIAN_MOVE;
        rcosts = new int[space.T][space.T];
        evts = new int[space.T];
    }

    //the event choosed in timeslot i
    public void setEvt(int i, int evt) {
        assert (i <= space.T);
        assert (i >= 0);
        assert (evt <= space.E);
        evts[i] = evt;
    }

    /**
     * @return The cost of assigning evt in timeslot without the room cost
     */
    public int computeFixCost(int evt, int timeslot) {
        if (evts[timeslot] == evt)
            return space.eventcost[evt];
        int ca = space.timeslotFeas[evt][timeslot];
        int cn = 0;
        for (int i = 0; i < space.nbept[timeslot]; i++) {
            int ei = space.listevts[timeslot][i];
            cn += space.conflicts[ei][evt];
        }
        if (evts[timeslot] != -1)
            cn -= space.conflicts[evts[timeslot]][evt];
        int cp = space.evalPrecedenceCostFromMove(evt, -1, space.timeslot[evt], timeslot);
        return ca + cn + cp;
    }

    /**
     * @param evt      the event considered
     * @param timeslot the timeslot where evt could be allocated
     * @return the cost of allocating evt in timeslot
     */
    public int computeCostForEvt(int evt, int timeslot) {
        return computeFixCost(evt, timeslot);
    }

    /**
     * Fill the cost matrix
     */
    public void computeMatrixCost() {
        for (int i = 0; i < space.T; i++) {
            if (evts[i] != -1) {
                int evt = evts[i];
                for (int j = 0; j < space.T; j++) {
                    if (evts[j] == -1 && space.nbept[j] == space.R) {
                        rcosts[i][j] = 100000;
                    } else rcosts[i][j] = computeCostForEvt(evt, j);
                }
            } else {
                Arrays.fill(rcosts[i], 0);
            }
        }
    }


    /**
     * Evaluate the objective function variation by optimally reassigning a set of
     * at most 45 events by taken one per timeslot.
     *
     * @return
     */
    public int evalCostMove() {
        int cost = 0;
        int ccost = computeCurrentCost();
        //if (ccost == 0) return 1; // no way to do better
        computeMatrixCost();
        HungarishAlgo = new Hongrois(rcosts, space);
        HungarishAlgo.solve();
        for (int i = 0; i < space.T; i++) {
            cost += rcosts[i][HungarishAlgo.getAffectedTask(i)];
        }
        delta = cost - ccost;
        return delta;
    }


    public void performMove() {
        int[] oldtimeslot = new int[space.T];
        for (int i = 0; i < space.T; i++) {
            if (evts[i] != -1) { //the event has to be removed
                oldtimeslot[i] = space.timeslot[evts[i]];
                space.removeEvent(evts[i]);
            }
        }
        for (int i = 0; i < space.T; i++) {
            int evt = evts[i];
            if (evt != -1) {
                int timeslot = HungarishAlgo.getAffectedTask(i);
                ((HardColoringSolutionSpace) space).addEvent(evt, timeslot);
            }
        }
        //maintain precedence cost
        for (int i = 0; i < space.T; i++) {
            int evt = evts[i];
            if (evt != -1) {
                space.maintainPrecedenceCostFromMove(evt, -1, oldtimeslot[i], HungarishAlgo.getAffectedTask(i));
            }
        }
    }

    public String toString() {
        return "Hungarian move inter timeslot";
    }

}
