package localsearch.moves.complete;

import localsearch.solvers.complete.AbstractSolver;
import localsearch.SolutionSpace;
import localsearch.moves.Move;
import localsearch.graphalgos.Hongrois;
import localsearch.graphalgos.MatchingInterface;
import localsearch.graphalgos.MatchingSolution;

import java.util.Arrays;
import java.util.BitSet;


/**
 * Created by IntelliJ IDEA.
 * User: hcambaza
 * Date: 20 aot 2007
 * Time: 17:48:19
 * To change this template use File | Settings | File Templates.
 */
public class VHungarianMove extends Move {

	//the cost of allocating event i in timeslot j
	public int[][] rcosts;

	//A table of 45 events to reallocate
	public int[] evts;

	//original room
	public int[] orooms;
	//Store for each timeslot and event, the room corresponding to the best
	//cost for the hungarian (-2 means that the matching has to be used to find
	//all new rooms of the timeslot)
	public int[][] frooms;

	//Matchings per timeslot to reoptimize the rooms of a timeslot if needed.
	protected MatchingInterface[] matchings;
	protected MatchingSolution[][] matchingFounds;

	// the Hungarish algorithm
	public Hongrois HungarishAlgo;

    public VHungarianMove() {}

    public VHungarianMove(SolutionSpace space) {
		super(space);
		this.id = AbstractSolver.VERT_HUNGARIAN_MOVE;
		rcosts = new int[space.T][space.T];
		evts = new int[space.T];
		frooms = new int[space.T][space.E];
		orooms = new int[space.T];
		matchings = new MatchingInterface[space.T];
		for (int i = 0; i < space.T; i++) {
			matchings[i] = new MatchingInterface(space);
		}
        matchingFounds = new MatchingSolution[space.T][space.E];
        for (int i = 0; i < space.T; i++) {
            for (int j = 0; j < space.E; j++) {
                matchingFounds[i][j] = new MatchingSolution();
            }
		}
	}

	//the event choosed in timeslot i
	public void setEvt(int i, int evt) {
		assert (i <= space.T);
		assert (i >= 0);
		assert (evt <= space.E);
		evts[i] = evt;
		if (evt != -1)
			orooms[i] = space.rooms[evt];
		else orooms[i] = -1;
	}

	/**
	 * @return The cost of assigning evt in timeslot without the room cost
	 */
	public int computeFixCost(int evt, int timeslot) {
		if (evts[timeslot] == evt) return space.eventcost[evt] - space.roomFeas[evt][space.rooms[evt]];
		int ca = space.timeslotFeas[evt][timeslot];
		int cn = 0;
		for (int i = 0; i < space.nbept[timeslot]; i++) {
			int ei = space.listevts[timeslot][i];
			cn += space.conflicts[ei][evt];
		}
		if (evts[timeslot] != -1)
			cn -= space.conflicts[evts[timeslot]][evt];
		int cp = space.evalPrecedenceCostFromMove(evt, -1, space.timeslot[evt], timeslot);
		return ca + cn + cp;
	}

	/**
	 * @param evt      the event considered
	 * @param timeslot the timeslot where evt could be allocated
	 * @return the cost of allocating evt in timeslot
	 */
	public int computeCostForEvt(int evt, int timeslot) {
		int fixcost = computeFixCost(evt, timeslot);
		int eir = orooms[timeslot];
		if (eir != -1) {
			if (space.roomFeas[evt][eir] == 0) { //the existing room is ok for evt
				frooms[timeslot][evt] = eir;
				return fixcost;
			}
		}
		for (int ri = 0; ri < space.R; ri++) {
			if (space.matrix[timeslot][ri] == -1) {
				if (space.roomFeas[evt][ri] == 0) { //a free room is ok for evt
					frooms[timeslot][evt] = ri;
					return fixcost;
				}
				if (eir == -1) {
					eir = ri; //store any other free room otherwise in the mean time
				}
			}
		}
		//check the matching to find a free room
		//because everything else has failed !
		if (evts[timeslot] == evt) {
			//matchings[timeslot].setTimeSlot(timeslot);
            matchings[timeslot].setPartialTimeSlot(timeslot,evt,-1);
		} else if (evts[timeslot] == -1) {
			//matchings[timeslot].setTimeSlot(timeslot, evt);
            matchings[timeslot].setPartialTimeSlot(timeslot,evt,-1);
		} else {
			//matchings[timeslot].setTimeSlot(timeslot, evts[timeslot], evt);
            matchings[timeslot].setPartialTimeSlot(timeslot,evt,evts[timeslot]);
        }
		if (matchings[timeslot].solveMatching()) {
            //matchingFounds[timeslot][evt] = matchings[timeslot].getMatchingSolution();
			matchings[timeslot].getMatchingSolutionWithLeftEvts(matchingFounds[timeslot][evt]);
			frooms[timeslot][evt] = -2; //means that the matching will be considered when reassigning the evt
            return fixcost;
		}
		frooms[timeslot][evt] = eir;
		return fixcost + 1;
	}

	/**
	 * Fill the cost matrix
	 */
	public void computeMatrixCost() {
		for (int i = 0; i < space.T; i++) {
			if (evts[i] != -1) {
				int evt = evts[i];
				for (int j = 0; j < space.T; j++) {
					if (evts[j] == -1 && space.nbept[j] == space.R) {
						rcosts[i][j] = 100000;
					} else rcosts[i][j] = computeCostForEvt(evt, j);
				}
			} else {
				Arrays.fill(rcosts[i], 0);
			}
		}
	}

	public void printCostMatrix() {
		System.out.println("COST MATRIX");
		for (int i = 0; i < space.E; i++) {
			System.out.print((space.eventcost[i]) + " ");
		}
		System.out.println("COST MATRIX");
		for (int i = 0; i < space.T; i++) {
			System.out.println(Arrays.toString(rcosts[i]));
		}
	}

	/**
	 * Compute the cost of the current assignment
	 *
	 * @return
	 */
	public int computeCurrentCost() {
		int cost = 0;
		for (int i = 0; i < evts.length; i++) {
			int evt = evts[i];
			if (evt != -1) {
				cost += space.eventcost[evt];
			}
		}
		return cost;
	}


	/**
	 * Evaluate the objective function variation by optimally reassigning a set of
	 * at most 45 events by taken one per timeslot.
	 *
	 * @return
	 */
	public int evalCostMove() {
		int cost = 0;
		int ccost = computeCurrentCost();
		//if (ccost == 0) return 1; // no way to do better
		computeMatrixCost();
		HungarishAlgo = new Hongrois(rcosts, space);
		HungarishAlgo.solve();
		for (int i = 0; i < space.T; i++) {
			cost += rcosts[i][HungarishAlgo.getAffectedTask(i)];
		}
		delta = cost - ccost;
		return delta;
	}


	public void performMove() {
        int previouscost = space.hardObj;
        int[] oldtimeslot = new int[space.T];
		for (int i = 0; i < space.T; i++) {
			if (evts[i] != -1) { //the event has to be removed
				oldtimeslot[i] = space.timeslot[evts[i]];
				space.removeEvent(evts[i]);
				space.addFreePlace(oldtimeslot[i], orooms[i]);
			}
		}
        for (int i = 0; i < space.T; i++) {
			int evt = evts[i];
			if (evt != -1) {
				int timeslot = HungarishAlgo.getAffectedTask(i);
				int r = frooms[timeslot][evt];
				if (r == -2) { //a matching has been needed here !
					reorganizeTimeslotBasedOnMatching(timeslot, evt);
					r = matchingFounds[timeslot][evt].getRoom(evt);
				}
				space.addEvent(evt, timeslot, r);
				space.removeFreePlace(timeslot, r);
			}
		}
		//maintain precedence cost
		for (int i = 0; i < space.T; i++) {
			int evt = evts[i];
			if (evt != -1) {
				space.maintainPrecedenceCostFromMove(evt, -1, oldtimeslot[i], HungarishAlgo.getAffectedTask(i));
			}
		}
        if (previouscost + delta < space.hardObj)
            System.out.println((previouscost + delta) + " " + space.hardObj);
        assert(previouscost + delta >= space.hardObj);
    }

	public void reorganizeTimeslotBasedOnMatching(int timeslot, int e) {
		MatchingSolution matchsol = matchingFounds[timeslot][e];
		for (int i = 0; i < space.nbept[timeslot]; i++) {
			int evt = space.listevts[timeslot][i];
			if (evt != evts[timeslot]) space.fastRemoveEvent(evt);
		}
        int[] matEvts = matchsol.getEvts();
		int[] matching = matchsol.getMatching();
		for (int i = 0; i < matEvts.length; i++) {
			int evt = matEvts[i];
			if (evt != e) space.fastAddEvent(evt, matching[i]);
		}
		//maintain the free place data structure
		space.maintainFreePlacesOnTimeslot(timeslot);
	}

	public String toString() {
		return "Hungarian move inter timeslot";
	}

}
