package localsearch.solvers.coloring;

import instance.TimetablingInstance;
import instance.Precedence;

import java.util.*;
import java.io.FileInputStream;
import java.io.FileNotFoundException;

import tools.InstanceReader;
import localsearch.SolutionSpace;
import localsearch.HardSolutionSpace;

/**
 * A solution space dedicated to coloring only
 */
public class HardColoringSolutionSpace extends SolutionSpace {


    public HardColoringSolutionSpace(TimetablingInstance ins, int seed) {
        this.ins = ins;
        this.rand = new Random(seed);
        this.T = 45;
        this.D = 5;
        this.H = 9;
        this.R = ins.getNRooms();
        this.S = ins.getNStudents();
        this.E = ins.getNEvents();
        initInstanceData();
    }


    //******** initialization of the data structure ************//

    public void initSolutionSpace() {
        initSolutionRepresentation();
    }

    public void initInstanceData() {
        timeslotFeas = new int[E][T];
        conflicts = new int[E][E];
        before = new boolean[E][E];
        greaterthan = new int[E][]; //add a fake -1 at the end to avoid testing the end of the table
        lessthan = new int[E][]; //add a fake -1 at the end to avoid testing the end of the table
        for (int i = 0; i < E; i++) {
            greaterthan[i] = new int[ins.getNInvprec(i)];
            lessthan[i] = new int[ins.getNPrec(i)];
            for (int j = 0; j < T; j++) {
                timeslotFeas[i][j] = ins.eventIsCompatibleWithTimeslot(i, j) ? 0 : 1;
            }
            for (int j = i + 1; j < E; j++) {
                conflicts[i][j] = (ins.eventsShareStudents(i, j) || ins.isRoomConflictBetweenEvents(i,j)) ? 1 : 0;
                conflicts[j][i] = conflicts[i][j];
            }
            Arrays.fill(greaterthan[i], -1);
            Arrays.fill(lessthan[i], -1);
        }
        for (Precedence p : ins.eventPrecedences()) {
            //e1 < e2
            int e1 = p.firstEvent;
            int e2 = p.secondEvent;
            before[e1][e2] = true;
            addEvtInPrecTable(greaterthan[e2], e1);
            addEvtInPrecTable(lessthan[e1], e2);
        }
    }

    public void addEvtInPrecTable(int[] ptable, int evt) {
        int i = 0;
        while (ptable[i] != -1) {
            i++;
        }
        ptable[i] = evt;
    }

    public void initSolutionRepresentation() {
        listevts = new int[T][R];
        timeslot = new int[E];
        hardObj = 0;
        eventcost = new int[E];
        conflictcost = new int[E];
        nbept = new int[T];
        for (int i = 0; i < T; i++) {
            Arrays.fill(listevts[i], -1);
        }
    }

    //******** Basic operation on events *********************//

    /**
     * Remove event evt from its current location while
     * maintaining all data structures and costs (except precedence cost)
     *
     * @param evt the event removed
     */
    public void removeEvent(int evt) {
        int e1t = timeslot[evt];
        timeslot[evt] = -1;
        hardObj -= eventcost[evt];
        int pevt = -1;
        for (int i = 0; i < nbept[e1t]; i++) { //reduce the cost of the neighbours
            int ei = listevts[e1t][i];
            if (ei != evt) {
                eventcost[ei] -= conflicts[ei][evt];
                conflictcost[ei] -= conflicts[ei][evt];
            } else pevt = i;
        }
        eventcost[evt] = 0;
        conflictcost[evt] = 0;
        maintainListEvt(evt, e1t, pevt);
        nbept[e1t]--;
    }


    /**
     * Place event evt in the EMPTY place corresponding to the given
     * timeslot and room while maintaining all data structures and
     * cost (except precedence cost).
     *
     * @param evt     the event to add
     * @param timeslo the timeslot where we want to put the event
     * @param room    the room where the event has to go
     */
    public void addEvent(int evt, int timeslo) {
        assert (nbept[timeslo] < R);
        int evtcost = timeslotFeas[evt][timeslo];
        int conflict = 0;
        for (int i = 0; i < nbept[timeslo]; i++) {
            int ei = listevts[timeslo][i];
            conflict += conflicts[ei][evt];
            conflictcost[ei] += conflicts[ei][evt];
            eventcost[ei] += conflicts[ei][evt];
        }
        eventcost[evt] = evtcost + conflict;
        conflictcost[evt] = conflict;
        hardObj += eventcost[evt];
        timeslot[evt] = timeslo;
        listevts[timeslo][nbept[timeslo]] = evt;
        nbept[timeslo]++;
        //System.out.println("add event " + evt + " in timeslot " + timeslo + " and room " + room);
    }


    /**
     * move an event without maintaining the precedence cost
     *
     * @param evt1 the two event to move with its corresponding previous and
     *             new precedence cost.
     */
    public void fastMoveTimeslot(int evt1, int e2t) {
        int e1t = timeslot[evt1];
        timeslot[evt1] = e2t;
        eventcost[evt1] -= timeslotFeas[evt1][e1t];
        eventcost[evt1] += timeslotFeas[evt1][e2t];
        hardObj -= timeslotFeas[evt1][e1t];
        hardObj += timeslotFeas[evt1][e2t];
    }

    /**
     * swap two evts without maintaining the precedence cost
     *
     * @param evt1,evt2 the two events to swap.
     */
    public void fastSwitchTimeslot(int evt1, int evt2) {
        int e1t = timeslot[evt1];
        int e2t = timeslot[evt2];
        timeslot[evt2] = e1t;
        timeslot[evt1] = e2t;
        eventcost[evt1] -= timeslotFeas[evt1][e1t];
        eventcost[evt1] += timeslotFeas[evt1][e2t];
        eventcost[evt2] -= timeslotFeas[evt2][e2t];
        eventcost[evt2] += timeslotFeas[evt2][e1t];
        hardObj += timeslotFeas[evt1][e2t] - timeslotFeas[evt1][e1t];
        hardObj += timeslotFeas[evt2][e1t] - timeslotFeas[evt2][e2t];
    }

    //******** initialization of the solution ****************//

    public void restart() {
        for (int i = 0; i < T; i++) {
            Arrays.fill(listevts[i], -1);
        }
        Arrays.fill(timeslot, -1);
        Arrays.fill(nbept, 0);
        Arrays.fill(eventcost, 0);
        Arrays.fill(conflictcost, 0);
        hardObj = 0;
    }

    /**
     * Randomly place each event in the assignment matrix
     */
    public void randomInit() {
        restart();
        int fullt = 0;
        for (int i = 0; i < E; i++) {
            int position = rand.nextInt((T - 1) - fullt);
            int cpt = 0;
            boolean mark = false;
            for (int j = 0; j < T && !mark; j++) {
                if (nbept[j] < R) {
                    if (cpt == position) {
                        addEvent(i, j);
                        if (nbept[j] == R) fullt++;
                        mark = true;
                    } else cpt++;
                }
            }
        }
        initMissingCosts();
    }


    /**
     * Place each event in the assignment matrix according to a timeslot
     * solution
     */
    public void initFromTimeslotSol(int[] timeslots) {
        restart();
        for (int i = 0; i < E; i++) {
            int timeslot = timeslots[i];
            addEvent(i, timeslot);
        }
        initMissingCosts();
    }


    /**
     * Costs involving greaterthan have to be computed when all position of the events
     * are known ! Thus a separate method to be called after everything has been
     * reassigned
     */
    public void addGreaterInCost(int evt) {
        for (int i = 0; i < greaterthan[evt].length; i++) {
            if (timeslot[greaterthan[evt][i]] >= timeslot[evt]) {
                eventcost[evt]++;
                hardObj++;
            }
        }
    }

    /**
     * init the free slots data structure
     */
    public void initMissingCosts() {
        for (int i = 0; i < E; i++) {
            addGreaterInCost(i);
        }
    }

    //******** print solution ********************************//

    public void printSolution() {
        for (int i = 0; i < T; i++) {
            System.out.println(Arrays.toString(listevts[i]));
        }
        System.out.println("violation " + hardObj);
        System.out.println("details of violations ");
        System.out.println(Arrays.toString(eventcost));
        System.out.println(Arrays.toString(conflictcost));
    }

    public void printCost() {
        int c = 0;
        for (int i = 0; i < E; i++) {
            c += conflictcost[i];
        }
        System.out.println("violation " + hardObj);
        System.out.println("cost conflict: " + c / 2);
        System.out.println("cost room: " + (hardObj - c / 2));
    }

    //******** simple tests **********************************//

    public static void main(String[] args) {
        InstanceReader reader = null;
        try {
            reader = new InstanceReader(new FileInputStream("data/Track2/comp-2007-2-2.tim.txt"));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
        TimetablingInstance ins = reader.getInstance();
        HardSolutionSpace sp = new HardSolutionSpace(ins, 1);
        sp.randomInit();
        sp.printSolution();
        System.out.println("fin");
    }

    //*************** Assert current data struture*************//

    public boolean assertAllSoftCosts() {
        return true;
    }


    public boolean assertFeasible() {
        return true;
    }


    public boolean assertAllHardCosts() {
        return assertHardCosts();
    }


    public boolean assertHardCosts() {
        int asserthardObj = 0;
        int totalconfcost = 0;
        int[] cardEvtsPerTimeslot = new int[T];
        for (int e = 0; e < E; e++) {
            int et = timeslot[e];
            cardEvtsPerTimeslot[et]++;
            assert (et != -1);
            int coste = timeslotFeas[e][et];
            int conflicte = 0;
            for (int r = 0; r < R; r++) {
                int e2 = listevts[et][r];
                if (e2 != -1 && e2 != e) {
                    conflicte += conflicts[e][e2];
                }
            }
            int preccoste = 0;
            for (int i = 0; i < greaterthan[e].length; i++) {
                if (timeslot[greaterthan[e][i]] >= et) {
                    coste++;
                }
            }
            //System.out.println("" + preccoste);
            totalconfcost += conflicte;
            coste += conflicte;
            if (coste != eventcost[e])
                System.out.println("event " + e + " " + coste + " != " + eventcost[e]);
            if (conflicte != conflictcost[e])
                System.out.println("event " + e + " " + conflicte + " != " + conflictcost[e]);
            //System.out.println(conflicte + " " + e + " " + timeslot[e]);
            assert (coste == eventcost[e]);
            assert (conflicte == conflictcost[e]);
            asserthardObj += coste;
        }
        for (int i = 0; i < T; i++) {
            assert (nbept[i] == cardEvtsPerTimeslot[i]);
        }
        asserthardObj -= totalconfcost / 2;
        if (asserthardObj != hardObj)
            System.out.println(asserthardObj + " " + hardObj);
        assert (asserthardObj == hardObj);
        //System.out.println(" " + costtf);
        for (int i = 0; i < E; i++) {
            int eit = timeslot[i];
            boolean isInList = false;
            for (int j = 0; j < nbept[eit] && !isInList; j++) {
                if (listevts[eit][j] == i) isInList = true;
            }
            if (!isInList) System.out.println("evt " + i + " in " + eit);
            assert (isInList);
        }
        return true;
    }
}
