package localsearch.solvers.complete;

import instance.TimetablingInstance;
import localsearch.*;
import localsearch.solvers.LSSolver;
import localsearch.graphalgos.matchings.Matching;
import localsearch.metah.*;
import localsearch.moves.complete.*;
import tools.InstanceReader;

import java.io.*;
import java.util.Collections;
import java.util.LinkedList;
import java.util.HashMap;

/**
 *
 * The root class for all local search solvers
 */
public abstract class AbstractSolver implements LSSolver {

	public static boolean trace = false, debug = false;

	public String name;
	public TimetablingInstance ins;
	public Tracer tracy;
	public static boolean turnTracerOn = false;

	//************* solver parameters ****************//

	public int starttime;
	public int endtime;
	public int timelimit;
	public int iter = 0;
	public int nonImprovIter = 0;
    public int maxNbIter = Integer.MAX_VALUE;
    public boolean firstRun = true;
	public int seed;

	public boolean chocoinit = false;
    public int[] eventTime;


    //solve the matchings in the initial solution or not
    public boolean initmatching = false;

    //definition of the neihbourhood
	public LinkedList<Integer> moves;
	public SwapEvents swape;
	public TranslateEvent transe;
	public TranslateMatchingEvent transem;
	public SwapTimeslots swapt;
	public VHungarianMove vhunge;
	public HHungarianMove hhunge;
	public CPMove cpmove;
	public static final int SWAP_EVENTS = 0;
	public static final int TRANSLATE_EVENTS = 1;
	public static final int SWAP_TIMESLOTS = 2;
	public static final int VERT_HUNGARIAN_MOVE = 3;
	public static final int HORIZ_HUNGARIAN_MOVE = 4;
	public static final int TRANSLATE_EVENTS_MATCHING = 5;
	public static final int CP_MOVE = 6;
	public boolean GREEDY = false;
	public boolean LNS = false;


	public boolean dynamicNeihbourhood = false;
	public static double pmt = 0.01d, phh = 0.001d, pst = 0.0005d, pvh = 0.0005d;
	protected int nbRandomHungarian = 20;

	protected boolean restartHardSolver = false;
	protected int restartThreshold = 121000;

	//************* solution managment ***************//

	//the current solution and its incremental data structure
	public SolutionSpace space;

	//the best solution is stored as the assignment matrix
	//with its best cost
	public int[][] bestsolution;
	public int bestCost = Integer.MAX_VALUE;

	//************* Metaheuristic parameters ********************//
	public MetaHeuristic mh;

	//default constructor
	public AbstractSolver() {

	}

	//boolean includeS1: include the soft constraint 1 as a hard constraint
	public AbstractSolver(TimetablingInstance ins, int seed) {
		buildAbstractSolver(ins,seed);
		this.name  = ins.getName();
	}

	public AbstractSolver(String name, int seed) {
		InstanceReader reader = null;
		try {
			reader = new InstanceReader(new FileInputStream(name));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
		buildAbstractSolver(reader.getInstance(),seed);
		this.name  = name;		
	}

	public void buildAbstractSolver(TimetablingInstance ins, int seed) {
		this.starttime = (int) System.currentTimeMillis();
		this.seed  = seed;
		this.ins = ins;
		this.space = new SolutionSpace(ins, seed);
		Matching.initNodeEvents(space);
		this.moves = new LinkedList<Integer>();
		this.swape = new SwapEvents(space);
		this.swapt = new SwapTimeslots(space);
		this.transe = new TranslateEvent(space);
		this.vhunge = new VHungarianMove(space);
		this.hhunge = new HHungarianMove(space);
		this.cpmove = new CPMove(space,this);
		this.transem = new TranslateMatchingEvent(space);
		this.bestsolution = new int[space.T][space.R];
		this.mh = new TrueMH(this);
	}

	public abstract void setLNS();

	public void setProbalisticTabu(int tsize) {
		mh = new ProbabilisticTabu(this, tsize);
	}

	public void setTabuMetaHeuristic(int tsize) {
		mh = new Tabu(this, tsize);
	}

    public void setStrictTabuMetaHeuristic(int tsize) {
		mh = new TabuStrictlyImprove(this, tsize);
	}

    public void setSAMetaHeuristic() {
		mh = new SimulatedAnnealing(this);
	}

	public void setSAMetaHeuristic(double ratio) {
		mh = new SimulatedAnnealing(this,ratio);
	}

	public void setRLMetaHeuristic() {
		mh = new ReinforcementLearning(this);
	}

	public HardSolutionSpace getSpace() {
		return space;
	}

	public void setGreedyInit(int[] eventTimeslot) {
		this.chocoinit = true;
		this.eventTime = eventTimeslot;
	}

	public int getTotalTimelimit() {
		return timelimit;
	}

	public void setTimelimit(int timelimit) {
		this.timelimit = timelimit;
	}

    public void setStartTime(int stime) {
        starttime = stime;
    }

    public int getStartTime() {
        return starttime;
    }

    public int getBestCost() {
		return bestCost;
	}

	public abstract int getCost();

	public void setBestCost(int bestCost) {
		this.bestCost = bestCost;
	}

	public void setGreedy() {
		GREEDY = true;
	}

    public void setMaxNbIter(int miter) {
        maxNbIter = miter;
    }

    public String getCompInfo() {
		return "";
	}

	public int getNbIter() {
		return iter;
	}

	public void dynamicNeighbourhood() {
		dynamicNeihbourhood = true;
	}

    public void addNeighbourhood(int m) {
		assert (m <= 6);
		moves.addFirst(m);
	}

	public boolean addLNSStep() {
		return false;
	}
	/**
	 * Define the neighbourhood used at the current iteration
	 * it is controlled by a probability p for each kind of move
	 */
	public void defineNeighbourhood() {
		if (dynamicNeihbourhood) {
			moves.clear();
			moves.add(TRANSLATE_EVENTS);
			moves.add(SWAP_EVENTS);
			if (space.rand.nextDouble() <= pmt) {
				moves.add(TRANSLATE_EVENTS_MATCHING);
			}
			if (space.rand.nextDouble() <= pst) {
				moves.add(SWAP_TIMESLOTS);
			}
			if (space.rand.nextDouble() <= phh) {
				moves.add(HORIZ_HUNGARIAN_MOVE);
			}
			if (space.rand.nextDouble() <= pvh) {
				moves.add(VERT_HUNGARIAN_MOVE);
			}
		}
		if (LNS && addLNSStep()) {
			moves.add(CP_MOVE);
		}
		Collections.shuffle(moves, space.rand);
	}

	public void start() {
		if (firstRun) {
			space.initSolutionSpace();
		}
		iter = 0;
		bestCost = Integer.MAX_VALUE;
		if (!dynamicNeihbourhood && moves.isEmpty()) {
			throw new Error("no neighbourhood defined !");
		}
	}

	public void end() {
		endtime = (int) System.currentTimeMillis();
		if (turnTracerOn) {
            tracy.print(getBestCost(),iter,true);
            tracy.close();
        }
	}

	public int getTotalTime() {
		return (endtime - starttime);
	}

	public boolean timelimitReached() {
		return (((int) System.currentTimeMillis() - starttime) > timelimit);
	}

	public boolean doWeStop() {
		return (bestCost == 0) || iter > maxNbIter || (((int) System.currentTimeMillis() - starttime) > timelimit);
	}

	public void storeNewImprovingSol() {
		for (int i = 0; i < space.T; i++) {
			for (int j = 0; j < space.R; j++) {
				bestsolution[i][j] = space.matrix[i][j];
			}
		}
		bestCost = getCost();
	}

	public void restoreBestSolution() {
		space.initFromGivenSol(bestsolution);
	}


	//**********************************************************//
	//******************** call the checker ********************//
	//**********************************************************//


	public void writeSolutionIn(String fileName) throws IOException {
		BufferedWriter bw = new BufferedWriter(new FileWriter(new File(fileName + ".sln")));
		for (int i = 0; i < space.E; i++) {
			Place p = getPosition(i);
			bw.write(p.t + " " + p.r);
			bw.newLine();
		}
		bw.flush();
		bw.close();
	}

	public void outputSolution(boolean soft) {
		try {
			String noutput = name.substring(0, name.length() - 4);
			writeSolutionIn(noutput);
			writeSolutionIn(noutput + "_" + seed + "_" + Benchmark.unique_ID);
			if (Benchmark.checkResults) {
				runCheckerFromJava(noutput, soft ? getBestCost() : -1);

			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public static void runCheckerFromJava(String pureName, int scost) throws IOException {
		String cmd = "lib/checksln3b " + " " + pureName;
		Runtime rt = Runtime.getRuntime();
		Process p = rt.exec(cmd);
		InputStream in = p.getInputStream();
		BufferedReader readrel = new BufferedReader(new InputStreamReader(in)); //new FileReader(new File(outname)));
		String line = readrel.readLine();
		boolean startreport = false;
		while (line != null && !line.equals("Press return to continue")) {
			if (line.length() > 6 && line.substring(0, 6).equals("Number"))
				startreport = true;
			line = readrel.readLine();
			if (startreport && SearchLoop.searchLoopTrace)
				System.out.println("" + line);
			if (line.equals("***This solution file does not give a valid timetable***"))
				throw new Error("checker failed on feasability");
			if (scost != -1 && line.startsWith("Total soft constraint penalty =")) {
				int softcost = Integer.parseInt(line.substring(32));
				//System.out.println(softcost + " " + scost);
				if (softcost != scost)
					throw new Error("softcost not properly maintained: " + scost + " != " + softcost);
			}
		}
		p.destroy();
	}


	public Place getPosition(int e) {
		if (space.eventcost[e] > 0)
			return new Place(-1, -1);
		for (int i = 0; i < space.T; i++) {
			for (int j = 0; j < space.R; j++) {
				if (bestsolution[i][j] == e)
					return new Place(i, j);
			}
		}
		return new Place(-1, -1);
	}


	//assertion
	public boolean assertCostComputation() {
		space.assertAllHardCosts();
        return bestCost <= space.hardObj;
	}

	public static void main (String[] args) {
		if (args.length < 2) {
			System.out.println("Usage :");
			System.out.println("Example : java -jar solverPatat08.jar -file A -t 600 -seed 0 -nbruns 1 ");
			System.out.println("-file A: the path of the instance or the path of a directory");
			System.out.println("-t B: the time limit in seconds");
			System.out.println("-nbruns C: the number of run performed on the same instance by varying the seed");
            System.out.println("-colost : apply the coloring strategy for the SA");
            System.out.println("-saph : the prediction of the cooling is made on time (default). this option enforces a deterministic prediction ");
			System.out.println("-seed E: the seed (if several runs, then it is incremented by one for each run)");
			System.out.println("-feas: only look for a feasible solution (default is false)");
			System.out.println("-check: specify to run the c++ checker (compiled for MAC in the .jar) (default is false)");
   		} else {

            HashMap<String, String> options = new HashMap<String, String>();
            for (int i = 0; i < args.length; i++) {
                String arg = args[i];
                String val = "";
                if (i < args.length-1) {
                    if (!args[i+1].startsWith("-"))
                      val = args[++i];
                }
                options.put(arg, val);
            }
            boolean feas = false;
            boolean runChecker = false;
            boolean coloringSA = false;
            boolean hardp = false;
            int nbr = 1;
            int seed = 0;
            int timl;

            String instance;
            if (options.containsKey("-file")) {
                instance = options.get("-file");
            } else {
                throw new Error("file option -file is missing");
            }
            if (options.containsKey("-t")) {
                timl = Integer.parseInt(options.get("-t")) * 1000;
            } else {
                throw new Error("Timelimit option -t is missing");
            }
            if (options.containsKey("-nbruns")) {
               nbr  = Integer.parseInt(options.get("-nbruns"));
            } 
            if (options.containsKey("-saph")) {
               hardp = true;
            }
            if (options.containsKey("-seed")) {
               seed = Integer.parseInt(options.get("-seed"));
            }
            if (options.containsKey("-feas")) {
                feas = true;        
            }
            if (options.containsKey("-check")) {
                runChecker = true;            
            }
            if (options.containsKey("-colost")) {
                coloringSA = true;
            }

            SimulatedAnnealing.hardPrediction  = hardp;
            //trace = true;
            Benchmark.ech = nbr;
            Benchmark.hardsolverOnly = feas;
            Benchmark.twoStepHardSolverOnly = true; //default hard solver is LS-Colouring
            Benchmark.checkResults = runChecker;
            Benchmark.fourStepsSolver = coloringSA;
            Benchmark.seed = seed;
            Benchmark.timelimit = timl;

            File f = new File(instance);
			if (f.isDirectory()) {
				File[] files = InstanceReader.listAllFiles(f);
				for (int i = 0; i < files.length; i++) {
					File file = files[i];
					Benchmark.test((i+1),file.getPath());
				}
			} else {
				Benchmark.test(0,instance);							
			}
		}
	}


}
