package util;


import java.util.AbstractList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Random;
import java.util.RandomAccess;

/**
 * This class implements an array of integers that
 * behaves as a linked list and as a bitset. More precisely, all the following
 * operations are performed in constant time: add, remove,
 * membership, next/previous element, cardinality, initialisation, but
 * order is not constant over time (modified by deletion).<br>
 * In the current implementation, the capacity is fixed at initialisation,
 * and is not dynamically growable.
 * @author alexp
 *
 */
public class IntSet implements Iterable<Integer>{
	
	/**
	 * the index of an elements in the
	 * elements array.
	 */
	private int[] indexes;
	
	/**
	 * the elements of this set, that are
	 * consistent for the indexes from
	 * 0 to card - 1
	 */
	private int[] elements;
	
	/**
	 * The cardinality of this set
	 */
	private int card;
	
	/**
	 * The minimal value of this set.
	 */
	private int offset;
	
	private Random random = new Random();
	
	/**
	 * Set a random number generator used
	 * by this set.
	 * @see #randomValue()
	 * @param r a random number generator
	 */
	public void setRandom(Random r){
		random = r;
	}
	
	/**
	 * Builds a new set with values ranging from <code>0</code> to
	 * <code>capacity - 1</code>.<br>
	 * For the moment, the implemention does
	 * not allow this capacity to be dynamically increased
	 * when needed, so any addition greater than or equal
	 * to the capacity will result in an error.
	 * @param capacity the maximal cardinality of this set
	 */
	public IntSet(int capacity){
		indexes = new int[capacity];
		elements = new int[capacity];
	}
	
	/**
	 * Builds a new set with value ranging from <code>offset</code>
	 * to <code>offset+capacity-1</code>.<br>
	 * For the moment, the implemention does
	 * not allow this capacity to be dynamically increased
	 * when needed, so any addition greater than or equal
	 * to the capacity will result in an error.
	 * @param capacity the maximal cardinality of this set
	 * @param offset
	 */
	public IntSet(int capacity, int offset){
		this(capacity);
		this.offset = offset;
	}
	
	/**
	 * Creates a new set with the specified capacity and the
	 * elements of the specified set.
	 * @param set an existing set
	 * @param newCapacity the new capacity the new set should have.
	 * This new capacity should be large enough to hold all the
	 * elements of the specified set, i.e. the biggest element in <code>set</code>
	 * should be less than <code>set.min()+newCapacity</code>, but this is not
	 * verified
	 */
	public IntSet(IntSet set, int newCapacity){
		this.indexes = new int[set.indexes.length];
		System.arraycopy(set.indexes, 0, this.indexes, 0, this.indexes.length);
		this.elements = new int[set.elements.length];
		System.arraycopy(set.elements, 0, this.elements, 0, this.elements.length);
		this.offset = set.offset;
		this.card = set.card;
	}
	
	/**
	 * Creates a copy of the specified set.
	 * @param set a set to copy.
	 */
	public IntSet(IntSet set){
		this(set,set.capacity());
	}
	
	/**
	 * Creates and initialises a set with the specified values.<br>
	 * The capacity and the size are set to the minimal
	 * required to fit the given values.
	 * @param values the values that the set will contain
	 */
	public static IntSet withValues(int ... values){
		int min=Integer.MAX_VALUE,max=0;
		for(int val : values){
			min = Math.min(val, min);
			max = Math.max(val, max);
		}
		int capacity = max-min+1;
		IntSet set = new IntSet();
		set.indexes = new int[capacity];
		set.elements = new int[capacity];
		set.offset = min;
		for(int val : values) set.add(val);
		
		return set;
	}
	//Just to have it
	private IntSet(){}

	/**
	 * Membership test.
	 * @param i a value
	 * @return whether it belongs to the set
	 */
	public boolean contains(int i){
		return containsNoOffset(i-offset);
	}
	
	private boolean containsNoOffset(int i){
		//rangeCheckNoOffset(i);
		return i >= 0 && i < indexes.length
		&& indexes[i] < card && elements[indexes[i]] == i;
	}
	
	/**
	 * Indicates whether this set contains
	 * all of the elements in the specified set.
	 * @param s a set
	 * @return <code>true</code> if this set contains
	 * all of its elements.
	 * @throws IndexOutOfBoundsException if the given
	 * set contains an element out of the range
	 * of this set.
	 */
	public boolean containsAll(IntSet s){
		for(int i=s.first();i>-1;i=s.next(i))
			if(!contains(i))
				return false;
		return true;
	}
	
	/**
	 * Adds a new element to this set.
	 * @param i a value
	 * @return <code>true</code> if the set didn't contain
	 * the specified value
	 * @throws IndexOutOfBoundsException if
	 * the value in invalid, ie negative or greater than
	 * or equal to the maximal capacity.
	 */
	public boolean add(int i){
		return addNoOffset(i-offset);
	}
	
	private boolean addNoOffset(int i){
		rangeCheckNoOffset(i);
		if(!containsNoOffset(i)){
			elements[card] = i;
			indexes[i] = card;
			card++;
			return true;
		}
		return false;
	}
	
	/**
	 * Removes an elements from this set.
	 * @param i a value
	 * @return <code>true</code> if the set contained
	 * the specified value
	 * @throws IndexOutOfBoundsException if
	 * the value in invalid, ie negative or greater than
	 * or equal to the maximal capacity. 
	 */
	public boolean remove(int i){
		return removeNoOffset(i-offset);
	}
	
	private boolean removeNoOffset(int i){
		rangeCheckNoOffset(i);
		if(containsNoOffset(i)){
			card--;
			elements[indexes[i]] = elements[card];
			indexes[elements[card]] = indexes[i];
			return true;
		}
		return false;
	}
	
	/**
	 * Removes all the elements from this set.
	 */
	public void clear(){
		indexes = new int[indexes.length];
		elements = new int[elements.length];
		card = 0;
	}
	
    /**
     * Indicates if this set contains no elements.
     *
     * @return <tt>true</tt> if this set contains no elements
     */
	public boolean isEmpty(){
		return card == 0;
	}
	
	/**
	 * Returns the current size of elements
	 * contained in this set.
	 * @return the cardinality of this set
	 */
	public int size(){
		return card;
	}
	
	/**
	 * Returns a random value from this set.
	 * @return a random value from this set.
	 */
	public int randomValue(){
		int index = random.nextInt(card);
		return elements[index]+offset;
	}
	
	/**
	 * Returns the capacity of this set, i.e.
	 * the maximal range of numbers it can hold.
	 * @return the capacity of this set
	 */
	public int capacity(){
		return indexes.length;
	}
	
	/**
	 * Returns the minimal value this set <b>may</b> hold.
	 * @return the minimal possible value of this set,
	 * and <b>not</b> the current minimal value
	 * @see #first()
	 */
	public int min(){
		return offset;
	}
	
	/**
	 * Returns the maximal value this set <b>may</b> hold.
	 * @return the maximal possible value of this set,
	 * and <b>not</b> the current maximal value
	 * @see #last() 
	 */
	public int max(){
		return offset+indexes.length-1;
	}
	
	/**
	 * Returns the first element of this set.
	 * Note that the order has no particular semantics,
	 * this method is to be used only for iterating
	 * over the set.
	 * @return the first element, -1 if this
	 * set is empty
	 */
	public int first(){
		int first;
		return (first=firstNoOffset())==-1?first:first+offset;
	}
	
	private int firstNoOffset(){
		if(card == 0) return -1;
		return elements[0];
	}
	
	/**
	 * Returns the last element of this set.
	 * Note that the order has no particular semantics,
	 * this method is to be used only for iterating
	 * over the set.
	 * @return the last element, -1 if this
	 * set is empty
	 */
	public int last(){
		int last;
		return (last=lastNoOffset())==-1?last:last+offset;
	}
	
	private int lastNoOffset(){
		if(card == 0) return -1;
		return elements[card-1];
	}
	
	/**
	 * Returns the element in this set next
	 * to the element specified. The specified element
	 * should belong to the set. If it doesn't, no error
	 * is raised, but the result
	 * might be unpredictable.
	 * Note that the order has no particular semantics,
	 * this method is to be used only for iterating
	 * over the set.
	 * @param i a value
	 * @return the next element or -1 if there is no
	 * such element
	 */
	public int next(int i){
		assert(contains(i));
		int next;
		//return (next=nextNoOffset(i-offset)+offset)==offset-1?-1:next;
		return (next=nextNoOffset(i-offset))==-1?next:next+offset;//equivalent
	}
	
	private int nextNoOffset(int i){
		rangeCheckNoOffset(i);
		if(indexes[i] == card - 1) return -1; 
		return elements[indexes[i]+1];
	}
	
	/**
	 * Returns the element in this set previous
	 * to the element specified. The specified element
	 * should belong to the set. If it doesn't, no error
	 * is raised, but the result
	 * might be unpredictable.
	 * Note that the order has no particular semantics,
	 * this method is to be used only for iterating
	 * over the set.
	 * @param i a value
	 * @return the previous element, or -1 if there is no
	 * such element.
	 */
	public int previous(int i){
		assert(contains(i));
		int previous;
		return (previous=previousNoOffset(i-offset))==-1?previous:previous+offset;
	}
	
	private int previousNoOffset(int i){	
		rangeCheckNoOffset(i);
		if(indexes[i] == 0) return -1;
		return elements[indexes[i]-1];
	}
	
	protected void rangeCheckNoOffset(int i){
		if(i < 0 || i >= indexes.length)
			throw new IndexOutOfBoundsException(
					"Value="+(i+offset)+
					" Index="+i+", Size="+indexes.length);
	}
	
	public Iterator<Integer> iterator(){
		return new LISIt(this);
	}
	
	/**
	 * Adds all the elements of the specified set
	 * in this set. Warning: the elements of this set
	 * and the specified set must be compatible in that
	 * the membership in one set of any element of the
	 * other set should raise no exception.
	 * @param s a set
	 * @return <code>true</code> if the set was modified
	 * after the call
	 */
	public boolean addAll(IntSet s){
		boolean modified = false;
		for(int v=s.first();v>-1;v=s.next(v))
			if(add(v)) modified = true;
		return modified;
	}
	
	/**
	 * Removes all the elements of the specified set
	 * in this set. Warning: the elements of this set
	 * and the specified set must be compatible in that
	 * the membership in one set of any element of the
	 * other set should raise no exception.
	 * @param s a set
	 * @return <code>true</code> if the set was modified
	 * after the call
	 */
	public boolean removeAll(IntSet s){
		boolean modified = false;
		for(int v=s.first();v>-1;v=s.next(v))
			if(remove(v)) modified = true;
		return modified;
	}
	
	/**
	 * Retains in this set only the elements of
	 * the specified set.
	 * @param s a set
	 * @return <code>true</code> if the set was modified
	 * after the call
	 */
	public boolean retainAll(IntSet s){
		boolean modified = false;
		for(Iterator<Integer> it=iterator();it.hasNext();)
			if(!s.contains(it.next())){
				it.remove();
				modified = true;
			}
		return modified;
	}
	
	/**
	 * Checks the equality between this set and the
	 * specified object. The equality holds unless:<br>
	 * <li> the object is not a <code>IntSet</code>,
	 * <li> the cardinality is not the same,
	 * <li> the minimal value is not the same,
	 * <li> the maximal value is not the same,
	 * <li> the two sets do not contain strictly the same elements.
	 */
	public boolean equals(Object o){
		if (o == this)
		    return true;
		if (!(o instanceof IntSet))
		    return false;
		
		IntSet s = (IntSet)o;
		
		if(card != s.card) return false;
		
		if(offset != s.offset) return false;
		
		if(indexes.length != s.indexes.length)
			return false;
		
		for(int i=first();i>=0;i=next(i))
			if(!s.contains(i)) return false;
		
		return true;
		
	}
	
	/**
	 * Fills the provided array with the elements of this
	 * collection, in the order they would have been returned
	 * by the iterator. If the set does not fit in the array,
	 * a new one is allocated an returned.
	 * @param array an array
	 * @return an array containing the elements of this
	 * set
	 */
	public int[] toArray(int[] array){
		if(array.length < card) array = new int[card];
		int i=0;
		for(int e=first();e>=0;e=next(e)) array[i++] = e;
		return array;
	}
	
	/**
     * Returns a fixed-size list backed by this set.<BR>
     * It is hazardous to use this list as it can make
     * the implementation inconsistent, for example,
     * a {@link List#set(int, Integer)} call with
     * an element already contained in this set
     * can have unpredictable results.
     * @return a list view of this set.
	 */
	public List<Integer> asList(){
		if(listView == null)
			listView = new SetList(this);
		return listView;
	}
	private List<Integer> listView;
	
	public String toString(){
		StringBuffer buffer = new StringBuffer(8*card + 2);
		String separator = "";
		buffer.append('{');

		for (int i = first();i>-1;i = next(i)) {
			buffer.append(separator);
			separator = ", ";
			buffer.append(i);
        }

		buffer.append('}');
		return buffer.toString();
	}
	
	private static class LISIt implements Iterator<Integer>{
		
		private IntSet set;
		private int curElem, lastReturned = -1;
		
		LISIt(IntSet set){
			this.set = set;
			curElem = set.first();
		}
		
		public boolean hasNext() {
			return curElem > -1;
		}

		public Integer next() {
			if(!hasNext()) throw new NoSuchElementException();
			lastReturned = curElem;
			curElem = set.next(curElem);
			return lastReturned;
		}

		public void remove() {
			if(lastReturned == -1)
				throw new IllegalStateException();
						
			set.remove(lastReturned);
			if(curElem > -1) curElem = set.elements[set.card]+set.offset;
			lastReturned = -1;
		}		
		
	}
	
	private static class SetList extends AbstractList<Integer>
	implements RandomAccess{
		
		private IntSet e;
		
		SetList(IntSet elements){
			e = elements;
		}

		private void rangeCheck(int i){
			if(i < 0 || i >= e.card)
				throw new IndexOutOfBoundsException("Index = "+i+" size = "+e.card);
		}
		
		public Integer get(int index) {
			rangeCheck(index);
			return e.elements[index]+e.offset;
		}
		
		public Integer set(int index, Integer v){
			Integer c = e.elements[index]+e.offset;
			e.elements[index] = v - e.offset;
			e.indexes[v-e.offset] = index;
			return c;
		}
		
		public boolean contains(Object v){
			if(!(v instanceof Integer)) return false;
			return e.contains((Integer)v);
		}

		public int size() {
			return e.card;
		}
		
	}
	
}
